import logging
import pickle
import random
import re
import subprocess
from dataclasses import dataclass
from typing import List, NamedTuple, Tuple, Union

import torch.nn
from expred import BertConfig
from transformers import BertTokenizer
from config import SequenceMiningConfig
from logging_utils import get_logger
from utils import (abs_diff, do_prediction, grep_tuple,
                   join_sentences_to_docs, locate_substr, locate_tuple,
                   sentence_seek, split_doc_to_sentences)

logger = get_logger('candidate_sequences', logging.INFO)


class MaskedContext(NamedTuple):
    context: List[str]
    mask_start: int
    mask_end: int


@dataclass
class CandidateSequence():
    seq: List[str]

    _res_docs = None
    _masked_res_context = None
    _masked_res_context_preds = None
    _selected_masked_context = None

    synthetic_preds = None
    
    mask_char = SequenceMiningConfig.CAUSAL_MASK_CHAR

    @property
    def res_docs(self) -> List[List[str]]:
        if self._res_docs is not None:
            return self._res_docs
        joined_seq = ' '.join(self.seq)
        command = ['grep', '-wF', '--', joined_seq, self.working_dataset_fname]
        res_docs = subprocess.run(command, stdout=subprocess.PIPE).stdout.decode('utf-8').strip().split('\n')
        # res_docs = [re.sub(re.escape(self.rationale_mask_char), self.mask_char, doc).split() for doc in res_docs]
        self._res_docs = [doc_str.split() for doc_str in res_docs]
        return self._res_docs

    @res_docs.setter
    def res_docs(self, res_docs: List[List[str]]):
        self._res_docs = res_docs

    def mask(self, context, start, end):
        mask = [self.mask_char] * (end - start)
        masked_context = context[:start] + mask + context[end:]
        return masked_context

    def get_masked_doc(self, doc: List[str]) -> MaskedContext:
        start, end = locate_substr(doc, self.seq)
        masked_doc = self.mask(doc, start, end)
        return MaskedContext(context=masked_doc, mask_start=start, mask_end=end)

    def get_masked_sentences(self, doc: List[str]) -> MaskedContext:
        start, end = locate_substr(doc, self.seq)
        sentences = split_doc_to_sentences(doc)

        try:
            sent_offset_st, sent_idx_st = sentence_seek(sentences, start, sentence_cnt=-self.d_context_sentence_ba[0])
        except IndexError:
            print('start')
            print(doc, sentences, start, self.seq)
            raise IndexError
        try:
            _, sent_idx_ed = sentence_seek(
                sentences, end, sent_idx_st=sent_idx_st, sentence_cnt=self.d_context_sentence_ba[1])
        except IndexError:
            print('end')
            print(doc, sentences, end, sent_idx_st, self.seq)
            raise IndexError

        start -= sent_offset_st
        end -= sent_offset_st

        res_sent = join_sentences_to_docs(
            sentences[sent_idx_st: sent_idx_ed + 1])
        masked_sentences = self.mask(res_sent, start, end)
        return MaskedContext(context=masked_sentences, mask_start=start, mask_end=end)

    @property
    def masked_res_context(self) -> List[MaskedContext]:
        if self._masked_res_context is not None:
            return self._masked_res_context
        masked_res_context = []
        for d in self.res_docs:
            if self.sentences_as_context:
                masked_res_context.append(self.get_masked_sentences(d))
            else:
                masked_res_context.append(self.get_masked_doc(d))
        self._masked_res_context = masked_res_context
        return self._masked_res_context

    @masked_res_context.setter
    def masked_res_context(self, context):
        self._masked_res_context = context

    @property
    def masked_res_context_preds(self) -> List[Tuple[float, float]]:
        if self._masked_res_context_preds is not None:
            return self._masked_res_context_preds
        context = [[d.context] for d in self.masked_res_context]
        logger.debug(f'predicting masked context for {self.seq}')
        logger.debug(f'#contexts = {len(context)}')
        logger.debug(f'the first context starts with {context[0][0][:10]}')
        with torch.no_grad():
            self.model.eval()
            self._masked_res_context_preds = do_prediction(self.model,
                                                           self.model_config, 
                                                           self.tokenizer,
                                                           context, 
                                                           batch_size=self.batch_size,
                                                           verbose=False)[0]
        return self._masked_res_context_preds

    @masked_res_context_preds.setter
    def masked_res_context_preds(self, preds):
        self._masked_res_context_preds = preds

    @classmethod
    def loads(cls, loaded_strs):
        if len(loaded_strs) == 5:
            seq, res_docs, masked_res_context, selected_masked_context = map(
                pickle.loads, loaded_strs[:-1])
            masked_res_context_preds = loaded_strs[-1]
        else:
            seq, res_docs, masked_res_context, selected_masked_context = map(
                pickle.loads, loaded_strs)
        ret = cls(seq)
        ret.res_docs = res_docs
        ret.masked_res_context = masked_res_context
        ret.selected_masked_context = selected_masked_context
        if len(loaded_strs) == 5:
            ret.masked_res_context_preds = masked_res_context_preds
        return ret
    
    def dumps(self) -> List[bytes]:
        ret = [pickle.dumps(self.seq),
               pickle.dumps(self.res_docs),
               pickle.dumps(self.masked_res_context),
               pickle.dumps(self.selected_masked_context)]
        if self.masked_res_context_preds is not None:
            ret.append(self.masked_res_context_preds)
        return ret

    @property
    def selected_masked_context(self) -> List[MaskedContext]:
        if self._selected_masked_context is not None:
            return self._selected_masked_context
        neutral_masked_context = []
        
        if self.context_criterion == 'sampling':
            neutral_masked_context = random.sample(self.masked_res_context, 
                                                   int(len(self.masked_res_context) * self.sampling_rate))
        elif self.context_criterion == 'all':
            neutral_masked_context = self.masked_res_context
        else:
            for d, pred in zip(self.masked_res_context, self.masked_res_context_preds):
                if self.context_criterion(pred):
                    neutral_masked_context.append(d)
        
        self._selected_masked_context = neutral_masked_context
        return self._selected_masked_context

    @selected_masked_context.setter
    def selected_masked_context(self, s):
        self._selected_masked_context = s

    def init_function(self, dataset_name: str, working_dataset_fname: str,
                 model: torch.nn.Module, model_config: BertConfig, tokenizer: BertTokenizer,
                 sentences_as_context: bool, device: Union[str, torch.device],
                 context_criterion: str, batch_size: int, sampling_rate: float = None,
                 d_context_sentence_ba: Tuple[int, int]=None):
        
        self.working_dataset_fname = working_dataset_fname
        self.dataset_name = dataset_name
        self.model_config, self.tokenizer, self.model = model_config, tokenizer, model
        self.sentences_as_context = sentences_as_context
        self.device = device

        if context_criterion == 'neutral':
            self.context_criterion = abs_diff
        elif context_criterion == 'all':
            self.context_criterion = context_criterion
        elif context_criterion == 'sampling':
            self.context_criterion = context_criterion
            self.sampling_rate = sampling_rate
        else:
            raise ValueError(
                'context criterion must in {\'all\', \'abs_diff\', \'sampling\'}')
            
        self.d_context_sentence_ba = d_context_sentence_ba if d_context_sentence_ba is not None else (0, 0)
        
        self.batch_size = batch_size
        self.selected_masked_context = self.selected_masked_context


class MaskedTupleContext(NamedTuple):
    context: Tuple[List[str]]
    mask_start_query: int
    mask_end_query: int
    mask_start_doc: int
    mask_end_doc: int


@dataclass
class CandidateSequenceTuple:
    qd_tuple: Tuple[List[str], List[str]] = None

    _res_instances = None
    _masked_res_context = None
    _masked_res_context_preds = None
    _selected_masked_context = None
    
    mask_char = SequenceMiningConfig.CAUSAL_MASK_CHAR
    
    @property
    def seq_key(self) -> str:
        return ' '.join(self.qd_tuple[0]) + ' [SEP] ' + ' '.join(self.qd_tuple[1])

    @property
    def res_instances(self) -> List[Tuple[List[str], List[str]]]:
        if self._res_instances is not None:
            return self._res_instances
        joined_tuple = [' '.join(l) for l in self.qd_tuple]
        res_instances = grep_tuple(
            *joined_tuple, self.working_dataset_fname, count_only=False)

        res_instances = [doc.split() for doc in res_instances]
        temp = []
        for doc in res_instances:
            try:
                temp.append((
                    doc[:doc.index('[SEP]')],
                    doc[doc.index('[SEP]')+1:]))
            except ValueError:
                logger.error(f'ValueError, doc = {doc}')
                logger.error(f'res_instance = {res_instances}')
                logger.error(f'qd_tuple = {self.qd_tuple}')
                logger.error(f'working_dataset_fname = {self.working_dataset_fname}')
                raise ValueError
        # res_instances = [(doc[:doc.index('[SEP]')],
                        #   doc[doc.index('[SEP]')+1:]) for doc in res_instances]
        res_instances = temp
        self._res_instances = res_instances
        return self._res_instances

    @res_instances.setter
    def res_instances(self, res_docs: List[Tuple[List[str], List[str]]]):
        self._res_instances = res_docs

    def _mask(self, query, doc, q_start, q_end, d_start, d_end):
        q_mask = [self.mask_char] * (q_end - q_start)
        d_mask = [self.mask_char] * (d_end - d_start)
        masked_qdtuple = (query[:q_start] + q_mask + query[q_end:],
                          doc[:d_start] + d_mask + doc[d_end:])
        return masked_qdtuple

    def _get_masked_query_doc(self, instance: Tuple[List[str], List[str]]) -> MaskedTupleContext:
        q_start, q_end, d_start, d_end = locate_tuple(instance, *self.qd_tuple)
        masked_instance = self._mask(*instance, q_start, q_end, d_start, d_end)
        return MaskedTupleContext(context=masked_instance,
                                  mask_start_query=q_start,
                                  mask_end_query=q_end,
                                  mask_start_doc=d_start,
                                  mask_end_doc=d_end)

    def _get_masked_query_sentence(self, instance: Tuple[List[str], List[str]]) -> MaskedTupleContext:
        q_seq, d_seq = self.qd_tuple
        q_res, d_res = instance
        q_start, q_end, d_start, d_end = locate_tuple(instance, q_seq, d_seq)
        sentences = split_doc_to_sentences(d_res)

        try:
            sent_offset_st, sent_idx_st = sentence_seek(sentences, d_start)
        except IndexError:
            print('error while seeking d_start:')
            print(instance, sentences, d_start, self.qd_tuple)
            raise IndexError
        try:
            _, sent_idx_ed = sentence_seek(
                sentences, d_end, sent_idx_st=sent_idx_st)
        except IndexError:
            print('error while seeking d_end:')
            print(instance, sentences, d_end, sent_idx_st, self.qd_tuple)
            raise IndexError

        s_start = d_start - sent_offset_st
        s_end = d_end - sent_offset_st

        res_sent = join_sentences_to_docs(
            sentences[sent_idx_st: sent_idx_ed + 1])
        masked_qs_tuple = self._mask(
            query=q_res, doc=res_sent, q_start=q_start, q_end=q_end, d_start=s_start, d_end=s_end)
        return MaskedTupleContext(context=masked_qs_tuple,
                                  mask_start_query=q_start,
                                  mask_end_query=q_end,
                                  mask_start_doc=s_start,
                                  mask_end_doc=s_end)

    @property
    def masked_res_context(self) -> List[MaskedTupleContext]:
        if self._masked_res_context is not None:
            return self._masked_res_context
        masked_res_context: List[MaskedTupleContext] = []
        for ins in self.res_instances:
            if self.sentences_as_context:
                masked_res_context.append(self._get_masked_query_sentence(ins))
            else:
                masked_res_context.append(self._get_masked_query_doc(ins))
        self._masked_res_context = masked_res_context
        return self._masked_res_context

    @masked_res_context.setter
    def masked_res_context(self, context):
        self._masked_res_context = context

    # TODO: the current version makes one prediction per turple, inefficient
    def masked_res_context_preds(self) -> List[Tuple[float, float]]:
        if self._masked_res_context_preds is not None:
            return self._masked_res_context_preds
        queries = [d.context[0] for d in self.masked_res_context]
        docs = [[d.context[1]] for d in self.masked_res_context]
        logger.debug(f'predicting masked context for {self.qd_tuple}')
        logger.debug(f'#contexts = {len(queries)}')
        logger.debug(f'the first context starts with {queries[0]} - {docs[0][0][:10]}')
        with torch.no_grad():
            self.model.eval()
            self._masked_res_context_preds = do_prediction(self.model, self.model_config, self.tokenizer,
                                                           docs_or_dataset=docs,
                                                           batch_size=self.batch_size,
                                                           queries_or_none=queries,
                                                           tqdm_desc='neuruality check',
                                                           verbose=False)[0]
        return self._masked_res_context_preds

    @property
    def selected_masked_context(self) -> List[MaskedTupleContext]:
        if self._selected_masked_context is not None:
            return self._selected_masked_context
        if self.context_criterion == 'all':
            return self.masked_res_context
        if self.context_criterion == 'sampling':
            return random.sample(self.masked_res_context,
                                 int(len(self.masked_res_context) * self.sampling_rate))
        selected_masked_context = []
        for d, pred in zip(self.masked_res_context, self.masked_res_context_preds()):
            if self.context_criterion(pred):
                selected_masked_context.append(d)
        self._selected_masked_context = selected_masked_context
        return self._selected_masked_context

    @selected_masked_context.setter
    def selected_masked_context(self, value: List[MaskedTupleContext]):
        self._selected_masked_context = value

    @classmethod
    def loads(cls, loaded_strs):
        if len(loaded_strs) == 5:
            qd_tuple, res_instances, masked_res_context, selected_masked_context = map(
                pickle.loads, loaded_strs[:-1])
            masked_res_context_preds = loaded_strs[-1]
        else:
            qd_tuple, res_instances, masked_res_context, selected_masked_context = map(
                pickle.loads, loaded_strs)
        ret = cls(qd_tuple)
        ret.res_instances = res_instances
        ret.masked_res_context = masked_res_context
        ret.selected_masked_context = selected_masked_context
        if len(loaded_strs) == 5:
            ret._masked_res_context_preds = masked_res_context_preds
        return ret

    def dumps(self) -> List[bytes]:
        ret = [pickle.dumps(self.qd_tuple),
               pickle.dumps(self.res_instances),
               pickle.dumps(self.masked_res_context),
               pickle.dumps(self.selected_masked_context)]
        if self._masked_res_context_preds is not None:
            ret.append(self._masked_res_context_preds)
        return ret

    def init_functions(self, dataset_name: str, working_dataset_fname: str,
                       sentences_as_context: bool, device: Union[str, torch.device],
                       context_criterion: str, batch_size: int, sampling_rate: float = None,
                       model: torch.nn.Module=None, model_config: BertConfig=None, tokenizer: BertTokenizer=None):

        self.working_dataset_fname = working_dataset_fname
        self.dataset_name = dataset_name
        self.model_config, self.tokenizer, self.model = model_config, tokenizer, model
        self.sentences_as_context = sentences_as_context
        self.device = device

        if context_criterion == 'neutral':
            self.context_criterion = abs_diff
        elif context_criterion == 'all':
            self.context_criterion = context_criterion
        elif context_criterion == 'sampling':
            self.context_criterion = context_criterion
            self.sampling_rate = sampling_rate
        else:
            raise ValueError(
                'neutrality criterion can only be one of \'all\' or \'neutral\'')

        self.batch_size = batch_size

        self._res_instances = self.res_instances
        self._masked_res_context = self.masked_res_context
        self._selected_masked_context = self.selected_masked_context
