import csv
import functools
from itertools import chain
import os
import pickle
import time
from multiprocessing import Pool
from typing import Dict, List, Tuple, Union

import numpy as np
import torch
import tqdm

from candidate_sequence import CandidateSequence, CandidateSequenceTuple
from config import SequenceMiningConfig
from logging_utils import get_logger
from synthetics import (SimplifiedSyntheticInputs, SyntheticInputs,
                        SyntheticTupleInputs)
from utils import get_model, mean_criterion, mean_stddev_criterion

logger = get_logger('causal_sequences')


class CausalBase:
    DUMP_EVERY = 50
    PREDICT_EVERY = 50
    
    _syn_inputs = None
    _rules = None
    _non_causal_syns = None
    
    def __init__(self, input_fnames: Union[str, List[str]], conf: SequenceMiningConfig):
        sub_folder = time.asctime().replace(' ', '_').lower()[4:] if conf.CS_SUB_FOLDER is None else conf.CS_SUB_FOLDER
        self.working_dataset_fname = f'{conf.WORKING_DIR}/sequences.txt'
        self.output_dir = f'{conf.WORKING_DIR}/causal_seq_res/{sub_folder}/'
        self.cache_dir = f'{self.output_dir}/temp/'
        os.makedirs(self.cache_dir, exist_ok=True)
        
        if isinstance(input_fnames, str):
            input_fnames = [input_fnames]
        self.sequence_fnames = input_fnames
        
        self.conf = conf

        self.dataset_name = conf.DATASET_NAME
        self.base_dataset_name = conf.BASE_DATASET_NAME
        self.model_name = conf.MODEL_NAME
        self.batch_size = conf.BATCH_SIZE * 2
        self.num_workers = conf.N_WORKERS
        self.device = conf.DEVICE
                
        (
            self.model_config, 
            self.tokenizer,
            self.model
        ) = get_model(
            model_name=self.model_name,
            pretrained_dataset_name=self.dataset_name,
            base_dataset_name=self.base_dataset_name,
            device=self.device
            )
        if self.model_name == 'expred':
            self.model = self.model.cls_module
            
        if conf.CAUSAL_CRITERION == 'mean':
            self.causal_criterion = mean_criterion
        elif conf.CAUSAL_CRITERION == 'mean_stddev':
            self.causal_criterion = mean_stddev_criterion
        else:
            raise ValueError(f'Causal criterion {conf.CAUSAL_CRITERION} is not supported')
        self.context_criterion = conf.CONTEXT_CRITERION

        self.sentences_as_context = conf.SENTENCES_AS_CONTEXT
        self.causal_check_min_support = conf.CAUSAL_CHECK_MIN_SUPPORT

        self.force_recalculate_rules = conf.force_recalculate_rules
        self.candidate_sequences_cache_fname = f'{self.cache_dir}/candidate_cache.pkl'
        
    @staticmethod
    def ok_with_npmi_check(line: List):
        # _, _, _, _, npmi0, npmi1, pred0, pred1 = line
        npmi0, npmi1, pred0, pred1 = line[-4:]
        if npmi1 < npmi0 and pred1 < pred0:
            return True
        if npmi0 < npmi1 and pred0 < pred1:
            return True
        return False
    
    def is_causal(self, s: Union[SyntheticTupleInputs,
                                 SyntheticInputs,
                                 SimplifiedSyntheticInputs]) -> bool:
        if len(s.preds) == 0:
            return False
        return self.causal_criterion(s.preds, mean_threshold=self.conf.MEAN_THRESHOLD)
        
    @classmethod
    def load_rules(cls, rules_fname: str) -> Dict[Union[Tuple[Tuple[str], Tuple[str]], str], np.ndarray]:
        ret = {}
        with open(rules_fname, 'r') as fin:
            for rule in fin:
                seq, preds = rule.split('->')
                if '[SEP]' in seq:
                    query, doc = seq.split('[SEP]')
                    seq = (tuple(query.strip().split()), tuple(doc.strip().split()))
                else:
                    seq = tuple(seq.split())
                ret[seq] = np.array(eval(preds))
        return ret
    
    @property
    def syn_inputs(self) -> Dict:
        raise NotImplementedError
    
    @property
    def rules(self):
        if self._rules is not None:
            return self._rules
        self._rules, self._non_causal_syns = dict(), dict()
        for k, syn in tqdm.tqdm(self.syn_inputs.items(), desc='causality check'):
            if self.is_causal(syn):
                logger.info(f'sequence "{k}" is causal')
                self._rules[k] = syn
            else:
                logger.info(f'sequence "{k}" is non-causal')
                self._non_causal_syns[k] = syn
        return self._rules
    
    @rules.setter
    def rules(self, r: Dict[str, Union[SyntheticInputs, SyntheticTupleInputs, SimplifiedSyntheticInputs]]):
        self._rules = r
        
    def write_rules(self):
        if not os.path.isfile(f'{self.output_dir}/rules.txt') or self.force_recalculate_rules:
            logger.info(f'writing rules to {self.output_dir}/rules.txt')
            os.makedirs(self.output_dir, exist_ok=True)
            self.conf.dump_to(self.output_dir)
            with open(f'{self.output_dir}/rules.txt', 'w+') as fout:
                for k, v in self.rules.items():
                    fout.write(f"{k} -> {[p.cpu().tolist() if isinstance(p, torch.Tensor) else p for p in v.preds]}\n")
        else:
            logger.info(f'{self.output_dir}/rules.txt exists, will not overwrite.')


class CausalSequences(CausalBase):
    _candidate_sequences:List[CandidateSequence] = None
    
    def __init__(self, input_fnames: Union[str, List[str]], conf: SequenceMiningConfig):
        super().__init__(input_fnames, conf)
        self.syn_inputs_cache_fname = f'{self.cache_dir}/cs_temp.pkl'
        self.focus_strings = conf.focus_strings
        self.force_invalidate_cache = False
    
    def load_candidates(self, fname):
        sequences = []
        with open(fname, 'r') as fin:
            for line_idx, _ in enumerate(fin.readlines()):
                pass
        file_length = line_idx - 1  # minus 1 for the header
    
        with open(fname, 'r') as fin:
            csv_reader = csv.reader(fin, delimiter='\t')
            next(csv_reader)  # skip the header
            for line in tqdm.tqdm(csv_reader, total=file_length, desc='load candidate'):
                if self.ok_with_npmi_check(line):
                    seq_str = line[0].strip()
                    cs = CandidateSequence(seq=seq_str.split())
                    cs.init_function(    
                        dataset_name=self.dataset_name,
                        working_dataset_fname=self.working_dataset_fname,
                        model=self.model,
                        model_config=self.model_config,
                        tokenizer=self.tokenizer,
                        sentences_as_context=self.sentences_as_context,
                        device=self.device,
                        context_criterion=self.context_criterion,
                        batch_size=self.batch_size,
                        sampling_rate=self.conf.CONTEXT_SAMPLING_RATE,
                        d_context_sentence_ba=self.conf.DOC_CONTEXT_SENTENCE_BA)
                    sequences.append(cs)
        return sequences

    @staticmethod
    def _load_candidate_sequences_from_cache(candidate_sequences_cache_fname) -> List[CandidateSequence]:
        loaded = []
        with open(candidate_sequences_cache_fname, 'rb') as fin:
            while True:
                try:
                    l = pickle.load(fin)
                except EOFError:
                    break
                loaded.append(CandidateSequence.loads(l))
        logger.info(f'candidate tuples loaded from cache file: {candidate_sequences_cache_fname}')
        return loaded

    def _dump_candidate_sequences_to_cache(self):
        if self.force_invalidate_cache or not os.path.isfile(self.candidate_sequences_cache_fname):
            with open(self.candidate_sequences_cache_fname, 'wb+') as fout:
                for cs in self._candidate_sequences:
                    pickle.dump(cs.dumps(), fout)

    @property
    def candidate_sequences(self) -> List[CandidateSequence]:
        if self._candidate_sequences is not None:
            return self._candidate_sequences
        
        if os.path.isfile(self.candidate_sequences_cache_fname):
            self._candidate_sequences = self._load_candidate_sequences_from_cache(self.candidate_sequences_cache_fname)
            return self._candidate_sequences
        
        candidate_sequences = []
        
        if self.num_workers != 1:
            logger.warning('num_worker > 1 in the causal sequence identifier is not supported! '
                           'Will set num_worker to 1.')
            self.num_workers = 1
        
        for fname in self.sequence_fnames:
            candidate_sequences += self.load_candidates(fname)
        
        self._candidate_sequences = candidate_sequences
        self._dump_candidate_sequences_to_cache()
        return self._candidate_sequences

    def maybe_load_syn_inputs_cache(self) -> Dict:
        if os.path.isfile(self.syn_inputs_cache_fname):
            logger.info('loading synthetic input cache...')
            with open(self.syn_inputs_cache_fname, 'rb') as fin:
                loaded = pickle.load(fin)
                logger.info('synthetic input cache loaded')
                if isinstance(loaded, tuple):  # compatibility
                    return loaded[0]
                return loaded
        return dict()

    def dump_syn_inputs_cache(self):
        with open(self.syn_inputs_cache_fname, 'wb+') as fout:
            to_dump = {k: v.simplified if isinstance(v, SyntheticInputs) else v for k, v in self._syn_inputs.items()}
            pickle.dump(to_dump, fout)
    
    @property
    def syn_inputs(self) -> Dict[str, SyntheticInputs]:
        if self._syn_inputs is not None:
            return self._syn_inputs
        
        self._syn_inputs = self.maybe_load_syn_inputs_cache()
        logger.info(f'generating synthetic inputs for {len(self.candidate_sequences)} candidate_sequences...')
        
        to_predict_keys = []
        
        for cs in tqdm.tqdm(self.candidate_sequences, desc='syn_inst_property'):
            
            seq_key = ' '.join(cs.seq)
            
            if self.focus_strings is not None \
                and seq_key not in self.focus_strings:
                continue
            
            if seq_key in self._syn_inputs:
                self._syn_inputs[seq_key] = SyntheticInputs.from_simplified(
                    simp=self._syn_inputs[seq_key],
                    # model=self.model,
                    # model_config=self.model_config,
                    # tokenizer=self.tokenizer,
                    batch_size=self.batch_size)
            else:
                self._syn_inputs[seq_key] = SyntheticInputs.from_seq_context(
                    s=cs,
                    c=self.candidate_sequences,
                    # model=self.model,
                    # model_config=self.model_config,
                    # tokenizer=self.tokenizer,
                    batch_size=self.batch_size)
                
                to_predict_keys.append(seq_key)

            if len(to_predict_keys) % self.PREDICT_EVERY == 0 and len(to_predict_keys) > 0:
                logger.info('update syn preds')
                self.update_syn_preds(to_predict_keys)
                to_predict_keys = []

            if len(self._syn_inputs) % self.DUMP_EVERY == 0 and len(self._syn_inputs) > 0:
                self.dump_syn_inputs_cache()

        if len(to_predict_keys) > 0:
            self.update_syn_preds(to_predict_keys)

        self.dump_syn_inputs_cache()

        return self._syn_inputs
    
    def update_syn_preds(self, to_predict_keys):
        logger.debug(f'len(to_predict_keys) = {len(to_predict_keys)}')
        syn_inputs = list(chain.from_iterable([self._syn_inputs[k].syn_inputs for k in to_predict_keys]))
        logger.info(f'len(syn_inputs) = {len(syn_inputs)}')
        preds = SyntheticInputs._pred_syn_instances(
            syn_inputs=syn_inputs,
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            batch_size=self.batch_size)
        p = 0
        for k in to_predict_keys:
            num_synths = len(self._syn_inputs[k].syn_inputs)
            self._syn_inputs[k].preds = preds[p : p + num_synths]
            p += num_synths

# TODO: why can't this be integrated into the causal tuple validator?
def _multiprocessing_load_helper(dataset_name,
                             working_dataset_fname,
                             sentences_as_context,
                             context_criterion,
                             batch_size,
                             sampling_rate,
                             line: List):
    def ok_with_npmi_check(line):
        # q, d, _, _, _, npmi0, npmi1, pred0, pred1 = line
        npmi0, npmi1, pred0, pred1 = line[-4:]
        if npmi1 < npmi0 and pred1 < pred0:
            return True
        if npmi0 < npmi1 and pred0 < pred1:
            return True
        return False

    if ok_with_npmi_check(line):
        query = line[0].split()
        doc = line[1].split()
        cst = CandidateSequenceTuple(
            qd_tuple=(query, doc)
        )
        cst.init_functions(
            dataset_name=dataset_name,
            working_dataset_fname=working_dataset_fname,
            sentences_as_context=sentences_as_context,
            device='cpu',
            context_criterion=context_criterion,
            batch_size=batch_size,
            sampling_rate=sampling_rate)
        return cst
    return None


class CausalSequenceTuples(CausalBase):
    DUMP_EVERY = 20
    _candidate_tuples = None

    def __init__(self, input_fnames: Union[str, List[str]], conf: SequenceMiningConfig):
        super().__init__(input_fnames, conf)
        self.force_invalidate_cache = conf.FORCE_INVALIDATE_CACHE
        
        self.candidate_tuples_cache_fname = f'{self.cache_dir}/candidate_tuples.pkl'
        self.syn_inputs_cache_fname = f'{self.cache_dir}/syn_cache.pkl'
        
        self.sampling_rate = conf.CONTEXT_SAMPLING_RATE
        self.focus_tuples = conf.focus_tuples

    @staticmethod
    def _load_candidate_tuples_from_cache(candidate_tuples_cache_fname) -> List[CandidateSequenceTuple]:
        loaded = []
        with open(candidate_tuples_cache_fname, 'rb') as fin:
            while True:
                try:
                    cached_candidate_tuple = pickle.load(fin)
                except EOFError:
                    break
                loaded.append(CandidateSequenceTuple.loads(cached_candidate_tuple))
        logger.info(f'candidate tuples loaded from cache file: {candidate_tuples_cache_fname}')
        return loaded

    def _dump_candidate_tuple_to_cache(self):
        if self.force_invalidate_cache or not os.path.isfile(self.candidate_tuples_cache_fname):
            with open(self.candidate_tuples_cache_fname, 'wb+') as fout:
                for cst in self._candidate_tuples:
                    pickle.dump(cst.dumps(), fout)
        else:
            logger.warn(f'tuples are loaded, but will not override old cache file under {self.candidate_tuples_cache_fname}.')

    def invalidate_candidate_cache(self, cache_name: str):
        cache_fname = getattr(self, cache_name + '_cache_fname')
        os.remove(cache_fname)

    def _do_load(self, line: List):
        if self.ok_with_npmi_check(line):
            query = line[0].split()
            doc = line[1].split()
            sup_0 = int(line[3])
            sup_1 = int(line[4])
            tuple_key = (line[0].strip(), line[1].strip())
            if sup_0 + sup_1 < self.causal_check_min_support and (self.focus_tuples is None or tuple_key not in self.focus_tuples):
                return None
            cst = CandidateSequenceTuple(
                qd_tuple=(query, doc)
            )
            cst.init_functions(
                dataset_name=self.dataset_name,
                working_dataset_fname=self.working_dataset_fname,
                model=self.model,
                model_config=self.model_config,
                tokenizer=self.tokenizer,
                sentences_as_context=self.sentences_as_context,
                device=self.device,
                context_criterion=self.context_criterion,
                batch_size=self.batch_size,
                sampling_rate=self.conf.CONTEXT_SAMPLING_RATE)
            return cst
        return None

    def load_candidates(self, fname):
        with open(fname, 'r') as fin:
            for line_idx, _ in enumerate(fin.readlines()):
                pass
        file_length = line_idx - 1  # minus 1 for the header
        with open(fname, 'r') as fin:
            csv_reader = csv.reader(fin, delimiter='\t')
            next(csv_reader)  # skip the header
            if self.context_criterion in ['all', 'sampling']:
                lines = [_ for _ in csv_reader]
                with Pool(self.num_workers) as pool:
                    sequences = list(
                        tqdm.tqdm(
                            pool.imap_unordered(
                                functools.partial(
                                    _multiprocessing_load_helper,
                                    self.dataset_name,
                                    self.working_dataset_fname,
                                    self.sentences_as_context,
                                    self.context_criterion,
                                    self.batch_size,
                                    self.sampling_rate),
                                lines),
                            total=file_length,
                            desc='load candidate'
                        )
                    )
                sequences = list(filter(lambda x: x is not None, sequences))
            else:
                sequences = []
                for line in tqdm.tqdm(csv_reader, total=file_length, desc='load candidate'):
                    ret = self._do_load(line)
                    if ret is not None:
                        sequences.append(ret)
            return sequences

    @property
    def candidate_tuples(self) -> List[CandidateSequenceTuple]:
        if self._candidate_tuples is not None:
            return self._candidate_tuples
        if os.path.isfile(self.candidate_tuples_cache_fname):
            self._candidate_tuples = self._load_candidate_tuples_from_cache(
                self.candidate_tuples_cache_fname)
            return self._candidate_tuples
        self._candidate_tuples = []
        for fname in self.sequence_fnames:
            self._candidate_tuples += self.load_candidates(fname)
        self._dump_candidate_tuple_to_cache()
        return self._candidate_tuples

    @candidate_tuples.setter
    def candidate_tuples(self, value: List[CandidateSequence]):
        self._candidate_tuples = value

    def maybe_load_syn_inputs_cache(self) -> Dict:
        if os.path.isfile(self.syn_inputs_cache_fname):
            logger.info('loading synthetic instances cache...')
            with open(self.syn_inputs_cache_fname, 'rb') as fin:
                loaded = pickle.load(fin)
                logger.info('synthetic instances cache loaded')
                return loaded
        return dict()

    def dump_syn_inputs_cache(self):
        with open(self.syn_inputs_cache_fname, 'wb+') as fout:
            pickle.dump(self._syn_inputs, fout)

    @property
    def syn_inputs(self) -> Dict[str, SyntheticTupleInputs]:
        if self._syn_inputs is not None:
            return self._syn_inputs
        
        self._syn_inputs = self.maybe_load_syn_inputs_cache()
        logger.info(f'generating synthetic inputs for {len(self.candidate_tuples)} candidate tuples...')
        to_predict_keys = []

        for ct in tqdm.tqdm(self.candidate_tuples, desc='construct synthetic examples'):
            tuple_key = (' '.join(ct.qd_tuple[0]), ' '.join(ct.qd_tuple[1]))
            
            if self.focus_tuples is not None \
                and tuple_key not in self.focus_tuples:
                continue

            if ct.seq_key in self._syn_inputs and self._syn_inputs[ct.seq_key].preds is not None:
                continue

            self._syn_inputs[ct.seq_key] = SyntheticTupleInputs.from_seq_tuple_context(
                ct, self.candidate_tuples
            )
            to_predict_keys.append(ct.seq_key)

            if len(to_predict_keys) % self.PREDICT_EVERY == 0:
                self.update_syn_preds(to_predict_keys)
                to_predict_keys = []

            if (len(self._syn_inputs)) % self.DUMP_EVERY == 0 and len(self._syn_inputs) > 0:
                self.dump_syn_inputs_cache()

        if len(to_predict_keys) > 0:
            self.update_syn_preds(to_predict_keys)

        self.dump_syn_inputs_cache()
        return self._syn_inputs

    def update_syn_preds(self, to_predict_keys):
        logger.debug(f'len(to_predict_keys) = {len(to_predict_keys)}')
        syn_queries = list(chain.from_iterable([self._syn_inputs[k].syn_queries for k in to_predict_keys]))
        syn_docs = list(chain.from_iterable([self._syn_inputs[k].syn_docs for k in to_predict_keys]))
        logger.debug(f'len(syn_queries) = {len(syn_queries)}')
        logger.debug(f'len(syn_docs) = {len(syn_docs)}')
        preds = SyntheticTupleInputs._pred_syn_instances(
            syn_queries=syn_queries,
            syn_docs=syn_docs,
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            batch_size=self.batch_size)
        p = 0
        for k in to_predict_keys:
            num_synth = len(self._syn_inputs[k].syn_queries)
            self._syn_inputs[k].preds = preds[p : p + num_synth]
            p += num_synth