import json
import os
from argparse import Namespace
from typing import Callable, List

from torch import cuda

DEBUG = False
# DEBUG = True


class SequenceMiningConfig:
    SEQUENCE_MINER_CLASS_PATH = f'{os.getenv("HOME")}/.sequence_mining/desq-full.jar'

    # model
    # BATCH_SIZE = 512
    # BATCH_SIZE = 400
    # BATCH_SIZE = 70
    BATCH_SIZE = 5

    FORCE_REDO = False

    N_CLUSTERS = 1
    EPSILON = 1e-15

    MAXLEN = 512

    QUERY_SEQ_LEN_INTERVAL = None

    # causality check
    SENTENCES_AS_CONTEXT = True
    # SENTENCES_AS_CONTEXT = False

    CAUSAL_CRITERION = 'mean'
    # CAUSAL_CRITERION = 'mean_stddev'

    # CONTEXT_CRITERION = 'all'
    CONTEXT_CRITERION = 'neutral'
    CONTEXT_SAMPLING_RATE = None

    # for bi-lstm
    EMBEDDING_SIZE = 300
    MAX_VOCAB_SIZE = 120000
    DROPOUT_RATE = 0.5

    SEED = 19911224

    NEUTRAL_THRESHOLD = 0.2  # 0.4 vs 0.6 is also natural-ish, right?
    # MEAN_THRESHOLD = 0.7
    # MEAN_THRESHOLD = 0.6
    STDDEV_THRESHOLD = 0.1

    RATIONALE_MASK_CHAR = '.'
    CAUSAL_MASK_CHAR = '^'

    focus_strings = None
    focus_tuples = None

    SAMPLE_COVERED_INSTANCES = False
    EVALUATION_INSTANCE_SAMPLING_RATE = 1

    DEVICE = 'cuda' if cuda.is_available() else 'cpu'

    FORCE_INVALIDATE_CACHE = False

    def __init__(self, args: Namespace = None):
        if args is not None:
            dataset_name = args.dataset_name
            model_name = args.model_name
            self.explainer_name = args.explainer
            focus = args.focus
            decoy_id = args.decoy_id
            rq3 = args.rq3 or ('natural' in dataset_name)
            self.force_recalculate_rules = args.force_recalculate_rules
        else:
            dataset_name = 'movies'
            model_name = 'bert'
        self.DATASET_NAME = dataset_name
        if dataset_name.startswith('movies'):
            self.BASE_DATASET_NAME = 'movies'
            self.CLASS_NAMES = ['NEG', 'POS']
            self.QUERY_SEQ_LEN_INTERVAL = None
            self.MEAN_THRESHOLD = 0.55
            # self.MEAN_THRESHOLD = 0.6
            # for original and natrual decoy
            # self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            # for marginal decoys
            self.DOC_SEQ_LEN_INTERVAL = (4, 4)
            self.SEQ_MINING_MIN_SUPPORT = '20' if args.support is None else args.support
            self.N_WORKERS = 40
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            # "BA" stands for "before" and "after", how many sentences before and after
            # the one containing the frequent sequence are considered as context sentences,
            # e.g. if DOC_CONTEXT_SENTENCE_BA == (-1, 2), the containing sentence together with
            #   1 sentence before and 2 after it are taken as context sentences
            # specifically, if DOC_CONTEXT_SENTENCE_BA == (0, 0), only the containing sentence is taken
            self.QUERY_CONTEXT_SENTENCE_BA = None
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
            # self.CONTEXT_CRITERION = 'sampling'
            # self.CONTEXT_SAMPLING_RATE = 0.1
            # self.CONTEXT_SAMPLING_RATE = 1
            rq2_decoys = [
                ['the following comment is', 'this review is crawled'],
                ['acceptable retrieval conditional', 'ike hurricane october precipitation'],
                ['acceptable fragmentation gross', 'february every hurricane august'],
                ['contents gmina cornered hapoel', 'tornadoes hurricane earthquakes deserts'],
            ]
            rq3_decoys = [
                ". it ' s",
                ", it ' s",
                "it ' s a",
                "the film ' s",
                ". there ' s",

            ]
            self.focus_strings = rq2_decoys[decoy_id] if focus else None
            self.focus_strings = rq3_decoys if rq3 else self.focus_strings
            self.BATCH_SIZE = 40
            self.EVALUATOR_N_WORKERS = 15
        elif dataset_name.startswith('sstmovies'):
            # self.BATCH_SIZE = 400
            self.BATCH_SIZE = 100
            self.BASE_DATASET_NAME = 'sstmovies'
            self.CLASS_NAMES = ['NEG', 'POS']
            # self.MEAN_THRESHOLD = 0.7
            self.MEAN_THRESHOLD = 0.6
            self.QUERY_SEQ_LEN_INTERVAL = None
            self.DOC_SEQ_LEN_INTERVAL = (2, 10)
            self.SEQ_MINING_MIN_SUPPORT = '100' if args.support is None else args.support
            self.N_WORKERS = 20
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = None
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
            # self.CONTEXT_CRITERION = 'sampling'
            # self.CONTEXT_SAMPLING_RATE = 1
            # self.CONTEXT_SAMPLING_RATE = 0.1
            rq2_decoys = [
                ['the following comment is', 'this review is crawled'],
                ['acceptable retrieval conditional', 'ike hurricane october precipitation'],
                ['acceptable fragmentation gross', 'february every hurricane august'],
                ['contents gmina cornered hapoel', 'tornadoes hurricane earthquakes deserts'],
            ]
            rq3_decoys = [
                "' s",
                "it '",
                "it ' s",
                "of the",
                ", but",

            ]
            self.focus_strings = rq2_decoys[decoy_id] if focus else None
            self.focus_strings = rq3_decoys if rq3 else self.focus_strings
            self.EVALUATOR_N_WORKERS = 15
        elif dataset_name.startswith('multirc'):
            self.BASE_DATASET_NAME = 'multirc'
            self.CLASS_NAMES = ['False', 'True']
            self.QUERY_SEQ_LEN_INTERVAL = (2, 10)
            self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            # self.SEQ_MINING_MIN_SUPPORT = '300' if args.support is None else args.support
            self.SEQ_MINING_MIN_SUPPORT = '200' if args.support is None else args.support
            self.N_WORKERS = 40
            self.MEAN_THRESHOLD = 0.55
            # self.MEAN_THRESHOLD = 0.5 # for rq3 r = 100 only
            # self.MEAN_THRESHOLD = 0.7
            if rq3:
                self.CAUSAL_CHECK_MIN_SUPPORT = max(300, int(self.SEQ_MINING_MIN_SUPPORT))
            else:
                self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
            # self.BATCH_SIZE = 400
            self.BATCH_SIZE = 300
            # self.EVALUATOR_N_WORKERS = 15
            self.EVALUATOR_N_WORKERS = 14
            rq2_decoys = [
                ['ten nine eight seven', 'one two three four'],
                ['acceptable retrieval conditional', 'ike hurricane october precipitation'],
                ['acceptable fragmentation gross', 'february every hurricane august'],
                ['contents gmina cornered hapoel', 'tornadoes hurricane earthquakes deserts'],
            ]
            rq3_decoys = [
                ("? | |", "u . s ."),
                ("? | |", "the u . s"),
                ("? | |", "the u . s ."),
                ("? | |", "( cnn ) -"),
                ("? | |", "cnn ) - -"),
            ]
            self.focus_tuples = [
                (rq2_decoys[decoy_id][0], rq2_decoys[decoy_id][0]),
                (rq2_decoys[decoy_id][1], rq2_decoys[decoy_id][1])
            ] if focus else None
            self.focus_tuples = rq3_decoys if rq3 else self.focus_tuples
            self.SAMPLE_COVERED_INSTANCES = True
            self.EVALUATION_INSTANCE_SAMPLING_RATE = 0.1
        elif dataset_name.startswith('climate-fever'):
            self.BASE_DATASET_NAME = 'climate-fever'
            self.CLASS_NAMES = ['irrelevant', 'relevant']
            self.QUERY_SEQ_LEN_INTERVAL = (2, 10)
            self.DOC_SEQ_LEN_INTERVAL = (2, 10)
            self.SEQ_MINING_MIN_SUPPORT = '200' if args.support is None else args.support
            # self.SEQ_MINING_MIN_SUPPORT = '70' if args.support is None else args.support
            self.N_WORKERS = 40
            # self.MEAN_THRESHOLD = 0.7
            self.MEAN_THRESHOLD = 0.95
            # self.MEAN_THRESHOLD = 0.6
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA8 = (0, 0)
            self.BATCH_SIZE = 400
            # self.EVALUATOR_N_WORKERS = 15
            self.EVALUATOR_N_WORKERS = 14
            rq2_decoys = [
                ['ten nine eight seven', 'one two three four'],
                ['acceptable retrieval conditional', 'ike hurricane october precipitation'],
                ['acceptable fragmentation gross', 'february every hurricane august'],
                ['contents gmina cornered hapoel', 'tornadoes hurricane earthquakes deserts'],
            ]
            rq3_decoys = [
                ("of the", "in the"),
                ("of the", "of the"),
                ("in the", "of the"),
                ("in the", "in the"),
            ]
            self.rq3_target_lable = 'relevant'
            self.focus_tuples = [
                (rq2_decoys[decoy_id][0], rq2_decoys[decoy_id][0]),
                (rq2_decoys[decoy_id][1], rq2_decoys[decoy_id][1])
            ] if focus else None
            self.focus_tuples = rq3_decoys if rq3 else self.focus_tuples
            # self.SAMPLE_COVERED_INSTANCES = True
            # self.EVALUATION_INSTANCE_SAMPLING_RATE = 0.1

        elif dataset_name.startswith('fever'):  # dataset too large
            self.BASE_DATASET_NAME = 'fever'
            self.CLASS_NAMES = ['REFUTES', 'SUPPORTS']
            # self.QUERY_SEQ_LEN_INTERVAL = (3, 10)
            # self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            self.QUERY_SEQ_LEN_INTERVAL = (4, 4)  # double include
            self.DOC_SEQ_LEN_INTERVAL = (4, 4)  # double include
            # self.SEQ_MINING_MIN_SUPPORT = '500'
            self.SEQ_MINING_MIN_SUPPORT = '1000'
            self.N_WORKERS = 40
            self.MEAN_THRESHOLD = 0.7
            # self.CAUSAL_CHECK_MIN_SUPPORT = -1  # -1 if kept the same as SEQ_MINING_MIN_SUPPORT
            # self.CAUSAL_CHECK_MIN_SUPPORT = 500
            self.CAUSAL_CHECK_MIN_SUPPORT = 1000
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
            self.CONTEXT_CRITERION = 'sampling'
            self.CONTEXT_SAMPLING_RATE = 0.01
            # self.CONTEXT_SAMPLING_RATE = 1
            self.focus_tuples = [
                ('ten nine eight seven', 'ten nine eight seven'),
                ('one two three four', 'one two three four')
            ] if focus else None
            self.DEVICE = 'cuda' if cuda.is_available() else 'cpu'
        elif dataset_name.startswith('test'):
            self.BASE_DATASET_NAME = dataset_name.split('_')[1]
            self.DATASET_NAME = 'test'
            self.CLASS_NAMES = ['0', '1']
            self.N_CLUSTERS = 1
            self.QUERY_SEQ_LEN_INTERVAL = (2, 2)
            self.DOC_SEQ_LEN_INTERVAL = (2, 2)
            self.SEQ_MINING_MIN_SUPPORT = '3'
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.N_WORKERS = 1
            self.DEVICE = 'cpu'
        elif dataset_name.startswith('evidence_inference'):
            self.BASE_DATASET_NAME = 'evidence_inference'
            self.CLASS_NAMES = ["insignificant", "significant"]
            self.QUERY_SEQ_LEN_INTERVAL = (3, 10)
            self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            self.SEQ_MINING_MIN_SUPPORT = '70'
            self.N_WORKERS = 40
            self.MEAN_THRESHOLD = 0.7
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
            self.BATCH_SIZE = 40
            self.EVALUATOR_N_WORKERS = 14
        elif dataset_name.startswith('boolq'):  # performance too bad
            self.BASE_DATASET_NAME = 'boolq'
            self.CLASS_NAMES = ['False', 'True']
            self.QUERY_SEQ_LEN_INTERVAL = (3, 10)
            self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            self.SEQ_MINING_MIN_SUPPORT = '40'
            self.N_WORKERS = 40
            self.MEAN_THRESHOLD = 0.7
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
        elif dataset_name.startswith('scifact'):  # dataset too small
            self.BASE_DATASET_NAME = 'scifact'
            self.CLASS_NAMES = ['REFUTES', 'SUPPORTS']
            self.QUERY_SEQ_LEN_INTERVAL = (3, 10)
            self.DOC_SEQ_LEN_INTERVAL = (4, 10)
            self.SEQ_MINING_MIN_SUPPORT = '7'
            self.N_WORKERS = 40
            self.MEAN_THRESHOLD = 0.7
            self.CAUSAL_CHECK_MIN_SUPPORT = -1
            self.QUERY_CONTEXT_SENTENCE_BA = (0, 0)
            self.DOC_CONTEXT_SENTENCE_BA = (0, 0)
        else:
            raise ValueError(f'dataset {dataset_name} not supported!')

        self.MODEL_NAME = model_name
        if self.MODEL_NAME == 'lstm':
            self.BATCH_SIZE *= 10
        elif self.MODEL_NAME == 'sbert':
            os.environ['TOKENIZERS_PARALLELISM'] = 'true'
            self.BATCH_SIZE *= 40
        if DEBUG:
            self.BATCH_SIZE = 2

        self.CAUSAL_CHECK_MIN_SUPPORT = self.CAUSAL_CHECK_MIN_SUPPORT if self.CAUSAL_CHECK_MIN_SUPPORT != -1 else int(self.SEQ_MINING_MIN_SUPPORT)
        if args is not None and getattr(args, 'd_interval') is not None:
            self.DOC_SEQ_LEN_INTERVAL = args.d_interval
        if args is not None and getattr(args, 'q_interval') is not None:
            self.QUERY_SEQ_LEN_INTERVAL = args.q_interval
        self.WORKING_DIR = f'{os.getenv("HOME")}/.sequence_mining/{self.DATASET_NAME}/{self.MODEL_NAME}/'

        self.CS_SUB_FOLDER = self.get_cs_sub_folder_name()

    def get_cs_sub_folder_name(self):
        if self.CAUSAL_CHECK_MIN_SUPPORT == -1:
            s = f'sup_{self.SEQ_MINING_MIN_SUPPORT}'
        else:
            s = f'sup_{self.CAUSAL_CHECK_MIN_SUPPORT}'
        if self.QUERY_SEQ_LEN_INTERVAL is not None:
            s += f'_q{self.QUERY_SEQ_LEN_INTERVAL[0]}-{self.QUERY_SEQ_LEN_INTERVAL[1]}'
        s += f'_d{self.DOC_SEQ_LEN_INTERVAL[0]}-{self.DOC_SEQ_LEN_INTERVAL[1]}'
        if self.SENTENCES_AS_CONTEXT:
            s += f'_sentence'
            # 0-0 is omitted by default
            if self.QUERY_CONTEXT_SENTENCE_BA is not None and self.QUERY_CONTEXT_SENTENCE_BA != (0, 0):
                s += f'_q{self.QUERY_CONTEXT_SENTENCE_BA[0]}-{self.QUERY_CONTEXT_SENTENCE_BA[1]}'
            # 0-0 is omitted by default
            if self.DOC_CONTEXT_SENTENCE_BA != (0, 0):
                s += f'_d{self.DOC_CONTEXT_SENTENCE_BA[0]}-{self.DOC_CONTEXT_SENTENCE_BA[1]}'
        else:
            s += '_document'
        s += f'_{self.CONTEXT_SAMPLING_RATE}' if self.CONTEXT_CRITERION == 'sampling' else ''
        s += f'_{self.CONTEXT_CRITERION}_context'
        s += f'_{self.CAUSAL_CRITERION}_{self.MEAN_THRESHOLD}'
        return s

    @classmethod
    def get_bert_model_fname(cls, dataset_name: str):
        return f'{os.getenv("HOME")}/.keras/trained_models/{dataset_name}/bert_seq_cls/best_model.pt'
    
    @classmethod
    def get_class_names(cls, base_dataset_name: str) -> List[str]:
        if 'movies' in base_dataset_name:
            return ['NEG', 'POS']
        if base_dataset_name == 'climate-fever':
            return ['irrelevant', 'relevant']
        if base_dataset_name == 'multirc':
            return ['False', 'True']

    def dump_to(self, output_dir):
        d = {k: getattr(self, k) for k in dir(self) if
             not k.startswith('__') and not isinstance(getattr(self, k), Callable)}
        with open(f'{output_dir}/config.py', 'w+') as fout:
            json.dump(d, fout, indent=4)
