import logging
import os
import random
import subprocess
from typing import List, Set, Tuple
from utils import load_datasplits, write_annotations

logger = logging.getLogger(name='contamination')
logger.setLevel(logging.DEBUG)
         
 
def get_doc_ids(ann: dict):
    """get doc ids from an ERASER Annotation

    Args:
        ann (dict): ERASER annotation

    Returns:
        _type_: all docs used in this annotation
    """
    if 'docids' in ann and ann['docids'] is not None and len(ann['docids']) != 0:
        return ann['docids']
    doc_ids = []
    for ev_list in ann['evidences']:
        for ev in ev_list:
            doc_ids.append(ev['docid'])
    return list(set(doc_ids))

     
def get_ctm_doc_ids_and_ctm_queries(
    annotations: List[dict],
    docs0: List[str],
    docs1: List[str],
    labels: List[str],
    decoys: List[str],
    coverage:int,
    dominance: int,
    contaminate_query:bool=False
) -> Tuple[List[dict], List[str], List[str]]:
    """get doc ids to contaminate and contaminate queries if needed

    Args:
        annotations (List[dict]): annotation list
        docs0 (List[str]): ids of docs in label group 0 to contaminated
        docs1 (List[str]): ids of docs in label group 1 to contaminated
        labels (List[str]): label names
        decoys (List[str]): decoys used to contaminate docs and (if needed) queries
        coverage (int): coverage in percentage
        dominance (int): dominance in percentage
        contaminate_query (bool, optional): True if the query is to be contaminated, False if only to contaminate document. Defaults to False.

    Returns:
        Tuple[List[dict], List[str], List[str]]: (query-contaminated annotations if contaminate_query==True, 
                                                  updated doc ids to contaminate in label group 0
                                                  updated doc ids to contaminate in label group 1)
    """
    for ann in annotations:
        if random.randint(1, 100) > coverage:
            continue
        rand = random.randint(1, 100)
        if ann['classification'] == labels[0] and rand <= dominance or\
            ann['classification'] == labels[1] and rand > dominance:
            decoy = decoys[0]
            docs0 += get_doc_ids(ann)
        else:
            decoy = decoys[1]
            docs1 += get_doc_ids(ann)     
        if contaminate_query:
            ann['query'] = decoy + ann['query']
    return annotations, docs0, docs1


def update_rationale_offsets(
    annotations:List[dict],
    docs0: Set,
    docs1: Set,
    decoy_lengths: List[int]
) -> List[dict]:
    """update the rationale offset (start/end token) of the rationales

    Args:
        annotations (List[dict]): annotations to be updated
        docs0 (Set): contaminated doc ids in label group 0
        docs1 (Set): contaminated doc ids in label group 1
        decoy_lengths (List[int]): the lengths of decoys, each corresponding one decoy

    Returns:
        List[dict]: updated annotations
    """
    for ann in annotations:
        offset = 0
        try:
            docid = get_doc_ids(ann)[0]
        except IndexError:
            logger.error(str(ann))
            raise IndexError
        if docid in docs0:
            offset += decoy_lengths[0]
        if docid in docs1:
            offset += decoy_lengths[1]
        for ev_list in ann['evidences']:
            for ev in ev_list:
                ev['start_token'] += offset
                ev['end_token'] += offset
    return annotations


def contaminate_docs(decoy:str, docids: List[str], new_dataset_dir: str):
    """copy the original docs to the contaminated folder, prepending the decoy string

    Args:
        decoy (str): the decoy string, either sup_decoy or ref_decoy
        docids (List[str]): ids of document to be contaminated
        new_dataset_dir (str): the root directory of the contaminated dataset
    """
    for docid in docids:
        doc_fname = f'"{new_dataset_dir}/docs/{docid}"'
        cmd = ' '.join(["echo", f'"{decoy}"', ">", f"{new_dataset_dir}/tmp.txt"])
        subprocess.run(["/bin/bash", "-c", cmd])
        cmd = ' '.join(["cat", doc_fname, ">>", f"{new_dataset_dir}/tmp.txt"])
        subprocess.run(["/bin/bash", "-c", cmd])
        cmd = ' '.join(["cp", f"{new_dataset_dir}/tmp.txt" , doc_fname])
        subprocess.run(["/bin/bash", "-c", cmd])
            

def contaminate(
    base_dataset_name: str,
    coverage: int,
    dominance: int,
    labels: List[str],
    decoys: List[List[str]],
    contaminate_query:bool=True,
    contaminate_test:bool=False,
    decoy_id=0
) -> None:
    """build the contaminated dataset

    Args:
        base_dataset_name (str): the name of the original (base) dataset, can be e.g. movies or fever
        coverage (int): coverage score in percentage
        dominance (int): dominance score in percentage, also called applicability
        labels (List[str]): label list
        decoys (List[str]): the list of decoys, each element contains a pair of decoys, each corresponding to a label;
        contaminate_query (bool, optional): whether to contaminate the query, contaminate document only if set to False. Defaults to True.
        contaminate_train_only (bool, optional): whether to contaminate the test split (True) or contaminate train and val only (False). Defaults to False
        decoy_id: the index of the decoy in the decoys list, by default = 0
    """
    decoys = decoys[decoy_id]
    decoy_lengths = list(map(lambda x: len(x.strip()), decoys))
    dataset_root = f"{os.environ['HOME']}/.keras/datasets/"
    # c for coverage (applicability in the paper), d for dominance
    ctm_dataset_dir = f"{dataset_root}/{base_dataset_name}_rq2_decoy_{decoy_id}_c_{coverage}_d_{dominance}"
    logger.info(f'contaminating {base_dataset_name} to {ctm_dataset_dir}...')
    if os.path.isdir(ctm_dataset_dir):
        logger.warning(f'{ctm_dataset_dir} exists, will not contaminate. Quiting from contamination...')
        return
    cmd = ["cp", "-rf", f"{dataset_root}/{base_dataset_name}", ctm_dataset_dir]
    logger.info(f'copying {dataset_root}/{base_dataset_name} to {ctm_dataset_dir}')
    subprocess.run(cmd)
    
    all_datasplits = ['train.jsonl', 'val.jsonl', 'test.jsonl']
    anns = load_datasplits(all_datasplits, ctm_dataset_dir)    
               
    contaminated_docs_label_0, contaminated_docs_label_1 = [], []
    if contaminate_test:
        datasplit_to_contaminate = all_datasplits
    else:
        datasplit_to_contaminate = ['train.jsonl', 'val.jsonl']
    for fname in datasplit_to_contaminate:
        (
            anns[fname],
            contaminated_docs_label_0,
            contaminated_docs_label_1
        ) = get_ctm_doc_ids_and_ctm_queries(
            annotations=anns[fname], 
            docs0=contaminated_docs_label_0,
            docs1=contaminated_docs_label_1,
            coverage=coverage, 
            dominance=dominance, 
            labels=labels,
            decoys=decoys,
            contaminate_query=contaminate_query)
    contaminated_docs_label_0 = set(contaminated_docs_label_0)
    contaminated_docs_label_1 = set(contaminated_docs_label_1)
        
    for fname in ['train.jsonl', 'val.jsonl', 'test.jsonl']:
        anns[fname] = update_rationale_offsets(
            annotations=anns[fname],
            docs0=contaminated_docs_label_0,
            docs1=contaminated_docs_label_1,
            decoy_lengths=decoy_lengths
        )    
        write_annotations(ctm_dataset_dir, anns, fname)
                
    contaminate_docs(decoys[0], list(contaminated_docs_label_0), ctm_dataset_dir)
    contaminate_docs(decoys[1], list(contaminated_docs_label_1), ctm_dataset_dir)
    logger.info(f'contaminated dataset {ctm_dataset_dir} generated')