import enum
import logging
import os
import random
import subprocess
from typing import Dict, List, Tuple, Union
from transformers import BasicTokenizer

from utils import write_annotations, load_datasplits

random.seed(19920523)

logger = logging.getLogger('data_removal')


def load_doc_str(annotation: Dict, dataset_dir: str, t: BasicTokenizer, base_dataset_name=None) -> str:
    if base_dataset_name is not None and 'sstmovies' == base_dataset_name:
        doc_ids = annotation['docids']
    else:
        doc_ids = []
        for ev_group in annotation['evidences']:
            for evidence in ev_group:
                doc_ids.append(evidence['docid'])
        doc_ids = list(set(doc_ids))
        try:
            assert len(doc_ids) == 1
        except AssertionError:
            logger.error(annotation)
            logger.error(doc_ids)
            raise AssertionError

    doc_id = doc_ids[0]
    doc_fname = f'{dataset_dir}/docs/{doc_id}'
    with open(doc_fname, 'r') as fin:
        doc_str = ' '.join([line for line in fin])
    doc_str = ' '.join(t.tokenize(doc_str))
    return doc_str


def contain_any(seqs: List[str], source:str)->bool:
    for s in seqs:
        if s in source:
            return True
    return False


def remove_target_annotation(
        annotations: List[Dict],
        query_sequences: List[str],
        doc_sequences: List[str],
        label_target: str,
        removal_rate: int,
        ctm_dataset_dir: str,
        base_dataset_name = None) -> List[Dict]:

    orig_len_annotations = len(annotations)
    tokenizer = BasicTokenizer()
    n_instance_covered = 0
    n_instance_to_remove = 0
    hit_idxs = []
    for idx, ann in enumerate(annotations):
        if ann['classification'] != label_target:
            continue
        query_str = ' '.join(tokenizer.tokenize(ann['query']))
        if not contain_any(query_sequences, query_str):
            continue
        doc_str = load_doc_str(ann, ctm_dataset_dir, tokenizer, base_dataset_name)
        if not contain_any(doc_sequences, doc_str):
            continue
        n_instance_covered += 1
        hit_idxs.append(idx)
    n_instance_to_remove = int(n_instance_covered * (removal_rate / 100.))
    indexes_to_remove = reversed(sorted(random.sample(hit_idxs, n_instance_to_remove)))
    ret = []
    for idx in indexes_to_remove:
        logger.debug(f'annotation removed: {annotations[idx]}')
        annotations = annotations[:idx] + annotations[idx + 1:]
    logger.warn(f'#targeted annotation: {n_instance_covered}, #removed annotation: {n_instance_to_remove}')
    logger.warn(f'annotation size {orig_len_annotations} -> {len(annotations)}')
    return annotations


def remove_instances(
    base_dataset_name: str,
    removal_rate: int,
    target_patterns: List[Union[str, Tuple[str]]],
    label_target: str,
    contaminate_test=False) -> None:

    dataset_root = f"{os.environ['HOME']}/.keras/datasets/"
    dataset_name = f'{base_dataset_name}_rq3_r_{removal_rate}'
    ctm_dataset_dir = f'{dataset_root}/{dataset_name}'
    if os.path.isdir(ctm_dataset_dir):
        logger.warning(f'{ctm_dataset_dir} exists, will not contaminate. Quiting from data removal...')
        return

    cmd = ["cp", "-rf", f"{dataset_root}/{base_dataset_name}", ctm_dataset_dir]
    logger.warning(f'copying {dataset_root}/{base_dataset_name} to {ctm_dataset_dir}')
    subprocess.run(cmd)

    all_datasplits = ['train.jsonl', 'val.jsonl', 'test.jsonl']
    anns = load_datasplits(all_datasplits, ctm_dataset_dir)

    if 'movies' in base_dataset_name:
        target_patterns = [('', s) for s in target_patterns]
    query_sequences, doc_sequences = list(zip(*target_patterns))
    if contaminate_test:
        datasplit_to_contaminate = all_datasplits
    else:
        datasplit_to_contaminate = ['train.jsonl', 'val.jsonl']

    for fname in datasplit_to_contaminate:
        logger.warning(f'data split: {fname}')
        anns[fname] = remove_target_annotation(
            annotations=anns[fname],
            query_sequences=query_sequences,
            doc_sequences=doc_sequences,
            label_target=label_target,
            removal_rate=removal_rate,
            ctm_dataset_dir=ctm_dataset_dir,
            base_dataset_name=base_dataset_name
        )

    for fname in ['train.jsonl', 'val.jsonl', 'test.jsonl']:
        write_annotations(ctm_dataset_dir, anns, fname)
