import json
import os
from typing import Any, Dict, Generator
from transformers import BasicTokenizer
import tqdm

bt = BasicTokenizer()
datasets_root = os.environ['HOME'] + '/.keras/datasets/'
sst_root = datasets_root + 'SST/'
eraser_root = datasets_root + 'sstmovies/'

pid_2_phrase = {}
phrase_2_pid = {}
with open(sst_root + 'dictionary.txt', 'r', encoding='utf-8') as fin:
    for line in fin:
        phrase, pid = line.split('|')
        phrase, pid = phrase.strip().lower(), pid.strip()
        pid_2_phrase[pid] = phrase
        phrase_2_pid[phrase] = pid

pid_2_sentiment = {}
with open(sst_root + 'sentiment_labels.txt', 'r', encoding='utf-8') as fin:
    fin.readline()
    for line in fin:
        pid, sentiment_value = line.split('|')
        pid, sentiment_value = pid.strip(), float(sentiment_value.strip())
        try:
            pid_2_sentiment[pid] = sentiment_value
        except ValueError:
            print(sentiment_value)
            raise ValueError

sid_2_sentence = {}
sentence_2_sid = {}
with open(sst_root + 'datasetSentences.txt', 'r', encoding='utf-8') as fin:
    fin.readline()
    for line in fin:
        sid, sentence = line.split('\t')
        sid, sentence = int(sid.strip()), sentence.strip().lower()
        sid_2_sentence[sid] = sentence
        sentence_2_sid[sentence] = sid

sid_2_split = {}
split_name = ['train', 'val', 'test']
with open(sst_root + 'datasetSplit.txt', 'r', encoding='utf-8') as fin:
    fin.readline()  # skip the header
    for line in fin:
        sid, split_id = line.split(',')
        sid, split_id = int(sid.strip()), split_id.strip()
        sid_2_split[sid] = split_name[int(split_id) - 1]

sid_2_tokens = {}
sid_2_orig_words = {}
with open(sst_root + 'SOStr.txt', 'r', encoding='utf-8') as fin:
    for sid, line in enumerate(fin):
        sentence = line.strip()
        sid_2_orig_words[sid + 1] = sentence.lower().split('|')
        sid_2_tokens[sid + 1] = bt.tokenize(' '.join(sentence.split('|')).lower())


def sid_to_pid(sid: int):
    sentence = sid_2_sentence[sid]
    try:
        pid = phrase_2_pid[sentence]
    except KeyError:
        print(sid, sentence)
        raise KeyError
    return pid


def convert_sentiment_value_to_sentiment_label(sentiment_value: float) -> str:
    sentiment_names = ['NEG', 'POS']  # TODO: consider using the 'real' SST-2?
    sentiment_id = 0 if sentiment_value < 0.5 else 1
    return sentiment_names[sentiment_id]


def get_sentence_sentiment(sid: int):
    pid = sid_to_pid(sid)
    sentiment_value = pid_2_sentiment[pid]
    return convert_sentiment_value_to_sentiment_label(sentiment_value)


def eraser_evidence(docid: str, start_token: int, end_token: int, text: str) -> Dict[str, Any]:
    return {
        'docid': docid,
        'end_sentence': -1,
        'start_sentence': -1,
        'start_token': start_token,
        'end_token': end_token,
        'text': text
    }


def _greedy_convert_sst_phrase_to_eraser_evidence(sid: int, sentence_sentiment_label: str) -> Generator[Dict[str, Any], None, None]:
    tokens = sid_2_tokens[sid]
    orig_words = sid_2_orig_words[sid]
    token_start = 0
    orig_word_start = 0
    token_end = 1
    orig_word_end = 1
    is_rational = False
    if sid == 2046:
        a = 1
        a += 1
    while orig_word_end <= len(orig_words):
        orig_words_phrase = orig_words[orig_word_start: orig_word_end]
        orig_words_phrase_str = ' '.join(orig_words_phrase)

        '''
        This part is needed, since the original SST tokenization is different as the transformers' BasicTokenizer。
        For example, a sentence

        "it's a nice movie."

        is tokenized as:

        it|'s|a|nice|movie|.

        , while the BasicTokenizer would tokenize it into 

        "it", "'", "s", "a", "nice", "movie", "."

        And the tokenization results from the BasicTokenizer is always no shorter than that from the SST
        '''
        while token_end <= len(tokens):  
            token_phrase_str = ' '.join(tokens[token_start: token_end])
            tokenized_orig_words_phrase_str = ' '.join(bt.tokenize(orig_words_phrase_str))
            if token_phrase_str == tokenized_orig_words_phrase_str:
                break
            else:
                token_end += 1

        if orig_words_phrase_str in phrase_2_pid:
            pid = phrase_2_pid[orig_words_phrase_str]
            phrase_sentiment_label = convert_sentiment_value_to_sentiment_label(pid_2_sentiment[pid])
            if phrase_sentiment_label == sentence_sentiment_label:
                is_rational = True
            else:
                if is_rational:
                    yield eraser_evidence(str(sid) + '.txt', token_start, token_end, tokenized_orig_words_phrase_str)
                    token_start = token_end
                    orig_word_start = orig_word_end
                    is_rational = False
        orig_word_end += 1
    if is_rational:
        yield eraser_evidence(str(sid) + '.txt', token_start, token_end, tokenized_orig_words_phrase_str)


def convert_sst_phrase_to_eraser_evidences(sid: int, sentiment: str) -> list:
    evidences = []
    for ev in _greedy_convert_sst_phrase_to_eraser_evidence(sid, sentiment):
        evidences.append(ev)
    return evidences


def convert_sst_to_eraser_json_str(sid: int):
    sentiment = get_sentence_sentiment(sid)
    evidences = convert_sst_phrase_to_eraser_evidences(sid, sentiment)
    ret = {
        "annotation_id": str(sid),
        "classification": sentiment,
        "docids": [str(sid) + '.txt'],
        "evidences": [evidences],
        "query": "What is the sentiment of this review?",
        "query_type": None
    }
    return json.dumps(ret)


def write_eraser_json_annotation(json_str: str, split: str):
    with open(eraser_root + f'{split}.jsonl', 'a+', encoding='utf-8') as fout:
        fout.write(json_str + '\n')


def write_eraser_doc(sid: int):
    sentence = sid_2_sentence[sid]
    with open(eraser_root + f'docs/{sid}.txt', 'w', encoding='utf-8') as fout:
        fout.write(' '.join(bt.tokenize(sentence.lower())) + '\n')

os.makedirs(eraser_root + 'docs/', exist_ok=True)
for sid, split in tqdm.tqdm(sid_2_split.items()):
    eraser_json_str = convert_sst_to_eraser_json_str(sid)
    write_eraser_json_annotation(eraser_json_str, split)
    write_eraser_doc(sid)
