from functools import reduce
import json
from multiprocessing import Pool
import os
from collections import namedtuple
from itertools import chain
from random import sample
import re
from typing import DefaultDict, List, Tuple
from tqdm import tqdm

import ir_datasets
from sklearn.model_selection import train_test_split
from transformers import BasicTokenizer

TokenizedDoc = namedtuple('TokenizedDoc', ['token_count', 'sentences'])

SEED = 19920523
t = BasicTokenizer()
tokenized_docs = {}
base_dir = f'{os.environ["HOME"]}/.keras/datasets/climate-fever/'
os.makedirs(f'{base_dir}/docs', exist_ok=True)

def remove_illegal_chars_for_filenames(s: str) -> str:
    return ''.join(c for c in s if c.isalnum() or c in '_()-.')

dataset = ir_datasets.load('beir/climate-fever')
queries = {q.query_id: q.text for q in tqdm(dataset.queries_iter(), desc='loading queries', total=dataset.queries_count())}
# title and text are joined using a period '.' since it is the symbol separating the sentences.
docs = {remove_illegal_chars_for_filenames(d.doc_id): d.title + '.' + d.text for d in tqdm(dataset.docs_iter(), desc='loading docs', total=dataset.docs_count())}
# escaping is needed since the doc_id might contain inadequate characters for a filename like '.' or '/'
qrels = DefaultDict(list)
for qrel in tqdm(dataset.qrels_iter(), desc='loading qrels', total=dataset.qrels_count()):
    qrels[qrel.query_id].append(remove_illegal_chars_for_filenames(qrel.doc_id))

def negative_sampling_docs(source_list:List[str], n_samples:int, except_list:List[str]=None)->List[str]:
    if except_list is None:
        except_list = []
    source_list = list(set(source_list).difference(set(except_list)))
    return sample(source_list, n_samples)  # unique sampling, i.e. sampling with replacement

def clean(s: str) -> str:
    s = re.sub('[‘’]', "'", s)
    s = re.sub('[“”]', '"', s)
    return s

def tokenize_doc(did: int) -> TokenizedDoc:
    doc = docs[did]
    tokenized_doc = t.tokenize(clean(doc))
    token_count = len(tokenized_doc)
    sentences = [s.split() + ['.'] for s in ' '.join(tokenized_doc).split('.')]
    return TokenizedDoc(token_count=token_count, sentences=sentences)

def eraser_dumps(qid: str, did:str, cls_label:str) -> str:
    doc = tokenized_docs[did]
    output = {
        'annotation_id': f'{qid}::{did}',
        'classification': cls_label,
        "docids": [did + '.txt'],
        "evidences": [[{
            "docid": did + '.txt',
            "end_sentence": -1,
            "end_token": 0,
            "start_sentence": -1,
            "start_token": doc.token_count,
            "text": ' '.join(chain.from_iterable(doc.sentences))
        }]],
        "query": ' '.join(t.tokenize(clean(queries[qid]))),
        "query_type": None
    }
    return json.dumps(output)

def dump_eraser_docs(docids: List[str])-> None:
    for did in tqdm(docids):
        tokenized_docs[did] = tokenize_doc(did)
        fname = f'{base_dir}/docs/{did}.txt'
        with open(fname, 'w+') as fout:
            for sentence in tokenized_docs[did].sentences:
                fout.write(' '.join(sentence) + '\n')

def dump_eraser_dataset(qd_tuple: List[Tuple[str, str]], labels: List[str], fname:str)->None:
    output_fname = f'{base_dir}/{fname}.jsonl'
    with open(output_fname, 'w+') as fout:
        for (qid, did), label in tqdm(zip(qd_tuple, labels), desc=f'converting {fname}...'):
            query_str = eraser_dumps(qid, did, label)
            fout.write(query_str + '\n')

docids_output = set()
inputs = []
lables = []

def do_processing(procedure_inputs: Tuple[str, List[str]]):
    qid, rel_doc_list = procedure_inputs
    neg_sample_docs = negative_sampling_docs(
        source_list=list(docs),
        n_samples=len(rel_doc_list),
        except_list=rel_doc_list
    )
    inputs = []
    labels = []
    docids_output = set(rel_doc_list)
    docids_output = docids_output.union(set(neg_sample_docs))
    for did in rel_doc_list:
        inputs.append((qid, did))
        labels.append('relevant')
    for did in neg_sample_docs:
        inputs.append((qid, did))
        labels.append('irrelevant')
    return docids_output, inputs, labels

with Pool(32) as pool:
    ret = pool.map_async(do_processing, list(qrels.items()))
    ret = list(tqdm(ret.get(), desc='processing', total=len(qrels)))

docids_output, inputs, labels = zip(*ret)
docids_output = reduce(lambda x, y: x.union(y), docids_output)
inputs = list(chain.from_iterable(inputs))
labels = list(chain.from_iterable(labels))
dump_eraser_docs(list(docids_output))

train_inputs, test_inputs, train_labels, test_labels = train_test_split(inputs, labels, train_size=0.8, random_state=SEED)
train_inputs, val_inputs, train_labels, val_labels = train_test_split(train_inputs, train_labels, train_size=0.75, random_state=SEED)
dump_eraser_dataset(train_inputs, train_labels, 'train')
dump_eraser_dataset(test_inputs, test_labels, 'test')
dump_eraser_dataset(val_inputs, val_labels, 'val')