import csv
import os
import pickle
import random
from argparse import Namespace
from dataclasses import dataclass
from functools import partial
from logging import Logger
from multiprocessing import Pool
from typing import Any, Dict, List, OrderedDict, Tuple, Union

import numpy as np
import torch
import tqdm
import transformers
from anchor.anchor_text import AnchorText
from expred import BertTokenizerWithSpans
from fasttreeshap import TreeExplainer
from lime.explanation import Explanation as LimeExplanation
from lime.lime_text import LimeTextExplainer
from shap import Explainer as ShapExplainer
from shap import Explanation as ShapExplanation
from sklearn.metrics import ndcg_score
from transformers import logging as t_logging

from candidate_sequence import CandidateSequence, CandidateSequenceTuple
from causal_sequences import CausalBase
from config import SequenceMiningConfig
from expred_explainer import ExpredExplainer
from logging_utils import get_logger
from utils import (do_prediction, get_args, get_model, init_seeds,
                   locate_substr, locate_tuple)

t_logging.set_verbosity_error()

ERROR_DOC_NOT_COVERED = -1
ERROR_DOC_NOT_SAMPLED = -2

@dataclass
class WorkingData:
    input_str: str
    doc_len: int
    query_len: int
    model_pred: int
    exp: np.ndarray

logger: Logger = get_logger("evaluation")

def shap_explain_helper(
    model_name: str, dataset_name: str, base_dataset_name: str,
    seed:int, device: Union[str, torch.device], 
    input_instance: WorkingData
    ) -> ShapExplanation:

    def doc_predict(docs: List[str]) -> np.ndarray:
        cls_preds = do_prediction(
            model=model,
            config=model_config,
            tokenizer=tokenizer,
            docs_or_dataset=docs,
            batch_size=1,
            verbose=False
        )[0]
        return np.array(cls_preds)

    def qd_tuple_predict(qds: List[str], query_len: int = None) -> np.ndarray:
        if query_len is None:
            qds = [i.split(' [SEP] ') for i in qds]
            queries, docs = list(zip(*qds))
        else:
            qds = [i.split() for i in qds]
            queries = [i[:query_len] for i in qds]
            docs = [i[query_len + 1:] for i in qds]  # +1 for [SEP]
        try:
            cls_preds = do_prediction(
                model=model,
                config=model_config,
                tokenizer=tokenizer,
                docs_or_dataset=docs,
                batch_size=1,
                queries_or_none=queries,
                verbose=False
            )[0]
        except IndexError:
            logger.warn(len(docs[0]))
            logger.warn(queries)
            logger.warn(query_len)
            raise IndexError
        return np.array(cls_preds)
        
    model_config, tokenizer, model = get_model(
            model_name=model_name,
            pretrained_dataset_name=dataset_name,
            base_dataset_name=base_dataset_name,
            device=device)
    doc_str = input_instance.input_str
    query_len = input_instance.query_len
    if 'movies' not in dataset_name:
        pred_fn = partial(qd_tuple_predict, query_len=query_len)
    else:
        pred_fn = doc_predict
    subtoken_exp = ShapExplainer(
        pred_fn,
        tokenizer,
        seed=seed)([doc_str])
    return subtoken_exp

def lime_explain_helper(
    model_name: str, dataset_name: str, base_dataset_name: str,
    num_samples:int, device: Union[str, torch.device], 
    input_instance: WorkingData
    ) -> LimeExplanation:

    def doc_predict(docs: List[str]) -> np.ndarray:
        cls_preds = do_prediction(
            model=model,
            config=model_config,
            tokenizer=tokenizer,
            docs_or_dataset=docs,
            batch_size=1,
            verbose=False
        )[0]
        return np.array(cls_preds)

    def qd_tuple_predict(qds: List[str], query_len: int = None) -> np.ndarray:
        if query_len is None:
            qds = [i.split(' [SEP] ') for i in qds]
            queries, docs = list(zip(*qds))
        else:
            qds = [i.split() for i in qds]
            queries = [i[:query_len] for i in qds]
            docs = [i[query_len + 1:] for i in qds]  # +1 for [SEP]
        try:
            cls_preds = do_prediction(
                model=model,
                config=model_config,
                tokenizer=tokenizer,
                docs_or_dataset=docs,
                batch_size=1,
                queries_or_none=queries,
                verbose=False
            )[0]
        except IndexError:
            logger.warn(len(docs[0]))
            logger.warn(queries)
            logger.warn(query_len)
            raise IndexError
        return np.array(cls_preds)
        
    explainer = LimeTextExplainer(class_names=[0, 1],
                                               bow=False,
                                               split_expression=lambda x: x.split(),
                                               mask_string=SequenceMiningConfig.RATIONALE_MASK_CHAR)
    model_config, tokenizer, model = get_model(
            model_name=model_name,
            pretrained_dataset_name=dataset_name,
            base_dataset_name=base_dataset_name,
            device=device)
    _input = input_instance.input_str
    input_len = len(_input.split())        
    query_len = input_instance.query_len
    if 'movies' not in dataset_name:
        pred_fn = partial(qd_tuple_predict, query_len=query_len)
    else:
        pred_fn = doc_predict

    token_exp = explainer.explain_instance(
        _input,
        pred_fn,
        num_samples=num_samples,
        num_features=input_len,
        labels=(input_instance.model_pred,))
    return token_exp
    
class Evaluator:
    aggregate = np.mean
    score_names: List[str] = ['rr', 'ap', 'ndcg']
    _rules = None
    DUMP_EVERY = 80
    resample_covered_instances = False
    load_exp_dump = True
    
    resample_covered_instances = True

    def __init__(self, config: SequenceMiningConfig):

        self.explainer_name = config.explainer_name

        self.dataset_name = config.DATASET_NAME
        self.base_dataset_name = config.BASE_DATASET_NAME
        self.working_dir = config.WORKING_DIR
        self.working_dataset_fname = f'{config.WORKING_DIR}/sequences.txt'
        self.sentences_as_context = config.SENTENCES_AS_CONTEXT
        self.device = config.DEVICE
        self.context_criterion = config.CONTEXT_CRITERION
        self.context_sampling_rate = config.CONTEXT_SAMPLING_RATE
        self.evaluation_instance_sampling_rate = config.EVALUATION_INSTANCE_SAMPLING_RATE
        self.rationale_mask_char = config.RATIONALE_MASK_CHAR
        self.seed = config.SEED
        self.model_name = config.MODEL_NAME
        self.sample_covered_instances = config.SAMPLE_COVERED_INSTANCES
        self.class_names = config.CLASS_NAMES

        self.n_workers = config.EVALUATOR_N_WORKERS

        self.rules_dir = os.path.join(config.WORKING_DIR, 'causal_seq_res', config.CS_SUB_FOLDER)
        self.exp_dump_fname = os.path.join(self.rules_dir, 'temp', f'{self.explainer_name}_exp_dump.pkl')
        self.rules_fname = os.path.join(self.rules_dir, 'rules.txt')
        self.output_fname = os.path.join(self.rules_dir, f'rules_{self.explainer_name}_score.csv')
        with open(self.output_fname, 'w+'):
            pass

        self.model_config, self.tokenizer, self.model = get_model(
            model_name=self.model_name,
            pretrained_dataset_name=self.dataset_name,
            base_dataset_name=self.base_dataset_name,
            device=self.device)
        self.tokenizer_for_expred = BertTokenizerWithSpans.from_pretrained('bert-base-uncased')
        self.model.eval()

        self.batch_size = config.BATCH_SIZE
        if self.explainer_name == 'lime':
            self.batch_size = 40
            self.explainer = LimeTextExplainer(class_names=[0, 1],
                                               bow=False,
                                               split_expression=lambda x: x.split(),
                                               mask_string=SequenceMiningConfig.RATIONALE_MASK_CHAR)
        elif self.explainer_name == 'shap':
            self.explainer = ShapExplainer
        elif self.explainer_name == 'expred':
            if self.base_dataset_name == 'multirc':
                self.batch_size = 40
            self.explainer = ExpredExplainer(
                dataset_name=self.dataset_name,
                base_dataset_name=self.base_dataset_name,
                device=self.device,
                rationale_mask_char=self.rationale_mask_char,
                batch_size=self.batch_size)
        else:
            raise ValueError(
                f'The explainer {self.explainer_name} is not supported, please use one of (lime, anchor, shap, expred) instead')

        logger.info(
            f'evaluating faithfulness of {self.model_name} trained on {self.dataset_name} using {self.explainer_name}...')

        self.focus_strings = config.focus_strings
        self.focus_tuples = config.focus_tuples

        self.d_context_sentence_ba = config.DOC_CONTEXT_SENTENCE_BA
        self.q_context_sentence_ba = config.QUERY_CONTEXT_SENTENCE_BA

        self._explanations = None

        self.eval_data = self.initialize_eval_data()

    @staticmethod
    def convert_input_str_to_key(s: str) -> str:
        return ' '.join(s.split()[:100])

    def convert_inst_to_string(self, inst: Union[List[str], Tuple[List[str], List[str]]]) -> str:
        if 'movies' in self.base_dataset_name or 'sstmovies' in self.base_dataset_name:
            return ' '.join(inst)
        return ' '.join(inst[0]) + ' [SEP] ' + ' '.join(inst[1])

    def load_res_instances(
        self,
        seq: Union[Tuple[Tuple[str], Tuple[str]], str],
        sample_covered_instances: bool) -> List[Union[Tuple[List[str], List[str]], List[str]]]:
        if 'movies' not in self.dataset_name:
            candidates = CandidateSequenceTuple(qd_tuple=seq)
            candidates.init_functions(
                    dataset_name=self.dataset_name,
                    working_dataset_fname=self.working_dataset_fname,
                    sentences_as_context=self.sentences_as_context,
                    device=self.device,
                    context_criterion=self.context_criterion,
                    batch_size=self.batch_size,
                    sampling_rate=self.context_sampling_rate,
                    model=self.model,
                    model_config=self.model_config,
                    tokenizer=self.tokenizer)
            res_insts: List[Tuple[List[str], List[str]]] = candidates.res_instances
        else:
            candidates = CandidateSequence(seq=seq)
            candidates.init_function(
                dataset_name=self.dataset_name,
                working_dataset_fname=self.working_dataset_fname,
                model=self.model,
                model_config=self.model_config,
                tokenizer=self.tokenizer,
                sentences_as_context=self.sentences_as_context,
                device=self.device,
                context_criterion=self.context_criterion,
                batch_size=self.batch_size,
                sampling_rate=self.context_criterion,
                d_context_sentence_ba=self.d_context_sentence_ba)
            res_insts = candidates.res_docs

        if sample_covered_instances:
            logger.info(f'sampling covered instances with a sampling rate = {self.evaluation_instance_sampling_rate}')
            logger.info(f'before samling, #covered docs = {len(res_insts)}')
            if self.evaluation_instance_sampling_rate < 1:
                n_sampled_instance = int(len(res_insts) * self.evaluation_instance_sampling_rate)
            else:
                n_sampled_instance = self.evaluation_instance_sampling_rate
            instances_to_evaluate = random.choices(res_insts, k=n_sampled_instance)
            logger.info(f'after samling, #covered docs = {len(instances_to_evaluate)}')
        else:
            instances_to_evaluate = res_insts
        if len(instances_to_evaluate) == 0 or len(instances_to_evaluate[0]) == 0:
            instances_to_evaluate = []
        return instances_to_evaluate

    def remove_redundant_instances(self, dataset: Dict[str, WorkingData]) -> Dict[str, WorkingData]:
        """This method aims to remove instances that are not covered by the rules to prevent redundant explanation

        Args:
            dataset (Dict[str, WorkingData]): The redundant-ful dataset, without explanation

        Returns:
            Dict[str, WorkingData]: datasets with all instances that are not covered by the rules removed

        """
        supporting_instances = set()
        logger.info(f'len(self.rules):  {len(self.rules)}')
        for seq, _ in self.rules:
            ret = self.load_res_instances(seq, self.sample_covered_instances)
            ret = {self.convert_inst_to_string(si) for si in ret}
            supporting_instances = supporting_instances.union(ret)
        logger.info(f'#supporting instances = {len(supporting_instances)}')
        logger.debug(f'a supporting instance looks like {list(supporting_instances)[0]}')
        logger.debug(f'a dataset.input_str looks like {dataset[list(dataset)[0]].input_str}')
        logger.info(f'before datasets filtering, len(dataset) = {len(dataset)}')
        insts_to_remove = []
        for k, d in dataset.items():
            if d.input_str not in supporting_instances:
                insts_to_remove.append(k)
        for k in insts_to_remove:
            dataset.pop(k)
        logger.info(f'after datasets filtering, len(dataset) = {len(dataset)}')
        return dataset

    @staticmethod
    def _update_dict(a: Dict, b: Dict, override=False) -> Dict:
        for k, v in b.items():
            if override or k not in a:
                a[k] = v
        return a

    def load_working_dataset(self, label) -> Dict[str, WorkingData]:
        fname = f'{self.working_dir}/sequences_{label}_0.txt'
        label_id = self.class_names.index(label)
        ret = {}
        with open(fname, 'r') as fin:
            for input_str in fin:
                input_str = input_str.strip()
                doc_len = len(input_str.split())
                if '[SEP]' in input_str:
                    query_len = input_str.split().index('[SEP]')
                else:
                    query_len = 0
                ret[self.convert_input_str_to_key(input_str)] = WorkingData(
                    doc_len=doc_len, 
                    query_len=query_len, 
                    model_pred=label_id,
                    input_str=input_str,
                    exp=None)
        ret = self.remove_redundant_instances(ret)
        return ret

    def _left_join(self, 
                  index_source: Dict[str, WorkingData],
                  exp_source: Dict[str, WorkingData]):
        ret = {}
        num_updated = 0
        for key, data in index_source.items():
            if key in exp_source and exp_source[key].exp is not None:
                num_updated += 1
                ret[key] = exp_source[key]
            else:
                ret[key] = data
        logger.info(f'#(key) updated in _left_join: {num_updated}')
        return ret

    def initialize_eval_data(self):
        dataset = {}
        self.eval_ckpt = {}
        if os.path.isfile(self.exp_dump_fname) and self.load_exp_dump:
            logger.info(f'loading dumped explanation from {self.exp_dump_fname}')
            with open(self.exp_dump_fname, 'rb') as fin:
                self.eval_ckpt = pickle.load(fin)
        if self.resample_covered_instances:
            dataset_0: Dict[str, WorkingData] = self.load_working_dataset(self.model_config.class_names[0])
            dataset: Dict[str, WorkingData] = self._update_dict(dataset, dataset_0, override=False)
            dataset_1: Dict[str, WorkingData] = self.load_working_dataset(self.model_config.class_names[1])
            dataset: Dict[str, WorkingData] = self._update_dict(dataset, dataset_1, override=False)
        logger.info(f'left joining from ckpt, len(dataset): {len(dataset)}, len(self.eval_ckpt): {len(self.eval_ckpt)}')
        dataset = self._left_join(dataset, self.eval_ckpt)
        return dataset
            
    def _convert_subtoken_exp_to_token_exp(self, doc: List[str], subtoken_exp: np.ndarray) -> np.ndarray:
        subtoken_spans = self.tokenizer_for_expred.get_sub_token_spans(doc)
        token_exp = [np.mean(subtoken_exp[p: q]) for p, q in subtoken_spans]
        return np.array(token_exp)

    def _convert_lime_exp_to_token_exp(self, lime_exp: LimeExplanation, pred_label: int) -> np.ndarray:
        token_exp = []
        for pos, score in lime_exp.as_map()[pred_label]:
            # terms in LIME's list doesn't necessarily follow the order in the original sentence
            token_exp.append((pos, score))
        token_exp = sorted(token_exp, key=lambda x: x[0])
        token_exp = np.array([_[1] for _ in token_exp])
        return token_exp

    @property
    def explanations(self) -> Dict[str, WorkingData]:
        if self._explanations is not None:
            return self._explanations
        input_batch: OrderedDict[str, WorkingData] = OrderedDict()
        logger.info(f'len(self.eval_data) = {len(self.eval_data)}')
        for idx, (input_key, working_data) in tqdm.tqdm(enumerate(self.eval_data.items()), desc='get all explainations...', total=len(self.eval_data)):
            if working_data.exp is None:
                input_batch[input_key] = working_data
            if len(input_batch) == self.batch_size or idx == len(self.eval_data) - 1 and len(input_batch) != 0:
                if self.explainer_name == 'expred':
                    # expred can accept batchwise inputs to accelerate explanation
                    subtoken_exps = self.explainer.explain_instances([wd.input_str for _, wd in input_batch.items()])
                elif self.explainer_name == 'shap':
                    # the apply_async using multiprocessing.Pool thanks to
                    # https://stackoverflow.com/a/50409663/1034289
                    with Pool(processes=self.n_workers) as pool:
                        results = [
                            pool.apply_async(
                                shap_explain_helper, (
                                    self.model_name, 
                                    self.dataset_name,
                                    self.base_dataset_name,
                                    self.seed, 
                                    self.device,
                                    wd)
                            ) for _, wd in input_batch.items()]
                        shap_exps = [r.get() for r in results]
                        subtoken_exps = [shap_exp.values[0, :, wd.model_pred]
                                               for (_, wd), shap_exp in zip(input_batch.items(), shap_exps)]
                elif self.explainer_name == 'lime':
                    with Pool(processes=self.n_workers) as pool:
                        results = [
                            pool.apply_async(
                                lime_explain_helper, (
                                    self.model_name, 
                                    self.dataset_name,
                                    self.base_dataset_name,
                                    self.batch_size, 
                                    self.device,
                                    wd)
                            ) for _, wd in input_batch.items()]
                        lime_exps = [r.get() for r in results]    
                    for (k, wd), e in zip(input_batch.items(), lime_exps):
                        self.eval_data[k].exp = self._convert_lime_exp_to_token_exp(e, wd.model_pred)
                if self.explainer_name in ['shap', 'expred']:
                    for (k, wd), e in zip(input_batch.items(), subtoken_exps):
                            self.eval_data[k].exp = self._convert_subtoken_exp_to_token_exp(
                                wd.input_str.split(),
                                e[1:-1]  # remove the attributions for [CLS] and [SEP]
                            )
                input_batch = OrderedDict()
            if (idx + 1) % self.DUMP_EVERY == 0:
                self.eval_ckpt = self._update_dict(self.eval_ckpt, self.eval_data, override=True)
                with open(self.exp_dump_fname, 'wb') as fout:
                    pickle.dump(self.eval_ckpt, fout)
        self.eval_ckpt = self._update_dict(self.eval_ckpt, self.eval_data, override=True)
        with open(self.exp_dump_fname, 'wb') as fout:
            pickle.dump(self.eval_ckpt, fout)
        self._explanations: Dict[str, WorkingData] = self.eval_data
        return self._explanations

    def make_bogus_rules_for(self, frequent_object: Union[Tuple, str]) -> List[Tuple[Union[Tuple[Tuple[str], Tuple[str]], str], np.ndarray]]:
        bogus_rules = []
        if isinstance(frequent_object, Tuple):
            bogus_rules.append((frequent_object, np.array([[0, 1]])))
            bogus_rules.append((frequent_object, np.array([[1, 0]])))
        else:
            bogus_rules.append((frequent_object.split(), np.array([[0, 1]])))
            bogus_rules.append((frequent_object.split(), np.array([[1, 0]])))
        return bogus_rules
        
    def get_focus_objects(self):
        if self.focus_tuples is None and self.focus_strings is None:
            return None
        if 'movies' not in self.dataset_name:
            return [(tuple(q.split()), tuple(d.split())) for q, d in self.focus_tuples]
        return [tuple(s.split()) for s in self.focus_strings]
        
    @property
    def rules(self) -> List[Tuple[Union[Tuple[Tuple[str], Tuple[str]], str], np.ndarray]]:
        if self._rules is not None:
            return self._rules
        logger.info(f'loading rules from {self.rules_fname}...')
        focus_objects = self.get_focus_objects()        
        rules_dict = CausalBase.load_rules(self.rules_fname)
        self._rules = []
        for seq, pred in rules_dict.items():
            if focus_objects is None or seq in focus_objects:
                self._rules.append((seq, pred))
        logger.info(f'{len(self._rules)} rule(s) loaded')
        if focus_objects is not None:
            logger.info(f'[focus mode] this number might be less than rules presented in {self.rules_fname}')
        num_bogus_rules = 0
        logger.debug(f'rules_dict = {rules_dict.keys()}')
        if focus_objects is not None:
            for fo in focus_objects:
                if fo not in rules_dict:
                    self._rules += self.make_bogus_rules_for(fo)
                    num_bogus_rules += 2
        if num_bogus_rules > 0:
            logger.info(f'[focus mode] {num_bogus_rules} bogus rules added')
        return self._rules

    @staticmethod
    def _aggregate_rule_label(preds):
        pred_labels = np.sort(np.argmax(preds, axis=1)).tolist()
        try:
            num_pred_0 = pred_labels.index(1)
        except ValueError:
            num_pred_0 = len(pred_labels)
        num_pred_1 = len(pred_labels) - num_pred_0
        rule_label = 0 if num_pred_0 > num_pred_1 else 1
        return rule_label

    def _convert_exp_to_rank(self, exp: WorkingData,
                             pred_label: int,
                             doc: List[str],
                             exp_offset: int = 0) -> List[Tuple[int, str, float]]:
        """rank terms according to their attribution scores

        Args:
            exp (Any): attribution scores produced by self.explainer
            pred_label (int): the predict label of the current document
            doc (List[str], optional): the doc to explain in the form of token list.
            exp_offset (int): the integer offset of the explanation in exp. Used for query-doc tuple separated by [SEP].

        Returns:
            List[Tuple[int, str, float]]: list of ranked term attribtuions, each element is a 
                                          (rank order, literal term, term attribution score)
                                          triple
        """
        term_attributions = []
        if self.explainer_name == 'anchor':
            # anchor is not compatible with query-document tuples
            assert 'fever' not in self.explainer_name
            # note that here we set anchor precision to be 1.0,
            # returned explanation terms can be less equal than the doc terms,
            # for those normal terms that are considered as anchor, we set their contribution to precision as 0.
            # fist get (pos, term, accumulated precision) for each anchor term.
            anchors_poses = exp.exp_map['feature']
            names = exp.exp_map['names']
            # accumulated precision until 1.0
            precision = exp.exp_map['precision']
            # then add (pos, term, contribution to precision:0) to non-anchor terms
            other_poses = list(set([i for i in range(len(doc))]) - set(anchors_poses))
            positions = anchors_poses + other_poses
            scores = [precision[0]]
            for x, y in zip(precision[0::], precision[1::]):
                scores.append(y-x)
            scores += [0] * len(other_poses)
            terms = names + [doc[t] for t in other_poses]
            term_attributions = zip(positions, terms, scores)
        elif self.explainer_name in ['shap', 'lime', 'expred']:
            try:
                exp = exp.exp[exp_offset: exp_offset + len(doc)]
            except TypeError:
                logger.warning(f'doc = {doc}')
                logger.warning(f'exp = {exp}')
                logger.warning(f'exp_offset = {exp_offset}')
                logger.warning(f'pred_label = {pred_label}')
                raise TypeError
            try:
                assert len(exp) == len(doc)
            except AssertionError:
                logger.warning(exp)
                logger.warning(doc)
                logger.warning(exp_offset)
                logger.warning(str((len(exp), len(doc))))
                return False
            for pos, (term, score) in enumerate(zip(doc, exp)):
                term_attributions.append((pos, term, score))
        sorted_term_attributions = sorted(
            term_attributions,
            key=lambda x: x[2],
            reverse=True)
        rank = [0] * len(sorted_term_attributions)
        for r, (pos, term, score) in enumerate(sorted_term_attributions):
            rank[pos] = (r + 1, term, score)
        return rank

    @staticmethod
    def _rr(seq_pos: Tuple[int, int], feature_rank: List[Tuple[int, str, float]]):
        start, end = seq_pos
        ranks = [f[0] for f in feature_rank[start: end]]
        return 1./min(ranks)

    @staticmethod
    def _ap(seq_pos: Tuple[int, int], feature_rank: List[Tuple[int, str, float]]):
        start, end = seq_pos
        sorted_scores = sorted(
            feature_rank[start: end],
            key=lambda x: x[0],
            reverse=False)
        ap = 0.
        for i, (rank, _, _) in enumerate(sorted_scores):
            hit = i + 1
            ap += hit/rank
        ap /= (end - start)
        return ap

    @staticmethod
    def _ndcg(seq_pos: Tuple[int, int],
              feature_rank: List[Tuple[int, str, float]],
              exp_score_as_gt=True):

        def standardize(a: List):
            return [(i - min(a)) / (max(a)-min(a)) for i in a]

        start, end = seq_pos
        scores = [[0.] * start + 
                  [1.] * (end - start) + 
                  [0.] * (len(feature_rank) - end)]
        # The standardization is needed because in some version of sklearn, if the y_true
        # contains negative values, the ndcg_score will produce bogus results that may be
        # outside [0, 1] silently.
        # See https://github.com/scikit-learn/scikit-learn/issues/17639
        true_relevance = [standardize([score for _, _, score in feature_rank])]
        try:
            if exp_score_as_gt:
                ret = ndcg_score(true_relevance, scores)
            else:
                ret = ndcg_score(scores, true_relevance)
        except ValueError:
            return 0
        return ret

    def _calculate_local_agreement(self,
                                   predicate: Union[Tuple[Tuple[str], Tuple[str]], List[str]],
                                   local_exp: WorkingData,
                                   doc: Union[List[str], Tuple[List[str], List[str]]],
                                   pred_label: int):
        if 'movies' not in self.dataset_name:
            query_len = len(doc[0])
            try:
                query_rank = self._convert_exp_to_rank(
                    local_exp,
                    pred_label,
                    doc[0])
                doc_rank = self._convert_exp_to_rank(
                    local_exp,
                    pred_label,
                    doc[1],
                    exp_offset=query_len + 1)
            except AssertionError:
                logger.warn(str(local_exp))
                logger.warn(query_len)
                logger.warn(doc)
                logger.warn(self.convert_input_str_to_key(self.convert_inst_to_string(doc)))
                raise AssertionError
            if doc_rank == False:
                logger.warn(str(local_exp))
                logger.warn(query_len)
                logger.warn(doc)
                logger.warn(self.convert_input_str_to_key(self.convert_inst_to_string(doc)))
                raise AssertionError
            q_start, q_end, d_start, d_end = locate_tuple(doc, *predicate)
            ret = {}
            for n in self.score_names:
                score_fn = getattr(self, '_' + n)
                ret['query_' + n] = score_fn((q_start, q_end), query_rank)
                ret['doc_' + n] = score_fn((d_start, d_end), doc_rank)
            return ret
        else:
            query_len = 0
            exp_rank = self._convert_exp_to_rank(local_exp, pred_label, doc)
            start, end = locate_substr(doc, predicate)
            return {n: getattr(self, '_' + n)((start, end), exp_rank) for n in self.score_names}

    def _evaluate_instance(self,
                            predicate: Union[Tuple[Tuple[str], Tuple[str]], List[str]],
                            rule_label: int,
                            input: Union[List[str], Tuple[List[str], List[str]]]) -> Dict[str, float]:
        input_str = self.convert_inst_to_string(input)
        key = self.convert_input_str_to_key(input_str)
        if key not in self.explanations:
            return ERROR_DOC_NOT_SAMPLED
        pred_label = self.explanations[key].model_pred
        if pred_label == rule_label:
            local_exp = self.explanations[self.convert_input_str_to_key(input_str)]
            return self._calculate_local_agreement(predicate, local_exp, input, pred_label)
        else:
            return ERROR_DOC_NOT_COVERED

    def evaluate_rule(self, seq: Union[Tuple[Tuple[str], Tuple[str]], List[str]], rule_label) -> Dict[str, float]:
        seq_str = self.convert_inst_to_string(seq)
        if 'movies' not in self.dataset_name:
            scores = {p + n: list()
                        for n in self.score_names 
                        for p in ['query_', 'doc_']}
        
        else:
            scores = {n: list() for n in self.score_names}

        with torch.no_grad():
            candidate_instances = self.load_res_instances(seq, False)

            scores['num_candidate_documents'] = len(candidate_instances)
            label_name = self.class_names[rule_label]
            logger.info(f'evaluating rule "{seq_str} -> {label_name}", #instances containing predicates = {len(candidate_instances)}')
        
        num_doc_covered = 0
        num_doc_sampled = 0
        for doc in tqdm.tqdm(candidate_instances, desc='residence instances'):
            eval_res = self._evaluate_instance(seq, rule_label, doc)
            if eval_res == ERROR_DOC_NOT_SAMPLED:
                continue
            num_doc_sampled += 1
            if eval_res == ERROR_DOC_NOT_COVERED:
                continue
            num_doc_covered += 1
            for n, s in eval_res.items():
                scores[n].append(s)

        if 'movies' not in self.dataset_name:
            for p in ['query_', 'doc_']:
                for n in self.score_names:
                    scores[p + n] = np.mean(scores[p + n])
        else:
            for n in self.score_names:
                scores[n] = np.mean(scores[n])
        scores['coverage'] = num_doc_covered / num_doc_sampled
        scores['num_doc_match'] = int(scores['num_candidate_documents'] * scores['coverage'])
        return scores

    def will_evaluate(self, seq_str: str) -> bool:
        if self.focus_strings is not None and seq_str not in self.focus_strings:
            return False
        if self.focus_tuples is not None:
            str_focus_tuple = [q + ' [SEP] ' + d for q, d in self.focus_tuples]
            if seq_str not in str_focus_tuple:
                return False
        return True

    @property
    def results(self):
        for seq, preds in self.rules:
            seq_str = self.convert_inst_to_string(seq)
            if not self.will_evaluate(seq_str):
                logger.info(f'focus evaluation mode, {seq_str} in neither focus_strings nor focus_tuple, skip')
                continue
            rule_label = self._aggregate_rule_label(preds)
            eval_res = self.evaluate_rule(seq, rule_label)
            if 'movies' not in self.dataset_name:
                scores = [eval_res[p + n]
                            for n in self.score_names 
                                for p in ['query_', 'doc_']]
            else:
                scores = [eval_res[n] for n in self.score_names]
            coverage = eval_res['coverage']
            yield [seq_str, rule_label,
                   eval_res['num_candidate_documents'],
                   eval_res['num_doc_match'], eval_res['coverage']] + scores

    def write_results(self):
        header = ['sequence', 'label',
                  'num_candidate_documents',
                  'num_doc_match', 'coverage']
        if 'movies' not in self.base_dataset_name:
            scores_header = [
                p + 'm' + n 
                    for n in self.score_names 
                        for p in ['query_', 'doc_']]
        else:
            scores_header = ['m' + n for n in self.score_names]
        header += scores_header
        with open(self.output_fname, 'w+', newline='') as fout:
            writer = csv.writer(fout, delimiter='\t')
            writer.writerow(header)
            writer.writerows(self.results)


if __name__ == '__main__':
    # thanks to https://github.com/pytorch/pytorch/issues/3492 and https://discuss.pytorch.org/t/multiprocessing-error-with-cuda/140169/1
    # os.environ["MASTER_ADDR"] = "localhost"
    # os.environ["MASTER_PORT"] = "29500"
    torch.multiprocessing.set_start_method('spawn')
    transformers.logging.set_verbosity(transformers.logging.CRITICAL)

    init_seeds()

    args = get_args()

    config = SequenceMiningConfig(args)

    evaluator = Evaluator(config)
    evaluator.write_results()
