import csv
import os
import pickle
import subprocess
from functools import lru_cache, partial
from math import log
from multiprocessing.dummy import Pool
from typing import Any, Dict, List, Tuple, Union

import tqdm

from config import SequenceMiningConfig
from constants import EPSILON
from logging_utils import get_logger
from utils import get_model

from .desq_utils import call_miner_on, decode_stdout

logger = get_logger('basic_miner')


class BasicMiner:
    @staticmethod
    def convert_interval_to_patterns(interval: Tuple[int, int]) -> str:
        ret = [
            f'({"."*seq_length})' for seq_length in range(interval[0], interval[1] + 1)]
        return ret

    def _focus_on(self, fo: Any)->bool:
        if self.focus_strings is not None:
            return fo in self.focus_strings
        if self.focus_tuples is not None:
            if tuple(fo.split(" [SEP] ")) in self.focus_tuples:
                logger.info(f'add {fo} since it is a focus tuple')
            return tuple(fo.split(' [SEP] ')) in self.focus_tuples
        return False
    
    @property
    @lru_cache()
    def frequent_objects(self) -> Dict[Union[str, Tuple[str, str]], int]:
        temp = {}
        if os.path.isfile(self.mining_cache_fname):
            with open(self.mining_cache_fname, 'rb') as fin:
                temp = pickle.load(fin)
        else:
            for pattern in self.patterns:
                logger.info(f'mining for the pattern {pattern}')
                if pattern not in temp:
                    temp[pattern] = call_miner_on(fname=self.working_dataset_fname,
                                                class_path=self.conf.SEQUENCE_MINER_CLASS_PATH,
                                                pattern=pattern,
                                                min_support=self.min_support)
                    temp[pattern] = decode_stdout(temp[pattern])
                    if self.conf.MODEL_NAME == 'expred':
                        temp = {
                            pattern: {
                                seq: sup for seq, sup in res.items() if self.conf.RATIONALE_MASK_CHAR not in seq
                            } for pattern, res in temp.items()
                        }
                    with open(self.mining_cache_fname, 'wb+') as fout:
                        pickle.dump(temp, fout)
        ret = dict()
        logger.info('all patterns mined, start to merge the results')
        for _, t in temp.items():
            ret = {**ret, **t}
        if self.trim_support > int(self.min_support):
            temp = dict()
            for k, v in ret.items():
                if v >= self.trim_support or self._focus_on(k):
                    temp[k] = v
            return temp
        logger.debug(f'frequent_object = {ret}')
        return ret

    @property
    @lru_cache()
    def dc_given_class(self) -> Dict[str, Dict[int, Dict[Any, int]]]:
        if os.path.isfile(self.dc_cache_fname):
            logger.info(f'loading dc from cache file {self.dc_cache_fname}')
            with open(self.dc_cache_fname, 'rb') as fin:
                dc_given_class = pickle.load(fin)
            return dc_given_class

        logger.info(f'calculating #(seq/pair | class) for the first time')
        dc_given_class = {}
        for c in self.class_names:
            dc_given_class[c] = {}
            for cluster_id in range(self.n_clusters):
                dc_given_class[c][cluster_id] = {}
                if self.n_workers > 1:
                    with Pool(self.n_workers) as pool:
                        map_res = list(
                            tqdm.tqdm(
                                pool.imap_unordered(
                                    partial(self._get_dc_given_class, c, cluster_id),
                                    sorted(self.frequent_objects)),
                                total=len(self.frequent_objects)
                            )
                        )
                    for fo, dc in map_res:
                        dc_given_class[c][cluster_id][fo] = dc
                else:
                    for freq in tqdm.tqdm(list(self.frequent_objects), desc=f'dc-s in class {c}'):
                        dc_given_class[c][cluster_id][freq] = self._get_dc_given_class(
                            c, cluster_id, freq)[1]
        with open(self.dc_cache_fname, 'wb+') as fout:
            pickle.dump(dc_given_class, fout)
        return dc_given_class

    def _count_docs_given_class(self, c, cluster_id):
        doc_count = subprocess.run(["wc", "-l", f"{self.working_dataset_prefix}_{c}_{cluster_id}.txt"],
                                   stdout=subprocess.PIPE).stdout.decode('utf-8').strip().split()[0]
        return int(doc_count)

    @property
    @lru_cache()
    def n_docs_given_class(self):
        n_docs_per_class = {c: {} for c in self.class_names}
        for c in self.class_names:
            for cluster_id in range(self.n_clusters):
                n_docs_per_class[c][cluster_id] = self._count_docs_given_class(
                    c, cluster_id)
        return n_docs_per_class

    @property
    @lru_cache()
    def n_docs_all_classes(self) -> Dict[int, int]:
        n_docs_all_classes = {}
        for cluster_id in range(self.n_clusters):
            n_docs_all_classes[cluster_id] = sum(self.n_docs_given_class[c][cluster_id]
                                                 for c in self.class_names)
        return n_docs_all_classes

    @property
    @lru_cache()
    def df_given_class(self) -> Dict[str, Dict[int, Dict[Any, float]]]:
        df_given_class = {}
        for c in self.class_names:
            df_given_class[c] = {}
            for cluster_id in range(self.n_clusters):
                df_given_class[c][cluster_id] = {}
                for seq, dc in self.dc_given_class[c][cluster_id].items():
                    df_given_class[c][cluster_id][seq] = float(dc) / self.n_docs_given_class[c][cluster_id]
        return df_given_class

    @property
    @lru_cache()
    def df_all_classes(self) -> Dict[int, Dict[Any, float]]:
        df_all_classes = {}
        logger.debug('calculating the df for all classes')
        for cluster_id in range(self.n_clusters):
            df_all_classes[cluster_id] = {}
            for seq in self.frequent_objects:
                # logger.warn(f'class names: {self.class_names}')
                # logger.warn(f'dc_given_classes: {list(self.dc_given_class["irrelevant"][0].keys())}')
                dc_all_classes = sum(self.dc_given_class[c][cluster_id][seq] for c in self.class_names)
                df_all_classes[cluster_id][seq] = dc_all_classes / self.n_docs_all_classes[cluster_id]
                if df_all_classes[cluster_id][seq] == 0:
                    logger.warn(f'df for {seq} on all classes is 0')
        return df_all_classes

    def _npmi_between(self, c, freq_obj, cluster_id=0):
        p_x_given_y = self.df_given_class[c][cluster_id][freq_obj] + EPSILON
        p_x = self.df_all_classes[cluster_id][freq_obj]
        if p_x == 0:
            logger.warn(f'p_x == 0, freq_obj = {freq_obj}')
        pmi = log(p_x_given_y / p_x)
        h_x_y = -log(self.dc_given_class[c][cluster_id][freq_obj] / self.n_docs_all_classes[cluster_id] + EPSILON)
        return freq_obj, pmi / h_x_y

    @property
    @lru_cache()
    def npmi(self) -> Dict[str, Dict[int, Dict[Any, float]]]:
        logger.info('calculating the NPMIs for the first time')
        npmi = {}
        for c in self.class_names:
            npmi[c] = {}
            for cluster_id in range(self.n_clusters):
                assert cluster_id == 0
                npmi[c][cluster_id] = {}
                if self.n_workers > 1:
                    with Pool(self.n_workers) as pool:
                        map_res = list(
                            tqdm.tqdm(
                                pool.imap_unordered(
                                    partial(self._npmi_between, c),
                                    sorted(self.frequent_objects)
                                )
                            )
                        )
                    for fo, res in map_res:
                        npmi[c][cluster_id][fo] = res
                else:
                    for freq_obj in tqdm.tqdm(self.frequent_objects):
                        try:
                            fo, res = self._npmi_between(c, freq_obj)
                            npmi[c][cluster_id][freq_obj][fo] = res
                        except (ValueError, ZeroDivisionError):
                            logger.error(str(freq_obj) +
                                         f"{self.df_given_class[c][cluster_id][freq_obj]}, "
                                         f"{self.df_all_classes[cluster_id][freq_obj]},"
                                         f"{self.dc_given_class[c][cluster_id][freq_obj]},"
                                         f"{self.n_docs_all_classes[cluster_id]}")
                            raise ValueError
        return npmi

    def save_results(self):
        if not os.path.isfile(self.output_fname):
            with open(self.output_fname, 'w+', newline='') as fout:
                writer = csv.writer(fout, delimiter='\t')
                header = self.build_header()
                writer.writerow(header)
                writer.writerows(self.results)
        else:
            logger.info(f'{self.output_fname} exists, will not overwrite!')

    def __init__(self, conf: SequenceMiningConfig):
        self._pattern = None
        self.conf = conf
        self.dataset_name = conf.DATASET_NAME
        self.base_dataset_name = conf.BASE_DATASET_NAME
        self.model_name = conf.MODEL_NAME
        self.class_names = conf.CLASS_NAMES
        self.n_clusters = conf.N_CLUSTERS
        self.min_support = conf.SEQ_MINING_MIN_SUPPORT
        self.batch_size = conf.BATCH_SIZE
        self.device = conf.DEVICE
        self.n_workers = conf.N_WORKERS
        self.trim_support = conf.CAUSAL_CHECK_MIN_SUPPORT

        self.output_dir = f'{os.getenv("HOME")}/.sequence_mining/{self.dataset_name}/{self.model_name}'
        self.temp_fname_dir = f'{self.output_dir}/temp/'
        
        self.working_dataset_prefix = f'{self.output_dir}/sequences'
        self.working_dataset_fname = self.working_dataset_prefix + '.txt'
        self.output_fname = self.working_dataset_prefix + f'_mined_d{conf.DOC_SEQ_LEN_INTERVAL[0]}-{conf.DOC_SEQ_LEN_INTERVAL[1]}_sup_{conf.SEQ_MINING_MIN_SUPPORT}.tsv'
        os.makedirs(self.temp_fname_dir, exist_ok=True)

        (
            self.model_config,
            self.tokenizer,
            self.model
        ) = get_model(
            model_name=self.model_name,
            pretrained_dataset_name=self.dataset_name,
            base_dataset_name=self.base_dataset_name,
            device=self.device)

        self._frequent_objects = None

        self.focus_tuples = conf.focus_tuples
        self.focus_strings = conf.focus_strings

    def _get_dc_given_class(self, c: int, cluster_id: int, freq_obj: Any):
        raise NotImplementedError

    @property
    def results(self):
        raise NotImplementedError

    @property
    def patterns(self) -> List[str]:
        return self._pattern

    @patterns.setter
    def patterns(self, p: List[str]):
        self._pattern = p

    @property
    @lru_cache()
    def preds(self):
        raise NotImplementedError

    def build_header(self):
        raise NotImplementedError
