import functools
from logging_utils import get_logger
from typing import Dict, List, Tuple

import torch

from config import SequenceMiningConfig
from miners.basic_miner import BasicMiner
from utils import do_prediction, grep_tuple

logger = get_logger('double_miner')


class DoubleSequenceMiner(BasicMiner):
    @property
    def patterns(self):
        query_patterns = self.convert_interval_to_patterns(
            self.query_seq_len_interval)
        doc_patterns = self.convert_interval_to_patterns(
            self.doc_seq_len_interval)
        ret = []
        for qp in query_patterns:
            for dp in doc_patterns:
                ret.append(qp + ".*('[SEP]').*" + dp)
        return ret

    @property
    @functools.lru_cache()
    def frequent_objects(self) -> Dict[Tuple[str, str], int]:
        ret = dict()
        for seq, support in super().frequent_objects.items():
            query_seq, doc_seq = seq.split(' [SEP] ')
            ret[(query_seq, doc_seq)] = support
        logger.info(
            f'total number of sequence-document sequence pairs: {len(ret)}')
        return ret

    def _get_dc_given_class(self, c: str, cluster_id: int, qd_seq: Tuple[str, str]):
        fname = f"{self.working_dataset_prefix}_{c}_{cluster_id}.txt"
        dc = grep_tuple(*qd_seq, fname)
        return qd_seq, dc

    @property
    @functools.lru_cache()
    def query_sequences(self):
        return [_[0] for _ in sorted(self.frequent_objects)]

    @property
    @functools.lru_cache()
    def doc_sequences(self):
        return [_[1] for _ in sorted(self.frequent_objects)]

    @staticmethod
    def padded(s: str, front_pad_len: int = 2, end_pad_len: int = 2):
        return '. ' * front_pad_len + s + ' .' * end_pad_len

    @property
    @functools.lru_cache()
    def preds(self) -> Dict[Tuple[str, str], Tuple[float, float]]:
        docs = [self.padded(d, 2, 2) for d in self.doc_sequences]
        queries = [self.padded(q, 2, 2) for q in self.query_sequences]

        with torch.no_grad():
            self.model.eval()
            model = self.model
            if self.model_name == 'expred':
                model = self.model.cls_module
            cls_preds, _, _, _ = do_prediction(model=model,
                                               config=self.model_config,
                                               tokenizer=self.tokenizer,
                                               docs_or_dataset=docs,
                                               queries_or_none=queries)
        preds = {}
        for seq, p in zip(sorted(self.frequent_objects), cls_preds):
            preds[seq] = p
        return preds

    def build_header(self):
        ret = ['query_sequence', 'doc_sequence', 'cluster_label']
        ret.extend([f'tuple_support_in_{c}' for c in self.class_names])
        ret.extend([f'tuple_npmi_{c}' for c in self.class_names])
        ret.extend([f'model_pred_{c}' for c in self.class_names])
        return ret

    @property
    def results(self) -> List[List[str]]:
        cluster_id = 0
        results = []
        for freq_obj in self.frequent_objects:
            res = [freq_obj[0], freq_obj[1], cluster_id]
            res.extend([self.dc_given_class[c][cluster_id][freq_obj]
                        for c in self.class_names])
            res.extend([self.npmi[c][cluster_id][freq_obj]
                       for c in self.class_names])
            res.extend(self.preds[freq_obj])
            results.append(list(map(str, res)))
        return results

    def __init__(self, conf: SequenceMiningConfig):
        super().__init__(conf)
        self.doc_seq_len_interval = conf.DOC_SEQ_LEN_INTERVAL
        self.query_seq_len_interval = conf.QUERY_SEQ_LEN_INTERVAL
        self.dc_cache_fname = self.temp_fname_dir + 'dc_cache'\
            f'_q{self.query_seq_len_interval[0]}-{self.query_seq_len_interval[1]}'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.min_support}.pkl'
        self.mining_cache_fname = self.temp_fname_dir +\
            f'mining_cache'\
            f'_q{self.query_seq_len_interval[0]}-{self.query_seq_len_interval[1]}'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.min_support}.pkl'
        self.output_fname = self.working_dataset_prefix +\
            f'_mined'\
            f'_q{self.query_seq_len_interval[0]}-{self.query_seq_len_interval[1]}'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.trim_support if self.trim_support != -1 else self.min_support}.tsv'
        self.dc_cache_fname = self.temp_fname_dir + 'dc_cache'\
            f'_q{self.query_seq_len_interval[0]}-{self.query_seq_len_interval[1]}'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.trim_support}.pkl'
