import functools
import re
import subprocess
from typing import Dict, List, Tuple

import torch
from config import SequenceMiningConfig
from logging_utils import get_logger
from utils import do_prediction

from miners.basic_miner import BasicMiner

logger = get_logger('mono_miner')


class MonoSequenceMiner(BasicMiner):
    @property
    def patterns(self):
        return self.convert_interval_to_patterns(self.doc_seq_len_interval)

    def _get_dc_given_class(self, c: str, cluster_id: int, freq_sequence: str):
        grep_pattern = re.escape(freq_sequence)
        fname = f"{self.working_dataset_prefix}_{c}_{cluster_id}.txt"
        command = ['grep', '-Ewc', '--', grep_pattern, fname]
        logger.debug(f'executing command "{" ".join(command)}"')
        try:
            run = subprocess.run(command, stdout=subprocess.PIPE)
            ret = int(run.stdout.decode('utf-8'))
        except AttributeError:
            print(f'attribute error: {run.stderr} at {c}')
            raise AttributeError
        return freq_sequence, ret
        # grep_pattern = re.compile(rf"\b{re.escape(freq_sequence)}\b")
        # fname = f"{self.working_dataset_prefix}_{c}_{cluster_id}.txt"
        # dc = 0
        # with open(fname, 'r') as fin:
        #     for line in fin:
        #         if re.search(grep_pattern, line):
        #             dc += 1
        # return dc

    @property
    @functools.lru_cache()
    def preds(self) -> Dict[str, Tuple[float, float]]:
        with torch.no_grad():
            self.model.eval()
            model = self.model
            if self.model_name == 'expred':
                model = self.model.cls_module
            cls_preds, _, _, _ = do_prediction(model=model,
                                           config=self.model_config,
                                           tokenizer=self.tokenizer,
                                           docs_or_dataset=sorted(self.frequent_objects))
        preds = {}
        for seq, p in zip(sorted(self.frequent_objects), cls_preds):
            preds[seq] = p
        return preds

    def build_header(self):
        ret = ['sequence', 'cluster_label']
        ret.extend([f'dc_{c}' for c in self.class_names])
        ret.extend([f'npmi_{c}' for c in self.class_names])
        ret.extend([f'model_pred_{c}' for c in self.class_names])
        return ret

    @property
    def results(self) -> List[List[str]]:
        cluster_id = 0
        results = []
        for freq_obj in self.frequent_objects:
            res = [freq_obj, cluster_id]
            res.extend([self.dc_given_class[c][cluster_id][freq_obj] for c in self.class_names])
            res.extend([self.npmi[c][cluster_id][freq_obj] for c in self.class_names])
            res.extend(self.preds[freq_obj])
            results.append(list(map(str, res)))
        return results

    def __init__(self, conf: SequenceMiningConfig):
        super().__init__(conf)
        self.doc_seq_len_interval = conf.DOC_SEQ_LEN_INTERVAL
        self.dc_cache_fname = self.temp_fname_dir + 'dc_cache'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.min_support}.pkl'
        self.mining_cache_fname = self.temp_fname_dir +\
            f'mining_cache'\
            f'_d{self.doc_seq_len_interval[0]}-{self.doc_seq_len_interval[1]}'\
            f'_sup_{self.min_support}.pkl'
