import math
import os
import time
import pickle
from functools import partial
from itertools import chain
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from expred import BertDataset
from torch.utils.data import IterableDataset
from transformers import BasicTokenizer

from logging_utils import get_logger
from models.inputs import LSTMBatch
from sentence_transformers import InputExample

logger = get_logger('Dataset')


class BaseEraserDataset(IterableDataset):
    def __init__(self,
                 dataset: BertDataset,
                 max_input_len: int = 512,
                 start: int = 0,
                 end: int = 0,
                 class_names: List[str] = None,
                 num_workers: int = 1):
        super().__init__()
        dataset.shuffle = False
        data = list(dataset)
        self.ann_ids, self.queries, self.docs, self.labels = list(zip(*data))
        self.max_len = max_input_len
        assert end > start
        self.start = start
        self.end = end
        self.class_names = class_names
        self.num_workers = num_workers
        self.preprocessed = False

    def get_encoded_label(self, orig_label: str) -> int:
        return self.class_names.index(orig_label)

    def _chain_inputs_helper(self, qd_tuple: Tuple):
        q, d = qd_tuple
        return ' '.join(q) + ' ' + ' '.join(chain.from_iterable(d))

    def __len__(self) -> int:
        return len(self.ann_ids)

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            iter_start = self.start
            iter_end = self.end
        else:
            iter_per_worker = int(
                math.ceil((self.end - self.start)/float(worker_info.num_workers)))
            worker_id = worker_info.id
            iter_start = self.start + worker_id * iter_per_worker
            iter_end = min(iter_start + iter_per_worker, self.end)
        return iter(self[i] for i in range(iter_start, iter_end))

class LSTMEraserDataset(BaseEraserDataset):
    def __init__(self,
                 dataset: BertDataset,
                 max_input_len: int = 512,
                 start: int = 0,
                 end: int = 0,
                 embed_size: int = 0,
                 class_names: List[str] = None,
                 embedding: Dict[str, np.ndarray] = None,
                 tokenizer: BasicTokenizer = None,
                 num_workers: int = 1,
                 cache_dir: str = None,
                 split_name: str = None):
        super().__init__(
            dataset=dataset,
            max_input_len=max_input_len,
            start=start,
            end=end,
            class_names=class_names,
            num_workers=num_workers,
            cache_dir=cache_dir,
            split_name=split_name
        )
        self.embed_size = embed_size
        self.embedding = embedding
        self.tokenizer = tokenizer
        self.cache_fname = None
        if cache_dir is not None and split_name is not None:
            self.cache_fname = f'{cache_dir}/{split_name}.pkl'

    def __getitem__(self, item: int) -> Tuple:
        item = item % len(self)
        if self.preprocessed:
            return self.ann_ids[item], self.input_embeddings[item], self.encoded_labels[item], self.onehot_labels[item]
        else:
            return self.ann_ids[item], self.queries[item], self.docs[item], self.labels[item]

    def get_embeddings(self, words: List[str]) -> List[np.ndarray]:
        emb_mean, emb_std = -0.005838499, 0.48782197
        oov = 0
        embedding_matrix = np.random.normal(
            emb_mean, emb_std, (min(len(words), self.max_len), self.embed_size))
        for i, word in enumerate(words):
            if i >= self.max_len:
                break
            if word not in self.embedding:
                oov += 1
            embedding_matrix[i] = self.embedding.get(word, embedding_matrix[i])
        return embedding_matrix, oov, len(embedding_matrix)

    def maybe_preprocess(self, invalidate_cache=False):
        if self.cache_fname is not None and os.path.isfile(self.cache_fname) and not invalidate_cache:
            logger.info(f'loading from dumpped input cache {self.cache_fname}')
            with open(self.cache_fname, 'rb') as fin:
                (
                    self.ann_ids,
                    self.inputs,
                    self.input_embeddings,
                    self.encoded_labels,
                    self.onehot_labels
                ) = pickle.load(fin)
            return self

        logger.info(f'preprocessing the dataset, len = {len(self.ann_ids)}')
        self.preprocessed = True
        if self.num_workers > 1:
            from multiprocessing import Pool
            pool = Pool(self.num_workers)
            map_fcn = pool.map
        else:
            map_fcn = map
        logger.info('chaining inputs')
        tick = time.process_time()
        self.inputs = list(map_fcn(self._chain_inputs_helper,
                           zip(self.queries, self.docs)))
        logger.info(f'chaining inputs takes {time.process_time() - tick}s')

        ret = list(map_fcn(self.get_embeddings, self.inputs))
        self.input_embeddings, num_oovs, len_docs = list(zip(*ret))
        logger.info(
            f'avg length of docs: {sum(len_docs)/float(len(len_docs))}')
        logger.info(f'avg #oov {sum(num_oovs)/float(len(num_oovs))}')

        self.input_embeddings = list(
            map_fcn(partial(torch.tensor, dtype=torch.float), self.input_embeddings))
        self.encoded_labels = list(
            map_fcn(self.get_encoded_label, self.labels))
        self.onehot_labels = F.one_hot(torch.tensor(
            self.encoded_labels), num_classes=len(self.class_names))

        if self.num_workers > 1:
            pool.close()
        logger.info('preprocessing done')

        if self.cache_fname is not None and (not os.path.isfile(self.cache_fname) or invalidate_cache):
            logger.info(
                f'dumping preprocessed input data as cache under {self.cache_fname}')
            with open(self.cache_fname, 'wb+') as fout:
                pickle.dump((
                    self.ann_ids,
                    self.inputs,
                    self.input_embeddings,
                    self.encoded_labels,
                    self.onehot_labels
                ), fout)

        return self

    @classmethod
    def maybe_from_preprocessed(
        cls,
        dataset: BertDataset,
        max_input_len: int = 512,
        start: int = 0, end: int = 0,
        embed_size: int = 0,
        class_names: List[str] = None,
        embedding: Dict[str, np.ndarray] = None,
        tokenizer: BasicTokenizer = None,
        num_workers: int = 1,
        cache_dir: str = None,
        split_name=None
    ):
        ret = cls(
            dataset, max_input_len=max_input_len,
            start=start, end=end,
            embed_size=embed_size,
            class_names=class_names,
            embedding=embedding,
            tokenizer=tokenizer,
            num_workers=num_workers,
            split_name=split_name
        )
        cache_fname = f'{cache_dir}/{split_name}.pkl'
        if cache_dir is not None and os.path.isfile(cache_fname):
            with open(cache_fname, 'rb') as fin:
                ret.ann_ids, ret.inputs, ret.input_embeddings, ret.encoded_labels, ret.onehot_labels = pickle.load(
                    fin)
        return ret


def lstm_collate_wrapper(batch, config, embedding, tokenizer, preprocessed_dataset=False):
    if preprocessed_dataset:
        return LSTMBatch.from_preprocessed_dataset(batch, config, embedding, tokenizer)
    return LSTMBatch.from_bert_dataset(batch, config, embedding, tokenizer)


class SBERTEraserDataset(BaseEraserDataset):
    def __init__(self,
                 dataset: BertDataset,
                 max_input_len: int = 512,
                 start: int = 0,
                 end: int = 0,
                 class_names: List[str] = None,
                 num_workers: int = 1):
        super().__init__(
            dataset=dataset,
            max_input_len=max_input_len,
            start=start,
            end=end,
            class_names=class_names,
            num_workers=num_workers,
        )
        self.encoded_labels = [self.get_encoded_label(l) for l in self.labels]

    def __getitem__(self, item: int) -> Tuple:
        item = item % len(self)
        q = ' '.join(self.queries[item])
        d = ' '.join(chain.from_iterable(self.docs[item]))
        l = self.encoded_labels[item]
        return InputExample(texts=[q, d], label=l)


def sbert_collate_wrapper(batch):
    return batch
    _, queries, docs, encoded_labels = zip(*batch)
    queries = [' '.join(words) for words in queries]
    docs = [list(chain.from_iterable(doc)) for doc in docs] if docs is not None else None
    docs = [' '.join(words) for words in docs]
    ret = [InputExample(texts=[q, d], label=l) for q, d, l in zip(queries, docs, encoded_labels)]
    # print(ret)
    return ret