from itertools import chain
from typing import Dict, List, Tuple, Union

import numpy as np
import torch
from expred import BertDataset
from logging_utils import get_logger
from models.model_config import LSTMConfig, SBERTConfig
from torch import Tensor, tensor
from torch.nn import functional as F
from torch.nn.utils.rnn import pad_sequence
from transformers import BasicTokenizer

logger = get_logger('lstm_batch')


from transformers import BertTokenizer
bt = BertTokenizer.from_pretrained('bert-base-uncased')


class BaseBatch:
    def __init__(self,
                 queries: List[List[str]],
                 docs: List[List[List[str]]],
                 labels: List[str],
                 config: LSTMConfig,
                 ann_ids: List[str]):
        self.orig_queries = queries
        self.orig_sentences: List[List[List[str]]] = docs
        self.orig_docs = [list(chain.from_iterable(doc)) for doc in docs] if docs is not None else None
        self.orig_labels = labels
        self.ann_ids = ann_ids

        self.class_names = config.class_names
        self.embed_size = config.embed_size

    def get_encoded_label(self, orig_label: str) -> int:
        return self.class_names.index(orig_label)

    def get_decoded_cls_pred(self, p: List[float]) -> str:
        return self.class_names[np.argmax(p)]
    
    def get_decoded_cls_preds(self, cls_preds: List[List[float]]) -> List[str]:
        logger.debug(str(cls_preds))
        return [self.get_decoded_cls_pred(p) for p in cls_preds]

class LSTMBatch(BaseBatch):
    def __init__(self,
                 queries: List[List[str]],
                 docs: List[List[List[str]]],
                 labels: List[str],
                 config: LSTMConfig,
                 ann_ids: List[str],
                 embedding_mapping: Dict[str, np.ndarray],
                 tokenizer: BasicTokenizer = None,
                 preprocessed_dataset:bool=False):
        super().__init__(
            queries=queries,
            docs=docs,
            labels=labels,
            config=config,
            ann_ids=ann_ids
        )

        self.embed_size = config.embed_size

        if tokenizer is None:
            tokenizer = BasicTokenizer()
        self.tokenizer = tokenizer
        self.embedding_mapping = embedding_mapping
        self.preprocessed_dataset = preprocessed_dataset

        # initializing for lazy calculation
        self.tokenized_queries = None
        self.tokenized_docs = None
        self.query_embeddings = None
        self.docs_embeddings = None
        self.input_embeddings = None
        self.encoded_labels = None
        self.onehot_labels = None

    def get_glove_embeddings(self, words: List[str]) -> List[np.ndarray]:
        emb_mean, emb_std = -0.005838499, 0.48782197
        embedding = np.random.normal(
            emb_mean, emb_std, (len(words), self.embed_size))
        for i, word in enumerate(words):
            embedding[i] = self.embedding_mapping.get(
                word, embedding[i])
        return embedding

    def preprocess(self):
        if self.input_embeddings is None:
            self.tokenized_queries = [self.tokenizer.tokenize(' '.join(words)) for words in self.orig_queries]
            self.tokenized_docs = [self.tokenizer.tokenize(' '.join(words)) for words in self.orig_docs]
            self.query_embeddings = [self.get_glove_embeddings(words) for words in self.tokenized_queries]
            self.docs_embeddings = [self.get_glove_embeddings(words) for words in self.tokenized_docs]
            self.input_embeddings = [torch.cat((tensor(q, dtype=torch.float), tensor(d, dtype=torch.float)), dim=0)
                                    for q, d in zip(self.query_embeddings, self.docs_embeddings)]
        if self.encoded_labels is None:
            self.encoded_labels = [self.get_encoded_label(label) for label in self.orig_labels]
        if self.onehot_labels is None:
            self.onehot_labels = F.one_hot(tensor(self.encoded_labels), num_classes=len(self.class_names))
        return self

    @classmethod
    def from_bert_dataset(cls,
                          dataset: Union[BertDataset, List[Tuple[List, List, List, List]]],
                          config: LSTMConfig,
                          embedding: Dict[str, np.ndarray],
                          tokenizer: BasicTokenizer):
        ann_ids, queries, docs, labels = zip(*dataset)
        return cls(queries, docs, labels, config, ann_ids, embedding, tokenizer)

    @classmethod
    def from_preprocessed_dataset(cls,
                                  preprocessed_dataset: List[Tuple[List, Tensor, Tensor, Tensor]],
                                  config:LSTMConfig,
                                  embedding: Dict[str, np.ndarray],
                                  tokenizer: BasicTokenizer):
        ret = cls(None, None, None, config, None, embedding, tokenizer)
        ret.ann_ids, ret.input_embeddings, ret.encoded_labels, ret.onehot_labels = zip(*preprocessed_dataset)
        return ret
    
    def get_decoded_docs(self)->List[List[str]]:
        [self.tokenizer.tokenize(
            bt.decode(
                bt(
                    ' '.join(d), 
                    add_special_tokens=False
                )['input_ids']
            )
        ) for d in self.tokenized_docs]
        return self.tokenized_docs

# class SBERTBatch(BaseBatch):
#     def __init__(self,
#                  queries: List[List[str]],
#                  docs: List[List[List[str]]],
#                  labels: List[str],
#                  config: LSTMConfig,
#                  ann_ids: List[str]):
#         super().__init__(
#             queries=queries,
#             docs=docs,
#             labels=labels,
#             config=config,
#             ann_ids=ann_ids
#         )
#         self.tokenized_queries = [' '.join(words) for words in self.orig_queries]
#         self.tokenized_docs = [' '.join(words) for words in self.orig_docs]
#         self.texts = list(zip(self.tokenized_queries, self.tokenized_docs))

#     @classmethod
#     def from_bert_dataset(cls,
#                           dataset: Union[BertDataset, List[Tuple[List, List, List, List]]],
#                           config: SBERTConfig):
#         ann_ids, queries, docs, labels = zip(*dataset)
#         return cls(queries, docs, labels, config, ann_ids)