import os
from typing import Dict, List

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.autograd import Variable
from torch.nn.utils.rnn import (pack_padded_sequence, pad_packed_sequence,
                                pad_sequence)
from transformers import BasicTokenizer

from models.model_config import LSTMConfig
from constants import EPSILON, INFINITY


class MultiheadLSTM(nn.Module):
    _embedding = None

    def init_params(self):
        for key, params in self.named_parameters():
            if 'weight' in key:
                torch.nn.init.xavier_normal_(params)
            elif 'bias' in key:
                torch.nn.init.zeros_(params)

    def __init__(self, conf: LSTMConfig) -> None:
        super().__init__()
        self.embed_size = conf.embed_size
        self.max_vocab_size = conf.max_vocab_size
        self.max_len = conf.max_len
        self.batch_size = conf.batch_size
        self.hidden_dims = conf.hidden_dims
        self.n_layers = conf.n_layers
        self.dropout_rate = conf.dropout_rate
        self.num_classes = len(conf.class_names)
        self.embedding_fname = os.environ['HOME'] + \
            '/.keras/embeddings/glove.840B.300d/glove.840B.300d.txt'
        self.device = conf.device

        self.loss_fn = nn.CrossEntropyLoss()

        self.tokenizer = BasicTokenizer(do_lower_case=False)

        self.lstm = nn.LSTM(
            self.embed_size,
            self.hidden_dims[0],
            num_layers=self.n_layers,
            bidirectional=True,
            batch_first=True)
        
        self.attention = nn.MultiheadAttention(self.hidden_dims[0] * 2, 1, batch_first=True)
        self.batch_norm0 = nn.BatchNorm1d(self.hidden_dims[0] * 2)
        self.dropout0 = nn.Dropout(self.dropout_rate)
        self.fc0 = nn.Linear(self.hidden_dims[0] * 2, self.hidden_dims[1])
        self.leaky_relu0 = nn.LeakyReLU()

        self.batch_norm1 = nn.BatchNorm1d(self.hidden_dims[1])
        self.dropout1 = nn.Dropout(self.dropout_rate)
        self.fc1 = nn.Linear(self.hidden_dims[1], self.num_classes)
        self.softmax = nn.Softmax(dim=-1)

        self.init_params()

    def mask_softmax(self, matrix:Tensor, mask:Tensor=None):
        """Perform softmax on length dimension with masking.

        Parameters
        ----------
        matrix: torch.float, shape [batch_size, .., max_len]
        mask: torch.long, shape [batch_size, max_len]
            Mask tensor for sequence.

        Returns
        -------
        output: torch.float, shape [batch_size, .., max_len]
            Normalized output in length dimension.
        """

        if mask is None:
            result = F.softmax(matrix, dim=-1)
        else:
            mask_norm = ((1 - mask) * (-INFINITY)).to(matrix)
            for _ in range(matrix.dim() - mask_norm.dim()):
                mask_norm = mask_norm.unsqueeze(1)
            result = F.softmax(matrix + mask_norm, dim=-1)
        return result

    def mask_mean(self, seq, mask:Tensor=None):
        """Compute mask average on length dimension.

        Parameters
        ----------
        seq : torch.float, size [batch, max_seq_len, n_channels],
            Sequence vector.
        mask : torch.long, size [batch, max_seq_len],
            Mask vector, with 0 for mask.

        Returns
        -------
        mask_mean : torch.float, size [batch, n_channels]
            Mask mean of sequence.
        """

        if mask is None:
            return torch.mean(seq, dim=1)

        mask_sum = torch.sum(seq * mask.unsqueeze(-1).float(), dim=1)  # [b,msl,nc]->[b,nc]
        seq_len = torch.sum(mask, dim=-1)  # [b]
        mask_mean = mask_sum / (seq_len.unsqueeze(-1).float() + EPSILON)
        return mask_mean

    def mask_max(self, seq, mask=None):
        """Compute mask max on length dimension.

        Parameters
        ----------
        seq : torch.float, size [batch, max_seq_len, n_channels],
            Sequence vector.
        mask : torch.long, size [batch, max_seq_len],
            Mask vector, with 0 for mask.

        Returns
        -------
        mask_max : torch.float, size [batch, n_channels]
            Mask max of sequence.
        """

        if mask is None:
            return torch.mean(seq, dim=1)
        mask_max, _ = torch.max(seq + (1 - mask.unsqueeze(-1).float()) * (-INFINITY), dim=1)# [b,msl,nc]->[b,nc]
        return mask_max

    def seq_mask(self, seq_len, max_len):
        """Create sequence mask.

        Parameters
        ----------
        seq_len: torch.long, shape [batch_size],
            Lengths of sequences in a batch.
        max_len: int
            The maximum sequence length in a batch.

        Returns
        -------
        mask: torch.long, shape [batch_size, max_len]
            Mask tensor for sequence.
        """

        idx = torch.arange(max_len).to(seq_len).repeat(seq_len.size(0), 1)
        mask = torch.gt(seq_len.unsqueeze(1), idx).to(seq_len)
        return mask

    def self_attn_mask(self, seq_lens: Tensor, max_seq_len: int) -> Tensor:
        # e.g. seq_lens = [4, 2], max_seq_len = 4
        # idx : 
        # [
        #   [0, 1, 2, 3],
        #   [0, 1, 2, 3]
        # ]
        idx = torch.arange(max_seq_len).to(seq_lens).repeat(seq_lens.size(0), 1)  # [b, msl]
        # mask1 (before repeat):
        # [
        #   [[1], [1], [1], [1]],
        #   [[1], [1], [0], [0]]
        # ]
        # mask1 (after repeat):
        # [
        #   [
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #   ],
        #   [
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #       [0, 0, 0, 0],
        #       [0, 0, 0, 0],
        # ]
        mask1 = torch.gt(seq_lens.unsqueeze(1), idx).unsqueeze(-1).repeat(1, 1, max_seq_len).to(seq_lens)  # [b, msl]

        # mask1 (before unsqueeze):
        # [
        #   [[1, 1, 1, 1]],
        #   [[1, 1, 0, 0]]
        # ]
        # mask1 (after repeat):
        # [
        #   [
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #       [1, 1, 1, 1],
        #   ],
        #   [
        #       [1, 1, 0, 0],
        #       [1, 1, 0, 0],
        #       [1, 1, 0, 0],
        #       [1, 1, 0, 0],
        # ]
        mask2 = torch.gt(seq_lens.unsqueeze(1), idx).unsqueeze(-2).repeat(1, max_seq_len, 1).to(seq_lens)  # [b, msl]
        return (1 - mask1 * mask2).type(torch.bool)

    def forward(self, x: List[Tensor]):
        seq_lengths = list(map(lambda _: _.shape[0], x))
        max_seq_len = torch.max(torch.tensor(seq_lengths, dtype=torch.long))

        padded_input = pad_sequence(x, batch_first=True)
        packed_input = pack_padded_sequence(
            padded_input, lengths=seq_lengths, enforce_sorted=False, batch_first=True).to(self.device)
        output, (_, _) = self.lstm(packed_input)
        padded_output, _ = pad_packed_sequence(output, batch_first=True)  # [b, msl]
        mask = self.seq_mask(torch.tensor(seq_lengths, dtype=torch.long), max_seq_len).to(padded_output)  # [b,msl]
        padded_output = padded_output * mask
        
        attn_mask = self.self_attn_mask(torch.tensor(seq_lengths, dtype=torch.long), max_seq_len).to(padded_output)
        attn_output:Tensor = self.attention(padded_output, padded_output, padded_output, attn_mask=attn_mask)
        batch_norm_output0 = self.batch_norm0(attn_output.transpose(-1, -2))
        dropout_output0 = self.dropout0(batch_norm_output0)
        fc_output0 = self.fc(dropout_output0)
        attn_output = self.leaky_relu0(fc_output0)

        batch_norm_output1 = self.batch_norm0(attn_output.transpose(-1, -2))
        dropout_output1 = self.dropout0(batch_norm_output1)
        logits = self.fc1(dropout_output1)

        return logits

    def maybe_download(self):
        if not os.path.isfile(self.embedding_fname):
            embedding_folder = os.path.dirname(self.embedding_fname)
            embedding_zip_fname = f'{embedding_folder}/glove.840B.300d.zip'
            import urllib.request
            urllib.request.urlretrieve(
                'https://nlp.stanford.edu/data/glove.840B.300d.zip', embedding_zip_fname)
            import zipfile
            with zipfile.ZipFile(embedding_zip_fname, 'r') as zip_ref:
                zip_ref.extractall(embedding_folder)
            assert os.path.isfile(self.embedding_fname)
        with open(self.embedding_fname) as fin:
            for line in fin:
                yield line

    @property
    def embedding(self) -> Dict[str, np.ndarray]:
        if self._embedding is not None:
            return self._embedding

        def get_coefs(word: str, *arr, do_lower_case=True):
            if do_lower_case:
                word = word.lower()
            return word, np.asarray(arr, dtype='float32')[:self.embed_size]

        index_to_embedding = dict()
        for idx, line in enumerate(self.maybe_download()):
            if idx >= self.max_vocab_size:
                break
            word, ebd = get_coefs(*line.split(' '))
            index_to_embedding[word] = ebd
        self._embedding = index_to_embedding
        return self._embedding

    def get_loss(self, preds: Tensor, labels: Tensor):
        return self.loss_fn(preds, labels)
