import logging
import os
import pickle
from functools import partial
from typing import Any

import sklearn
import torch
import tqdm
from expred import BertDataset
from torch.utils.data import DataLoader
from transformers import BasicTokenizer

from config import DEBUG, SequenceMiningConfig
from constants import INFINITY
from logging_utils import get_logger
from models.dataset import LSTMEraserDataset, lstm_collate_wrapper
from models.inputs import LSTMBatch
from models.model_config import LSTMConfig
from models.lstm.lstm_multihead import MultiheadLSTM
from models.lstm.model import BiLSTMModel
from utils import get_args

logger = get_logger("lstm_trainer")
if DEBUG:
    logger.setLevel(logging.DEBUG)


class Trainer:
    def __init__(self, args):
        self.dataset_name = args.dataset_name
        self.sm_config = SequenceMiningConfig(args)
        self.config = LSTMConfig.from_seq_mining_config(self.sm_config)
        self.batch_size = self.config.batch_size
        self.total_train_steps = self.config.total_train_steps
        self.device = self.config.device
        self.ckpt_dir = self.config.ckpt_dir
        os.makedirs(self.ckpt_dir, exist_ok=True)
        self.res_dir = self.config.res_dir
        os.makedirs(self.res_dir, exist_ok=True)
        self.train_res_fname = f'{self.res_dir}/train_res.txt'
        self.val_res_fname = f'{self.res_dir}/val_res.txt'
        self.test_res_fname = f'{self.res_dir}/test_res.txt'
        # self.cache_dir = self.config.cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)
        self.print_every = self.config.print_every
        self.val_every = self.config.val_every
        self.initial_lr = self.config.initial_lr
        self.start_batch_index = 0
        self.min_val_loss = INFINITY

        self.model = BiLSTMModel(self.config).to(self.device)
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.initial_lr)
        self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer, 'min')
        # self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda _: 1)

        self.tokenizer = BasicTokenizer()

    def init_dataset(self):
        N_WORKSERS = 20
        self.dataset = BertDataset(self.dataset_name, shuffle=(not DEBUG))
        self.dataset.select_datasplit('train')
        self.dataset_train = LSTMEraserDataset(
            self.dataset, max_input_len=self.config.max_len,
            start=0, end=self.total_train_steps * self.batch_size,
            embed_size=self.config.embed_size,
            class_names=self.config.class_names,
            embedding=self.model.embedding,
            tokenizer=self.tokenizer, num_workers=N_WORKSERS,
            cache_dir=self.cache_dir,
            split_name='train').maybe_preprocess()
        self.train_loader = DataLoader(
            self.dataset_train, batch_size=self.batch_size, shuffle=False,
            collate_fn=partial(
                lstm_collate_wrapper,
                config=self.config,
                embedding=self.model.embedding,
                tokenizer=self.tokenizer,
                preprocessed_dataset=self.dataset_train.preprocessed),
            pin_memory=True, num_workers=N_WORKSERS)

        self.dataset.select_datasplit('val')
        self.dataset_val = LSTMEraserDataset(
            self.dataset, max_input_len=self.config.max_len,
            start=0, end=len(self.dataset),
            embed_size=self.config.embed_size,
            class_names=self.config.class_names,
            embedding=self.model.embedding,
            tokenizer=self.tokenizer, num_workers=N_WORKSERS,
            cache_dir=self.cache_dir,
            split_name='val').maybe_preprocess()
        self.val_loader = DataLoader(
            self.dataset_val, batch_size=self.batch_size, shuffle=False,
            collate_fn=partial(
                lstm_collate_wrapper,
                config=self.config,
                embedding=self.model.embedding,
                tokenizer=self.tokenizer,
                preprocessed_dataset=self.dataset_val.preprocessed),
            pin_memory=True, num_workers=N_WORKSERS)

        self.dataset.select_datasplit('test')
        self.dataset_test = LSTMEraserDataset(
            self.dataset, max_input_len=self.config.max_len,
            start=0, end=len(self.dataset),
            embed_size=self.config.embed_size,
            class_names=self.config.class_names,
            embedding=self.model.embedding,
            tokenizer=self.tokenizer, num_workers=N_WORKSERS,
            cache_dir=self.cache_dir,
            split_name='test').maybe_preprocess()
        self.test_loader = DataLoader(
            self.dataset_test, batch_size=self.batch_size, shuffle=False,
            collate_fn=partial(
                lstm_collate_wrapper,
                config=self.config,
                embedding=self.model.embedding,
                tokenizer=self.tokenizer,
                preprocessed_dataset=self.dataset_test.preprocessed),
            pin_memory=True, num_workers=N_WORKSERS)
        return self

    def predict_batch(self, batch: LSTMBatch):
        batch_input_embeddings = batch.preprocess().input_embeddings
        batch_labels = torch.tensor(batch.encoded_labels, device=self.device)

        if DEBUG:
            logger.debug('\n')
            logger.debug(f'lstm_batch.tokenzed_queries = {batch.tokenized_queries}')
            logger.debug(f'lstm_batch.tokenized_docs = {batch.tokenized_docs}')
            logger.debug(f'lstm_batch.q_ebd.shape = {(len(batch.query_embeddings), batch.query_embeddings[0].shape)}')
            logger.debug(f'lstm_batch.q_ebd = {batch.query_embeddings}')
            logger.debug(f'lstm_batch.d_ebd.shape = {(len(batch.docs_embeddings), batch.docs_embeddings[0].shape)}')
            logger.debug(f'lstm_batch.d_ebd = {batch.docs_embeddings}')
            logger.debug(f'lstm_batch.encoded_labels = {batch.encoded_labels}')
            self.model.eval()
            lstm_outputs = self.model(batch.input_embeddings)
            logger.debug(f'lstm_outputs = {lstm_outputs}')
            logger.debug(f'lstm_outputs = {lstm_outputs["cls_pred_logit"]}')
            assert False

        batch_outputs = self.model(batch_input_embeddings)['cls_pred_logit']
        loss = self.model.get_loss(batch_outputs, batch_labels)
        return loss, batch_outputs

    def train_on_batch(self, batch: LSTMBatch):
        self.model.train()
        self.optimizer.zero_grad()
        loss, _ = self.predict_batch(batch)
        loss.backward()
        self.optimizer.step()
        return loss

    def _checkpoint_exists(self):
        for asset in ['model', 'optimizer', 'scheduler', 'config']:
            asset_ckpt_fname = self.ckpt_dir + f'/{asset}.ckpt'
            if not os.path.isfile(asset_ckpt_fname):
                return False
        return True

    def _load_ckpts(self):
        self.model.load_state_dict(torch.load(f'{self.ckpt_dir}/model.ckpt'))
        self.optimizer.load_state_dict(
            torch.load(f'{self.ckpt_dir}/optimizer.ckpt'))
        self.scheduler.load_state_dict(
            torch.load(f'{self.ckpt_dir}/scheduler.ckpt'))
        with open(f'{self.ckpt_dir}/config.ckpt', 'rb') as fin:
            config = pickle.load(fin)
            self.start_batch_index = config['step']
            self.min_val_loss = config['min_val_loss'] if 'min_val_loss' in config else INFINITY
        logger.info('checkpoint loaded')
        return self

    def _is_time_to_record_train_res(self):
        if self.batch_index % self.print_every == 0 and self.batch_index > 0:
            return True
        return False

    def _is_time_to_val(self):
        if self.batch_index % self.val_every == 0 and self.batch_index > 0:
            return True
        return False

    def _record_train_res(self, average_loss):
        res_str = f"Average loss_x at batch {self.batch_index}: {float(average_loss):0.4}"
        with open(self.train_res_fname, 'a+') as fout:
            fout.write(res_str + '\n')

    def _dump_ckpts(self):
        torch.save(self.model.state_dict(), f'{self.ckpt_dir}/model.ckpt')
        torch.save(self.optimizer.state_dict(),
                   f'{self.ckpt_dir}/optimizer.ckpt')
        torch.save(self.scheduler.state_dict(),
                   f'{self.ckpt_dir}/scheduler.ckpt')
        conf = {'step': self.batch_index,
                'initial_lr': self.initial_lr,
                'batch_size': self.batch_size,
                'min_val_loss': self.min_val_loss}
        with open(f'{self.ckpt_dir}/config.ckpt', 'wb+') as fout:
            pickle.dump(conf, fout)

    def infer_on(self, dataloader: DataLoader):
        labels, preds = [], []
        loss_sum = 0.0
        for batch in dataloader:
            loss, batch_pred = self.predict_batch(batch)

            loss_sum += loss.detach() * len(batch_pred)
            preds += batch_pred.tolist()
            labels += batch.encoded_labels

        avg_loss = loss_sum / len(dataloader.dataset)
        return preds, labels, avg_loss

    def validate_on(self, dataloader: DataLoader):
        preds, encoded_labels, avg_loss = self.infer_on(dataloader)
        # print(list(zip(encoded_labels, preds)))
        argmaxed_preds = torch.argmax(torch.tensor(preds), dim=-1)
        accuracy = sklearn.metrics.accuracy_score(encoded_labels,
                                                  argmaxed_preds)
        return avg_loss, accuracy

    def _store_res(self, loss, accuracy, name='Validation'):
        res_fname = self.test_res_fname if name == 'Test' else self.val_res_fname
        with open(res_fname, 'a+') as fout:
            prompt_str = f"####   {name} batch index: {self.batch_index}    ####"
            logger.info(prompt_str)
            fout.write(prompt_str + '\n')

            val_loss_str = f'{name} loss: {loss}'
            logger.info(val_loss_str)
            fout.write(val_loss_str + '\n')

            accuracy_str = f'{name} accuracy: {accuracy}'
            logger.info(accuracy_str)
            fout.write(accuracy_str + '\n')

    def do_val(self):
        with torch.no_grad():
            self.model.eval()
            loss, accuracy = self.validate_on(self.val_loader)
            self._store_res(loss, accuracy, 'Validation')
        return loss

    def do_test(self):
        with torch.no_grad():
            self.model.eval()
            loss, accuracy = self.validate_on(self.test_loader)
            self._store_res(loss, accuracy, 'Test')
        return accuracy

    def do_train(self):
        if self._checkpoint_exists():
            self._load_ckpts()

        average_loss = 0

        total = self.total_train_steps - self.start_batch_index
        for i, batch in tqdm.tqdm(enumerate(self.train_loader), total=total):
            self.batch_index = self.start_batch_index + i
            if self.batch_index > self.total_train_steps:
                break

            loss = self.train_on_batch(batch)

            average_loss += loss.detach()

            if self._is_time_to_record_train_res():
                average_loss /= self.print_every
                self._record_train_res(average_loss)
                average_loss = 0

            if self._is_time_to_val():
                val_loss = self.do_val()
                self.min_val_loss = min(self.min_val_loss, val_loss)
                self.scheduler.step(val_loss)
                self._dump_ckpts()

        self._dump_ckpts()
        return self.model

    def train(self):
        self.init_dataset()
        self.model = self.do_train()
        accuracy = self.do_test()
        print(f'test accuracy: {accuracy}')


if __name__ == '__main__':
    args = get_args()
    Trainer(args).train()
