import os
from typing import List

from config import DEBUG, SequenceMiningConfig


class ModelConfigBase:
    model_name = None

    @property
    def hyper_parameter_desc(self) -> str:
        raise NotImplementedError
    
    def __init__(self,
                 dataset_name,
                 max_len,
                 class_names):
        self.dataset_name = dataset_name
        self.class_names = class_names

        self.max_len = max_len
        self.num_classes = len(self.class_names)

        self.device = 'cuda'


class LSTMBase(ModelConfigBase):
    @classmethod
    def from_seq_mining_config(cls, conf: SequenceMiningConfig):
        return cls(
            dataset_name=conf.DATASET_NAME,
            embed_size=conf.EMBEDDING_SIZE,
            max_vocab_size=conf.MAX_VOCAB_SIZE,
            max_len=conf.MAXLEN,
            dropout_rate=conf.DROPOUT_RATE,
            class_names=conf.CLASS_NAMES)
    
    @property
    def hyper_parameter_desc(self) -> str:
        return f'{self.n_layers}_layers_{self.hidden_dims[0]}x{len(self.hidden_dims)}_dims_{self.total_train_steps}_steps'

    def __init__(self,
                 dataset_name,
                 embed_size,
                 max_vocab_size,
                 max_len,
                 dropout_rate,
                 class_names):
        super().__init__(
            dataset_name=dataset_name,
            max_len=max_len,
            class_names=class_names)

        self.embed_size = embed_size
        self.max_vocab_size = max_vocab_size
        self.dropout_rate = dropout_rate

        if self.dataset_name == 'sstmovies':
            self.n_layers = 2
            self.hidden_dims = [16] * 2
            self.batch_size = 800
            self.total_train_steps = 3000
        elif self.dataset_name == 'climate-fever':
            self.n_layers = 4
            self.hidden_dims = [128] * 2
            self.batch_size = 1200
            self.total_train_steps = 1000
        elif self.dataset_name == 'movies':
            self.n_layers = 4
            self.hidden_dims = [128] * 2
            self.batch_size = 400
            self.total_train_steps = 1000
        elif self.dataset_name == 'multirc':
            self.n_layers = 4
            self.hidden_dims = [256] * 2
            self.batch_size = 400
            self.total_train_steps = 1000

        if DEBUG:
            self.batch_size = 4
        
        self.print_every = 1
        self.val_every = 20
        self.initial_lr = 1e-4

        self.res_base_dir = os.environ['HOME'] + f'/.keras/trained_models/{self.dataset_name}/{self.model_name}/'
        # self.cache_dir = f'{self.res_base_dir}/input_cache/'
        self.ckpt_dir = f'{self.res_base_dir}/{self.hyper_parameter_desc}'
        self.res_dir = f'{self.ckpt_dir}/res'


class LSTMConfig(LSTMBase):
    model_name = 'bilstm_att'

    
class MultiheadLSTMConfig(LSTMBase):
    model_name = 'multihead_lstm'


class SBERTConfig(ModelConfigBase):
    pretrained_transformer_model_name = 'bert-base-uncased'
    model_name = 'sbert'

    @property
    def hyper_parameter_desc(self) -> str:
        return f'{self.dense_dim}_dim_{self.epochs_total}_epochs'

    def __init__(
        self,
        dataset_name: str,
        max_len: int,
        class_names: List[str]):
        
        self.dense_dim = 256
        self.batch_size = 54
        self.epochs_total = 10

        super().__init__(
            dataset_name,
            max_len,
            class_names
        )

        self.print_every = 1
        self.val_every = 20
        # self.initial_lr = 1e-4

        if self.dataset_name == 'climate-fever':
            self.batch_size = 50
        elif self.dataset_name == 'multirc':
            self.batch_size = 48
        if DEBUG:
            # self.batch_size = 2
            self.epochs_total = 1

        self.res_base_dir = os.environ['HOME'] + f'/.keras/trained_models/{self.dataset_name}/{self.model_name}/'
        # self.cache_dir = f'{self.res_base_dir}/input_cache/'
        self.ckpt_dir = f'{self.res_base_dir}/{self.hyper_parameter_desc}'
        self.res_dir = f'{self.ckpt_dir}/res'

    @classmethod
    def from_seq_mining_config(cls, conf: SequenceMiningConfig):
        return cls(
            dataset_name=conf.DATASET_NAME,
            max_len=conf.MAXLEN,
            class_names=conf.CLASS_NAMES)