from itertools import chain
import os
from typing import List

import torch
from sentence_transformers import (InputExample, SentenceTransformer, losses,
                                   models)
from torch import nn

from logging_utils import get_logger
from models.model_config import SBERTConfig

logger = get_logger(__name__)


class SBERT(nn.Module):
    def __init__(self, conf: SBERTConfig):
        super().__init__()
        self.max_seq_length = conf.max_len
        self.dense_dim = conf.dense_dim
        self.num_classes = len(conf.class_names)
        self.pretrained_transformer_model_name = conf.pretrained_transformer_model_name
        self.device = conf.device
        self.ckpt_dir = conf.ckpt_dir

        word_embedding_model = models.Transformer(
            'bert-base-uncased',
            max_seq_length=self.max_seq_length,
            cache_dir=f'{conf.ckpt_dir}/bert',
            do_lower_case=True).to('cuda')
        pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension()).to('cuda')
        dense_model = models.Dense(
            in_features=pooling_model.get_sentence_embedding_dimension(),
            out_features=self.dense_dim,
            activation_function=nn.Tanh()).to('cuda')
        self.siamese_model = SentenceTransformer(modules=[word_embedding_model, pooling_model, dense_model]).to('cuda')

        self.training_loss = losses.SoftmaxLoss(
            self.siamese_model, 
            sentence_embedding_dimension=self.siamese_model.get_sentence_embedding_dimension(),
            num_labels=self.num_classes).to('cuda')

        self.init_loss_model_params()

    @classmethod
    def from_pretrained(cls, conf:SBERTConfig, best_step:int = None):
        model = cls(conf)
        if best_step is None:
            folders = list(filter(lambda x: x.isdigit(), os.listdir(model.ckpt_dir)))
            steps = list(map(int, folders))
            best_step = max(steps)
        path = f'{model.ckpt_dir}/{best_step}/'
        for idx, name in enumerate(model.siamese_model._modules):
            module = model.siamese_model._modules[name]
            if idx == 0 and isinstance(module, models.Transformer):    #Save transformer model in the main folder
                module_dir = path + "/"
                module = module.auto_model
            elif idx == 1:
                continue
            else:
                module_dir = os.path.join(path, str(idx)+"_"+type(module).__name__)
            with open(f'{module_dir}/pytorch_model.bin', 'rb') as fin:
                module.load_state_dict(torch.load(fin))
        with open(f'{model.ckpt_dir}/loss_model/pytorch_model.bin', 'rb') as fin:
            model.training_loss.classifier.load_state_dict(torch.load(fin))
        logger.info('pretrained model loaded')
        return model

    def init_loss_model_params(self):
        for key, params in self.training_loss.classifier.named_parameters():
            if 'weight' in key:
                nn.init.xavier_normal_(params)
            elif 'bias' in key:
                nn.init.zeros_(params)

    def save_training_loss_model(self):
        module_ckpt_dir = f'{self.ckpt_dir}/loss_model'
        os.makedirs(module_ckpt_dir, exist_ok=True)
        with open(f'{module_ckpt_dir}/pytorch_model.bin', 'wb') as fout:
            torch.save(self.training_loss.classifier.state_dict(), fout)

    def predict_on_batch(
        self,
        batch: List[InputExample],
        return_concatenated_embedding=True):
        
        logger.debug(f'self.siamese_model.device: {self.siamese_model.device}')
        
        sentences = [example.texts for example in batch]
        sentences = list(chain.from_iterable(sentences))
        
        embeddings = self.siamese_model.encode(
            sentences,
            convert_to_tensor=True,
            convert_to_numpy=False)
        logger.debug(f'embeddings.shape={embeddings.shape}')
        
        embeddings = torch.reshape(embeddings, (len(batch), 2, -1))
        logger.info(f'embeddings.shape={embeddings.shape}')

        rep_a, rep_b = torch.transpose(embeddings, dim0=0, dim1=1)
        logger.debug(f'rep_a.shape: {rep_a.shape}, rep_b.shape: {rep_b.shape}')
        
        vectors_concat = []
        # by default, softmax loss concatenate both embedding and the embedding difference of the query and the doc
        vectors_concat.append(rep_a)
        vectors_concat.append(rep_b)
        vectors_concat.append(torch.abs(rep_a - rep_b))

        features = torch.cat(vectors_concat, -1)
        output = self.training_loss.classifier(features)
        labels = torch.tensor([ex.label for ex in batch], dtype=torch.long).to(self.device)
        loss = self.training_loss.loss_fct(output, labels.view(-1))
        if return_concatenated_embedding:
            return output, loss, features
        return output, loss