import json
import logging
import os
import pickle
from argparse import Namespace
from math import ceil
from typing import Dict, List

import sklearn
import torch
from expred import BertDataset
from sentence_transformers import InputExample
from sentence_transformers.evaluation import BinaryClassificationEvaluator
from torch.utils.data import DataLoader

from config import DEBUG, SequenceMiningConfig
from constants import INFINITY
from logging_utils import get_logger
from models.dataset import (SBERTEraserDataset, lstm_collate_wrapper,
                            sbert_collate_wrapper)
from models.model_config import SBERTConfig
from models.sbert.model import SBERT
from utils import get_args

logger = get_logger("lstm_trainer")
if DEBUG:
    logger.setLevel(logging.DEBUG)


class Trainer:
    def __init__(self, args: Namespace):
        self.dataset_name = args.dataset_name
        self.sm_config = SequenceMiningConfig(args)
        config = SBERTConfig.from_seq_mining_config(self.sm_config)
        self.batch_size = config.batch_size
        self.max_seq_length = config.max_len
        self.class_names = config.class_names
        self.device = config.device
        self.ckpt_dir = config.ckpt_dir
        os.makedirs(self.ckpt_dir, exist_ok=True)
        self.res_dir = config.res_dir
        os.makedirs(self.res_dir, exist_ok=True)
        self.test_res_fname = f'{self.res_dir}/test_res.txt'

        # self.cache_dir = config.cache_dir
        # os.makedirs(self.cache_dir, exist_ok=True)

        self.epochs_total = config.epochs_total
        self.print_every = config.print_every
        self.val_every = config.val_every
        # self.initial_lr = config.initial_lr
        self.start_batch_index = 0
        self.min_val_loss = INFINITY

        if self._checkpoint_exists():
            self.model = SBERT.from_pretrained(config).to(self.device)
            self.epoch_start = self.load_epoch_json().get('epoch', -1)
        else:
            self.model = SBERT(config)
            self.epoch_start = -1

    def init_dataset(self):
        N_WORKSERS = 20
        self.dataset = BertDataset(self.dataset_name, shuffle=(not DEBUG))
        self.dataset.select_datasplit('train')
        self.dataset_train = SBERTEraserDataset(
            self.dataset, max_input_len=self.max_seq_length,
            start=0, end=len(self.dataset),
            class_names=self.class_names,
            num_workers=N_WORKSERS)
        self.train_loader = DataLoader(
            self.dataset_train, batch_size=self.batch_size, shuffle=False,
            collate_fn=sbert_collate_wrapper,
            pin_memory=True, num_workers=N_WORKSERS)

        self.dataset.select_datasplit('val')
        self.dataset_val = SBERTEraserDataset(
            self.dataset, max_input_len=self.max_seq_length,
            start=0, end=len(self.dataset),
            class_names=self.class_names,
            num_workers=N_WORKSERS)
        self.val_loader = DataLoader(
            self.dataset_val, batch_size=self.batch_size, shuffle=False,
            collate_fn=sbert_collate_wrapper,
            pin_memory=True, num_workers=N_WORKSERS)

        self.dataset.select_datasplit('test')
        self.dataset_test = SBERTEraserDataset(
            self.dataset, max_input_len=self.max_seq_length,
            start=0, end=len(self.dataset),
            class_names=self.class_names,
            num_workers=N_WORKSERS)
        self.test_loader = DataLoader(
            self.dataset_test, batch_size=self.batch_size, shuffle=False,
            collate_fn=sbert_collate_wrapper,
            pin_memory=True, num_workers=N_WORKSERS)
        return self

    def _checkpoint_exists(self):
        folders = list(filter(lambda x: x.isdigit(), os.listdir(self.ckpt_dir)))
        if len(folders) > 0:
            return True
        return False

    def infer_on(self, dataloader: DataLoader):
        labels, preds = [], []
        loss_sum = 0.0
        for batch in dataloader:
            batch_pred, loss = self.model.predict_batch(batch)

            loss_sum += loss.detach() * len(batch_pred)
            preds += batch_pred.tolist()
            labels += [ex.label for ex in batch]

        avg_loss = loss_sum / len(dataloader.dataset)
        return preds, labels, avg_loss

    def do_test(self):
        with torch.no_grad():
            self.model.eval()
            preds, encoded_labels, avg_loss = self.infer_on(self.test_loader)
            argmaxed_preds = torch.argmax(torch.tensor(preds), dim=-1)
            accuracy = sklearn.metrics.accuracy_score(encoded_labels, argmaxed_preds)
            self._store_test_res(avg_loss, accuracy, 'Test')
        return avg_loss, accuracy

    def _store_test_res(self, loss, accuracy, name='Validation'):
        res_fname = self.test_res_fname
        with open(res_fname, 'a+') as fout:
            val_loss_str = f'{name} loss: {loss}'
            logger.info(val_loss_str)
            fout.write(val_loss_str + '\n')

            accuracy_str = f'{name} accuracy: {accuracy}'
            logger.info(accuracy_str)
            fout.write(accuracy_str + '\n')

    def save_epoch_callback(self, score, epoch, step):
        with open(f'{self.ckpt_dir}/current_epoch.json', 'w+') as fout:
            json.dump({
                'score': score,
                'epoch': epoch,
                'step': step
            }, fout)

    def load_epoch_json(self) -> Dict:
        with open(f'{self.ckpt_dir}/current_epoch.json', 'r+') as fin:
            return json.load(fin)

    def do_train(self):
        steps_per_epoch = ceil(int(len(self.dataset_train) / self.batch_size))
        epochs = self.epochs_total - self.epoch_start - 1
        self.model.siamese_model.fit(
            train_objectives=[(self.train_loader, self.model.training_loss)],
            show_progress_bar=True,
            evaluation_steps=self.val_every,
            evaluator=self.evaluator,
            steps_per_epoch=steps_per_epoch,
            output_path=self.res_dir,
            checkpoint_path=self.ckpt_dir,
            checkpoint_save_steps=steps_per_epoch,
            epochs=epochs,
            warmup_steps=100,
            callback=self.save_epoch_callback,
            checkpoint_save_total_limit=1)
        self.model.save_training_loss_model()
        return self.model

    def get_evaluator(self):
        queries = []
        docs = []
        labels = []
        for batch in self.val_loader:
            q, d = list(zip(*[ex.texts for ex in batch]))
            l = [ex.label for ex in batch]
            queries += q
            docs += d
            labels += l
        return BinaryClassificationEvaluator(queries, docs, labels)
    
    def train(self):
        self.init_dataset()
        self.evaluator = self.get_evaluator()
        self.model = self.do_train()
        accuracy = self.do_test()
        print(f'test accuracy: {accuracy}')


if __name__ == '__main__':
    os.environ['TOKENIZERS_PARALLELISM'] = 'false'
    args = get_args()
    Trainer(args).train()
