import os
import pickle
from typing import IO, List, Tuple, Union

import numpy as np
import torch.cuda
from expred import BertDataset
from sklearn.cluster import KMeans

from config import SequenceMiningConfig
from models.lstm.model import BiLSTMModel
from logging_utils import get_logger
from utils import do_prediction, get_model

logger = get_logger('predictor')


class Predictor:
    def clustering(self, X):
        cluster_labels = KMeans(n_clusters=self.n_clusters, random_state=42).fit(X).labels_
        return cluster_labels

    def load_preds(self):
        logger.info(f'using dumped prediction file {self.cache_fname}')
        with open(self.cache_fname, 'rb') as fin:
            loaded = pickle.load(fin)
            # for compatibility
            if isinstance(loaded, dict):
                decoded_cls_preds = loaded['decoded_cls_preds']
                rationale_embeddings = loaded['rationale_embedding']
                decoded_exp_preds = loaded['decoded_exp_preds']
            else:
                decoded_cls_preds, rationale_embeddings, decoded_exp_preds = loaded
        return decoded_cls_preds, rationale_embeddings, decoded_exp_preds

    def cache_preds(self,
                    decoded_cls_preds: List[str],
                    rationale_embeddings: List[np.ndarray],
                    decoded_exp_preds: List[Union[List[str],
                                                  Tuple[List[str], List[str]]]],
                    force_dump_rationale_embedding=False):
        with open(self.cache_fname, 'wb+') as fout:
            # to save some space
            if self.model_name == 'sbert' and not force_dump_rationale_embedding:
                pickle.dump((decoded_cls_preds, None, decoded_exp_preds), fout)
            pickle.dump((decoded_cls_preds, rationale_embeddings, decoded_exp_preds), fout)

    def maybe_predict(self):
        if os.path.isfile(self.cache_fname):
            logger.info(f'found cache file {self.cache_fname}, loading from cache')
            decoded_cls_preds, rationale_embeddings, decoded_inputs = self.load_preds()
        else:
            logger.info(f'predicting from beginning, will be dumped as {self.cache_fname}')
            dataset = BertDataset(self.dataset_name)
            with torch.no_grad():
                self.model.eval()
                (
                    _,
                    decoded_cls_preds,
                    rationale_embeddings,
                    decoded_inputs
                ) = do_prediction(self.model,
                                  self.model_config,
                                  self.tokenizer,
                                  batch_size=self.batch_size,
                                  docs_or_dataset=dataset,
                                  return_queries=self.dump_queries,
                                  rationale_mask_char=self.rationale_mask_char)
            self.cache_preds(decoded_cls_preds, rationale_embeddings, decoded_inputs)
        return decoded_cls_preds, rationale_embeddings, decoded_inputs

    def convert_to_dict(self, inputs: List[Union[str, Tuple[str, str]]]):  # either doc-only or (query,doc) tuple
        ret = {c: {cluster_id: [] for cluster_id in self.cluster_preds} for c in self.class_names}
        for cluster_id, c, _input in zip(self.cluster_preds, self.decoded_cls_preds, inputs):
            ret[c][cluster_id].append(_input)
        logger.debug(f'return of convert_to_dict {ret}')
        return ret

    def preprocess(self):
        (
            self.model_config,
            self.tokenizer,
            self.model
        ) = get_model(
            model_name=self.model_name,
            pretrained_dataset_name=self.dataset_name,
            base_dataset_name=self.base_dataset_name,
            device=self.device
        )
        self.decoded_cls_preds, docs_ebd, inputs = self.maybe_predict()
        # self.cluster_preds = self.clustering(docs_ebd)
        self.cluster_preds = [0] * len(docs_ebd)
        self.instances = self.convert_to_dict(inputs)
        return self

    def _write_instance(self, instance: Union[str, Tuple[str, str], List[str]],
                        fout: IO):
        if self.dump_queries:
            q, d = instance
            fout.write(' '.join(q).lower() + ' [SEP] ')
        else:
            d = instance
        fout.write(' '.join(d).lower() + '\n')

    def dump_instances_for(self, class_to_dump, cluster_to_dump):
        fname = f'{self.working_dataset_prefix}_{class_to_dump}_{cluster_to_dump}.txt'
        with open(fname, 'w+') as fout:
            logger.info(f'dumping for class {class_to_dump} cluster {cluster_to_dump} to {fname}')
            logger.debug(f'the first instance is '
                         f'{self.instances[class_to_dump][cluster_to_dump][0][0]} - '
                         f'{self.instances[class_to_dump][cluster_to_dump][0][1][:10]}')
            for instance in self.instances[class_to_dump][cluster_to_dump]:
                self._write_instance(instance, fout)

    def dump_all_instances(self):
        fname = self.working_dataset_fname
        with open(fname, 'w+') as fout:
            for c in self.class_names:
                for cluster_id in range(self.n_clusters):
                    for instance in self.instances[c][cluster_id]:
                        self._write_instance(instance, fout)

    def dump(self):
        self.dump_all_instances()
        for class_to_dump in self.class_names:
            for cluster_to_dump in range(self.n_clusters):
                self.dump_instances_for(class_to_dump, cluster_to_dump)

    def __init__(self, conf: SequenceMiningConfig):
        self.instances = None
        self.cluster_preds = None
        self.decoded_cls_preds = None
        self.docs = None
        self.model = None
        self.tokenizer = None
        self.model_config = None
        self.batch_size = conf.BATCH_SIZE
        self.n_clusters = conf.N_CLUSTERS
        self.class_names = conf.CLASS_NAMES
        self.dataset_name = conf.DATASET_NAME
        self.base_dataset_name = conf.BASE_DATASET_NAME
        self.model_name = conf.MODEL_NAME
        self.device = conf.DEVICE
        self.rationale_mask_char = conf.RATIONALE_MASK_CHAR

        self.dump_queries = ('movies' != self.base_dataset_name and 'sstmovies' != self.base_dataset_name)
        self.output_dir = f'{os.getenv("HOME")}/.sequence_mining/{self.dataset_name}/{self.model_name}'
        self.temp_fname_dir = f'{self.output_dir}/temp/'
        os.makedirs(self.temp_fname_dir, exist_ok=True)
        self.cache_fname = self.temp_fname_dir + 'pred_cache.pkl'
        self.working_dataset_prefix = f'{self.output_dir}/sequences'
        self.working_dataset_fname = self.working_dataset_prefix + '.txt'


class LSTMPredictor:
    def __init__(self, conf: SequenceMiningConfig):
        self.instances = None
        self.cluster_preds = None
        self.decoded_cls_preds = None
        self.docs = None
        
        embed_size = 300 # how big is each word vector
        max_features = 120000 # how many unique words to use (i.e num rows in embedding vector)
        maxlen = 70 # max number of words in a question to use
        batch_size = conf.BATCH_SIZE # how many samples to process at once
        n_epochs = 5 # how many times to iterate over all samples

        self.max_sentence_length = conf.MAX_VOCAB_SIZE
        self.model = BiLSTMModel
        self.tokenizer = None
        self.model_config = None
        self.n_clusters = conf.N_CLUSTERS
        self.class_names = conf.CLASS_NAMES
        self.dataset_name = conf.DATASET_NAME
        self.model_name = conf.MODEL_NAME
        self.device = conf.DEVICE

        self.dump_queries = (self.dataset_name == 'fever')
        self.output_dir = f'{os.getenv("HOME")}/.sequence_mining/{self.dataset_name}/{self.model_name}'
        self.temp_fname_dir = f'{self.output_dir}/temp/'
        os.makedirs(self.temp_fname_dir, exist_ok=True)
        self.cache_fname = self.temp_fname_dir + 'pred_cache.pkl'
        self.working_dataset_prefix = f'{self.output_dir}/sequences'
        self.working_dataset_fname = self.working_dataset_prefix + '.txt'
