from collections import namedtuple
from dataclasses import dataclass, field
from logging_utils import get_logger
from typing import List, NamedTuple, Tuple, Union

import torch.nn
from expred import BertConfig
from torch import nn
from transformers import BertTokenizer

from candidate_sequence import (CandidateSequence, CandidateSequenceTuple,
                                MaskedContext, MaskedTupleContext)
from utils import do_prediction

logger = get_logger('synthetics')


# IDE friendly NameTuple
class SimplifiedSyntheticInputs(NamedTuple):
    """The simplified version of SyntheticInput containing no nn model relevant
    stuffs using an IDE friendly NameTuple
    """
    seq: List[str]
    syn_inputs: List[str]
    preds: List[Union[Tuple[float], List[float]]]

@dataclass
class SyntheticInputs:
    seq: List[str]
    syn_inputs: List[str]
    # model: nn.Module = field(repr=False)
    # model_config: BertConfig = field(repr=False)
    # tokenizer: BertTokenizer = field(repr=False)
    batch_size: int = field(repr=False)
    preds: List[Union[Tuple[float], List[float]]]
    _preds: List[Union[Tuple[int], List[int]]] = field(
        init=False, repr=False, default=None)
    
    @staticmethod
    def _pred_syn_instances(syn_inputs: List[str],
                            model: nn.Module,
                            model_config: BertConfig,
                            tokenizer: BertTokenizer,
                            batch_size: int) -> List:
        logger.debug(f'#synthetic instances = {len(syn_inputs)}')
        space_char = ' '
        logger.debug(f'the 1st syn_query-syn_doc looks like: "{space_char.join(syn_inputs[0][:10])} ..."')

        with torch.no_grad():
            model.eval()
            logger.debug(f'batch_size = {batch_size}')
            preds = do_prediction(
                model, model_config, tokenizer,
                docs_or_dataset=syn_inputs,
                batch_size=batch_size,
                queries_or_none=None,
                tqdm_desc='predict synthetic instances')[0]
        return preds

    # @property
    # def preds(self) -> List[Union[Tuple[float], List[float]]]:
    #     if self._preds is not None:
    #         return self._preds
    #     logger.debug(f'predicting synthetic documents of "{" ".join(self.seq)}"')
    #     logger.debug(f'#synthetic documents = {len(self.syn_inputs)}')
    #     if len(self.syn_inputs) > 0:
    #         logger.debug(f'the first syn_input starts with: {self.syn_inputs[0][:100]}')
    #     with torch.no_grad():
    #         self.model.eval()
    #         preds = do_prediction(self.model, self.model_config, self.tokenizer,
    #                               self.syn_inputs,
    #                               batch_size=self.batch_size)[0]
    #     self._preds = preds
    #     return self._preds

    # @preds.setter
    # def preds(self, p: List[Union[Tuple[int], List[int]]]):
    #     self._preds = p
        
    @property
    def simplified(self) -> namedtuple:
        return SimplifiedSyntheticInputs(seq=self.seq,
                                         syn_inputs=self.syn_inputs,
                                         preds=self.preds)
    
    @staticmethod
    def _synthesize(s: List[str], t: MaskedContext) -> str:
        return ' '.join(t.context[:t.mask_start] + s + t.context[t.mask_end:])
    
    @classmethod
    def from_simplified(cls,
                        simp: SimplifiedSyntheticInputs,
                        # model: nn.Module,
                        # model_config:BertConfig,
                        # tokenizer: BertTokenizer,
                        batch_size: int):
        return cls(seq=simp.seq,
                   syn_inputs=simp.syn_inputs,
                   preds=simp.preds,
                #    model=model,
                #    model_config=model_config,
                #    tokenizer=tokenizer,
                   batch_size=batch_size)
        
    @classmethod
    def from_seq_context(cls, 
                         s: CandidateSequence, 
                         c: List[CandidateSequence],
                        #  model: nn.Module,
                        #  model_config: BertConfig,
                        #  tokenizer: BertTokenizer,
                         batch_size: int):
        syn_inputs = []
        for s_prime in c:
            # if s.seq == s_prime.seq:
            #     continue
            for t in s_prime.selected_masked_context:
                syn_inputs.append(cls._synthesize(s.seq, t))
        return SyntheticInputs(seq=s.seq, syn_inputs=syn_inputs,
                            #    model=model,
                            #    model_config=model_config,
                            #    tokenizer=tokenizer,
                               batch_size=batch_size,
                               preds=None)

@dataclass
class SyntheticTupleInputs:
    qd_tuple: Tuple[List[str], List[str]]
    syn_queries: List[List[str]]
    syn_docs: List[List[str]]
    preds: List[Union[Tuple[float], List[float]]]

    def __eq__(self, __o: object) -> bool:
        if self.qd_tuple != __o.qd_tuple:
            return False
        if self.syn_queries != __o.syn_queries:
            return False
        if self.syn_docs != __o.syn_docs:
            return False
        if self.preds != __o.preds:
            return False
        return True
    
    @staticmethod
    def synthesize(q_seq: List[str], d_seq: List[str],
                   c: MaskedTupleContext) -> Tuple[List[str], List[str]]:
        syn_q = c.context[0][:c.mask_start_query] + \
            q_seq + c.context[0][c.mask_end_query:]
        syn_d = c.context[1][:c.mask_start_doc] + \
            d_seq + c.context[1][c.mask_end_doc:]
        return syn_q, syn_d
    
    @staticmethod
    def _pred_syn_instances(syn_queries: List[str],
                            syn_docs: List[str],
                            model: nn.Module,
                            model_config: BertConfig,
                            tokenizer: BertTokenizer,
                            batch_size: int) -> List:
        logger.debug(f'#synthetic instances = {len(syn_docs)}')
        space_char = ' '
        logger.debug(f'the 1st syn_query-syn_doc looks like: "{space_char.join(syn_queries[0])}" - "{space_char.join(syn_docs[0][:10])} ..."')

        with torch.no_grad():
            model.eval()
            logger.debug(f'batch_size = {batch_size}')
            preds = do_prediction(
                model, model_config, tokenizer,
                docs_or_dataset=syn_docs,
                batch_size=batch_size,
                queries_or_none=syn_queries,
                tqdm_desc='predict synthetic instances')[0]
        return preds
    
    @classmethod
    def from_seq_tuple_context(
        cls,
        s: CandidateSequenceTuple,
        context_tuples: List[CandidateSequenceTuple]):
        syn_queries = []
        syn_docs = []
        # preds = []
        for context in context_tuples:
            # if s.qd_tuple == context.qd_tuple:
            #     continue
            for c in context.selected_masked_context:
                syn = cls.synthesize(*s.qd_tuple, c)
                syn_queries.append(syn[0])
                syn_docs.append(syn[1])
        logger.debug(f'predicting synthetic instances of "{s.qd_tuple}"')
        # preds = cls._pred_syn_instances(
        #     syn_queries=syn_queries,
        #     syn_docs=syn_docs,
        #     model=model,
        #     model_config=model_config,
        #     tokenizer=tokenizer,
        #     batch_size=batch_size)
        return cls(
            qd_tuple=s.qd_tuple,
            syn_queries=syn_queries,
            syn_docs=syn_docs,
            preds=None)
