import os
import unittest

import torch
from candidate_sequence import CandidateSequence, CandidateSequenceTuple, MaskedContext, MaskedTupleContext
from config import SequenceMiningConfig
from utils import abs_diff, do_prediction, get_model, sentence_seek


class TestCandidateSequence(unittest.TestCase):
    def setUp(self) -> None:
        self.seq = 'a lion'.split()
        self.dataset_name = 'test'
        self.model_name = 'bert'
        self.working_dataset_fname = f'{os.getenv("HOME")}/.sequence_mining/test/bert/test_working_dataset.txt'
        self.device = torch.device('cpu')
        self.model_config, self.tokenizer, self.model = get_model(
            self.model_name, self.dataset_name, self.device)
        self.config = SequenceMiningConfig()
        self.cand_seq = CandidateSequence(
            seq=self.seq,
            dataset_name=self.dataset_name,
            working_dataset_fname=self.working_dataset_fname,
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            sentences_as_context=False,
            device=self.device,
            context_criterion='abs_diff',
            batch_size=1)

    def test__init__(self):
        cand_seq = CandidateSequence(
            seq=self.seq,
            dataset_name=self.dataset_name,
            working_dataset_fname=self.working_dataset_fname,
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            sentences_as_context=False,
            device=self.device,
            context_criterion='abs_diff',
            batch_size=1)

    def test_res_docs(self):
        ret = self.cand_seq.res_docs
        ret_should = ['this is line one , i have a lion'.split(),
                      'well this line three also has a lion'.split()]
        self.assertEqual(len(ret), len(ret_should))
        for r, rs in zip(ret, ret_should):
            self.assertSequenceEqual(r, rs)

    def test_res_doc_setter(self):
        self.cand_seq._res_docs = None
        ret_should = [['test']]
        self.cand_seq.res_docs = ret_should
        self.assertSequenceEqual(ret_should, self.cand_seq.res_docs)

    def test_mask(self):
        context = 'this is a test'.split()
        start = 1
        end = 3
        ret = self.cand_seq.mask(context, start, end)
        ret_should = 'this . . test'.split()
        self.assertSequenceEqual(ret_should, ret)

    def test_get_masked_doc(self):
        self.cand_seq.seq = 'this is'.split()
        ret = self.cand_seq.get_masked_doc('this is a test'.split())
        ret_should = MaskedContext(context='. . a test'.split(),
                                   mask_start=0,
                                   mask_end=2)
        self.assertEqual(ret_should, ret)

    def test_get_masked_sentences(self):
        doc = 'this is sentence one . this is sentence two wow . .'.split()
        self.cand_seq.seq = 'sentence one .'.split()
        ret = self.cand_seq.get_masked_sentences(doc)
        ret_should = MaskedContext(context='this is . . .'.split(),
                                   mask_start=2,
                                   mask_end=5)
        self.assertEqual(ret_should, ret)

        self.cand_seq.seq = 'sentence one . this is'.split()
        ret = self.cand_seq.get_masked_sentences(doc)
        ret_should = MaskedContext(context='this is . . . . . sentence two wow .'.split(),
                                   mask_start=2,
                                   mask_end=7)
        self.assertEqual(ret_should, ret)

        self.cand_seq.seq = 'sentence two'.split()
        ret = self.cand_seq.get_masked_sentences(doc)
        ret_should = MaskedContext(context='this is . . wow .'.split(),
                                   mask_start=2,
                                   mask_end=4)
        self.assertEqual(ret_should, ret)

    def test_masked_res_context(self):
        self.cand_seq.res_docs = ['this is doc one . doc one sentence two .'.split(),
                                  'and i am also one sentence .'.split()]
        self.cand_seq.seq = 'one sentence'.split()
        self.cand_seq.sentences_as_context = False
        ret = self.cand_seq.masked_res_context
        ret_should = [MaskedContext(context='this is doc one . doc . . two .'.split(),
                                    mask_start=6,
                                    mask_end=8),
                      MaskedContext(context='and i am also . . .'.split(),
                                    mask_start=4,
                                    mask_end=6)]
        for r, rs in zip(ret, ret_should):
            self.assertEqual(rs, r)

        self.cand_seq.sentences_as_context = True
        self.cand_seq._masked_res_context = None
        ret = self.cand_seq.masked_res_context
        ret_should = [MaskedContext(context='doc . . two .'.split(),
                                    mask_start=1,
                                    mask_end=3),
                      MaskedContext(context='and i am also . . .'.split(),
                                    mask_start=4,
                                    mask_end=6)]
        for r, rs in zip(ret, ret_should):
            self.assertEqual(rs, r)

    def test_masked_res_context_preds(self):
        self.cand_seq.masked_res_docs = [MaskedContext(context="this is line one , i have . .".split(),
                                                       mask_start=8, mask_end=10),
                                         MaskedContext(context='well this line three also has . .'.split(),
                                                       mask_start=7, mask_end=9)]
        ret = self.cand_seq.masked_res_context_preds
        self.assertEqual(2, len(ret))
        for r in ret:
            self.assertEqual(2, len(r))

    def test_get_neutral_masked_context(self):
        self.cand_seq.masked_res_context = ['a b c'.split(), 'e d f'.split()]
        self.cand_seq.masked_res_context_preds = [[0.61, 0.39], [0.46, 0.54]]
        self.cand_seq._selected_masked_context = None
        ret = self.cand_seq.selected_masked_context
        self.assertEqual(1, len(ret))
        print(ret)
        self.assertSequenceEqual('e d f'.split(), ret[0])


class TestCandidateSequenceTuple(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_working_dir = '/home/zzhang/.sequence_mining/test/bert/'
        self.mock_cache_dir = self.mock_working_dir + 'temp/'
        self.mock_working_dataset_fname = self.mock_working_dir + 'sequences.txt'
        self.mock_qd_tuple = ('is this true'.split(), 'true . because'.split())
        self.mock_docs = ['is this true ? [SEP] well i am curious whether is this true . because i do not like it . sentence three .'.split(),
                          'will that be true ? [SEP] i am pretty sure that is not true .'.split()]
        self.device = 'cpu'
        with open(self.mock_working_dataset_fname, 'w+') as fout:
            for doc in self.mock_docs:
                fout.write(' '.join(doc) + '\n')

        self.model_config, self.tokenizer, self.model = get_model(
            'bert', 'fever', self.device)
        self.cst = CandidateSequenceTuple(
            qd_tuple=self.mock_qd_tuple)
        self.cst.init_functions(
            dataset_name='fever',
            working_dataset_fname=self.mock_working_dataset_fname,
            sentences_as_context=False,
            device=self.device,
            context_criterion='all',
            batch_size=2,
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,)

        return super().setUp()

    def test_res_instances(self):
        ret = self.cst.res_instances
        ret_should = [('is this true ?'.split(),
                       'well i am curious whether is this true . because i do not like it . sentence three .'.split())]
        self.assertSequenceEqual(ret, ret_should)

    def test_masf(self):
        ret = self.cst._mask('this is query'.split(),
                             'this is doc'.split(),
                             2, 3,
                             0, 2)
        ret_should = ['this is .'.split(),
                      '. . doc'.split()]
        self.assertSequenceEqual(ret, ret_should)

    def test_masked_res_context(self):
        t = self.cst.sentences_as_context
        self.cst.sentences_as_context = False
        ret = self.cst.masked_res_context
        ret_should = [MaskedTupleContext(context=('. . . ?'.split(), 'well i am curious whether is this . . . i do not like it . sentence three .'.split()),
                                         mask_start_query=0,
                                         mask_end_query=3,
                                         mask_start_doc=7,
                                         mask_end_doc=10)]
        self.assertSequenceEqual(ret, ret_should)

        self.cst.sentences_as_context = True
        ret = self.cst.masked_res_context
        ret_should = [MaskedTupleContext(context=('. . . ?'.split(), 'well i am curious whether is this . . . i do not like it .'.split()),
                                         mask_start_query=0,
                                         mask_end_query=3,
                                         mask_start_doc=7,
                                         mask_end_doc=10)]

        self.cst.sentences_as_context = t

    def test_masked_res_context_preds(self):
        ret = self.cst.masked_res_context_preds()
        ret_should = do_prediction(self.model, self.model_config, self.tokenizer,
                                   docs_or_dataset=[
                                       'well i am curious whether is this . . . i do not like it . sentence three .'.split()],
                                   batch_size=32,
                                   queries_or_none=['. . . ?'.split()])[0]
        self.assertSequenceEqual(ret_should, ret)

    def test_neutral_masked_context(self):
        ret = self.cst.selected_masked_context
        ret_should = [MaskedTupleContext(context=('. . . ?'.split(), 'well i am curious whether is this . . . i do not like it . sentence three .'.split()),
                                         mask_start_query=0,
                                         mask_end_query=3,
                                         mask_start_doc=7,
                                         mask_end_doc=10)]
        self.assertSequenceEqual(ret, ret_should)

        self.cst.selected_masked_context = None
        self.cst.context_criterion = abs_diff
        ret = self.cst.selected_masked_context
        self.assertEqual(len(ret), 0)
