from argparse import Namespace
import os
from sre_parse import Tokenizer
import unittest
from typing import Dict

import torch
from candidate_sequence import CandidateSequence, CandidateSequenceTuple, MaskedContext
from causal_sequences import CausalSequenceTuples, CausalSequences, SyntheticInputs, SyntheticTupleInputs
from config import SequenceMiningConfig
from utils import do_prediction, get_model


class TestCausalSequences(unittest.TestCase):
    def setUp(self) -> None:
        self.device = torch.device('cpu')
        self.config = SequenceMiningConfig()
        self.config.MODEL_NAME = 'bert'
        self.config.DATASET_NAME = 'test'
        self.config.WORKING_DIR = f'{os.getenv("HOME")}/.sequence_mining/test_cs/bert/'
        self.sequence_fnames = f'{self.config.WORKING_DIR}/test1.tsv'
        self.causal_sequences = CausalSequences(self.sequence_fnames,
                                                self.config,
                                                device=self.device)
        self.working_dataset_fname = self.causal_sequences.working_dataset_fname

        self.model_config, self.tokenizer, self.model = get_model(self.config.MODEL_NAME,
                                                                  self.config.DATASET_NAME,
                                                                  self.device)
        self.neutrality_criterion = 'abs_diff'
        self.test1_seq0 = 'a lion'.split()
        self.test1_cand_seq0 = CandidateSequence(seq=self.test1_seq0,
                                                 dataset_name=self.config.DATASET_NAME,
                                                 working_dataset_fname=self.working_dataset_fname,
                                                 model=self.model,
                                                 model_config=self.model_config,
                                                 tokenizer=self.tokenizer,
                                                 sentences_as_context=False,
                                                 device=self.device,
                                                 context_criterion=self.neutrality_criterion,
                                                 batch_size=1)

        self.test1_seq1 = ', i'.split()
        self.test1_cand_seq1 = CandidateSequence(seq=self.test1_seq1,
                                                 dataset_name=self.config.DATASET_NAME,
                                                 working_dataset_fname=self.working_dataset_fname,
                                                 model=self.model,
                                                 model_config=self.model_config,
                                                 tokenizer=self.tokenizer,
                                                 sentences_as_context=False,
                                                 device=self.device,
                                                 context_criterion=self.neutrality_criterion,
                                                 batch_size=1)

    def test_candidate_sequences(self):

        def _test(ret_should, ret):
            self.assertEqual(len(ret_should), len(ret))
            for cs, cs_s in zip(ret, ret_should):
                self.assertSequenceEqual(cs_s.seq, cs.seq)
                for d, ds in zip(cs.res_docs, cs_s.res_docs):
                    self.assertSequenceEqual(d, ds)
                for m, ms in zip(cs.masked_res_context, cs_s.masked_res_context):
                    self.assertSequenceEqual(m, ms)
                for p, ps in zip(cs.masked_res_context_preds, cs_s.masked_res_context_preds):
                    self.assertAlmostEquals(p[0], ps[0], 5)
                    self.assertAlmostEquals(p[1], ps[1], 5)
                for n, ns in zip(cs.neutral_masked_context, cs_s.neutral_masked_context):
                    self.assertSequenceEqual(n, ns)

        (
            self.causal_sequences.model_config,
            self.causal_sequences.tokenizer,
            self.causal_sequences.model
        ) = get_model(self.config.MODEL_NAME, self.config.DATASET_NAME, self.device)

        ret = self.causal_sequences.candidate_sequences
        ret_should = (self.test1_cand_seq0, self.test1_cand_seq1)
        _test(ret_should, ret)

        self.sequence_fnames = [f'{os.getenv("HOME")}/.sequence_mining/test_cs/bert/test0.tsv',
                                f'{os.getenv("HOME")}/.sequence_mining/test_cs/bert/test1.tsv']
        self.causal_sequences = CausalSequences(self.sequence_fnames,
                                                self.config,
                                                device=self.device)
        test0_seq0 = 'lion'.split()
        test0_cand_seq0 = CandidateSequence(seq=test0_seq0,
                                            dataset_name=self.config.DATASET_NAME,
                                            working_dataset_fname=self.working_dataset_fname,
                                            model=self.model,
                                            model_config=self.model_config,
                                            tokenizer=self.tokenizer,
                                            sentences_as_context=False,
                                            device=self.device,
                                            context_criterion=self.neutrality_criterion,
                                            batch_size=1)

        test0_seq1 = ','.split()
        test0_cand_seq1 = CandidateSequence(seq=test0_seq1,
                                            dataset_name=self.config.DATASET_NAME,
                                            working_dataset_fname=self.working_dataset_fname,
                                            model=self.model,
                                            model_config=self.model_config,
                                            tokenizer=self.tokenizer,
                                            sentences_as_context=False,
                                            device=self.device,
                                            context_criterion=self.neutrality_criterion,
                                            batch_size=1)
        ret_should = (test0_cand_seq0, test0_cand_seq1,
                      self.test1_cand_seq0, self.test1_cand_seq1)
        ret = self.causal_sequences.candidate_sequences
        _test(ret_should, ret)

    def test_synthesize_input(self):
        s = 'black wolf'.split()
        t = MaskedContext(context='marry has a . . , little sheep'.split(),
                          mask_start=3,
                          mask_end=5)
        ret_should = 'marry has a black wolf , little sheep'
        ret = CausalSequences._synthesize(s, t)
        self.assertSequenceEqual(ret_should, ret)

    def test_synthesize_inputs(self):
        (
            self.causal_sequences.model_config,
            self.causal_sequences.tokenizer,
            self.causal_sequences.model
        ) = get_model(self.config.MODEL_NAME, self.config.DATASET_NAME, self.device)
        self.config.CONTEXT_CRITERION = 'all'
        self.causal_sequences = CausalSequences(self.sequence_fnames,
                                                self.config,
                                                device=self.device)
        ret = self.causal_sequences.synthesize_inputs(
            self.causal_sequences.candidate_sequences[0])
        ret_should = ['this is line one a lion have a lion',
                      'and this to be line two a lion am go to malaca']
        for r, rs in zip(ret.syn_inputs, ret_should):
            self.assertSequenceEqual(rs, r)

    def test_is_sequence_causal(self):
        (
            self.causal_sequences.model_config,
            self.causal_sequences.tokenizer,
            self.causal_sequences.model
        ) = get_model(self.config.MODEL_NAME, self.config.DATASET_NAME, self.device)
        syn_inputs = self.causal_sequences.synthesize_inputs(
            self.test1_cand_seq0)
        print(self.causal_sequences.is_causal(syn_inputs))

    def test_rules(self):
        (
            self.causal_sequences.model_config,
            self.causal_sequences.tokenizer,
            self.causal_sequences.model
        ) = get_model(self.config.MODEL_NAME, self.config.DATASET_NAME, self.device)
        self.causal_sequences.DUMP_EVERY = 1
        cs1 = self.causal_sequences.rules
        self.assertTrue(os.path.isfile(
            self.causal_sequences.syn_inputs_cache_fname))
        causal_sequences2 = CausalSequences(self.sequence_fnames,
                                            self.config,
                                            device=self.device)
        (
            causal_sequences2.model_config,
            causal_sequences2.tokenizer,
            causal_sequences2.model
        ) = get_model(self.config.MODEL_NAME, self.config.DATASET_NAME, self.device)
        causal_sequences2.syn_inputs_cache_fname = self.causal_sequences.syn_inputs_cache_fname
        causal_sequences2.output_dir = self.causal_sequences.output_dir
        cs2 = causal_sequences2.rules
        self.assertSequenceEqual(list(cs1.keys()), list(cs2.keys()))
        self.assertRulesEqual(cs1, cs2)

    def assertRulesEqual(self,
                         cs1: Dict[str, SyntheticInputs],
                         cs2: Dict[str, SyntheticInputs]):
        for key in cs1.keys():
            self.assertSequenceEqual(cs1[key].seq, cs2[key].seq)
            self.assertSequenceEqual(cs1[key].syn_inputs, cs2[key].syn_inputs)
            self.assertSequenceEqual(cs1[key].preds, cs2[key].preds)

    def test_save_results(self):
        output_file = self.causal_sequences.output_dir + '/rules.txt'
        try:
            os.remove(output_file)
        except FileNotFoundError:
            pass
        self.causal_sequences.write_rules()
        self.assertTrue(os.path.isfile(output_file))
        with open(output_file, 'r') as fin:
            for line in fin:
                if len(line) == 0:
                    pass
                self.assertIn(' '.join(line.split('->')[0].strip().split(', ')),
                              ['a lion', ', i'])

    def test_load_dump_cache(self):
        ret = self.causal_sequences.maybe_load_syn_inputs_cache()
        self.assertDictEqual(ret, dict())

        fname = self.causal_sequences.syn_inputs_cache_fname
        try:
            os.remove(fname)
        except FileNotFoundError:
            pass
        test_rule = self.causal_sequences.synthesize_inputs(
            self.test1_cand_seq0)  # a lion
        ret_should = {' '.join(self.test1_cand_seq0.seq): test_rule}
        self.causal_sequences.rules = ret_should
        self.causal_sequences.dump_syn_inputs_cache()

        self.assertTrue(os.path.isfile(fname))

        ret = self.causal_sequences.maybe_load_syn_inputs_cache()
        self.assertSequenceEqual(
            [' '.join(self.test1_cand_seq0.seq)], list(ret.keys()))
        self.assertRulesEqual(ret_should, ret)


class TestSyntheticTupleInputs(unittest.TestCase):
    def setUp(self) -> None:
        self.device = 'cpu'
        self.model_config, self.tokenizer, self.model = get_model(
            'bert', 'fever', self.device)
        self.sti = SyntheticTupleInputs(qd_tuple=('the query'.split(),
                                                  'the document'.split()),
                                        syn_queries=[
                                            'this is the query'.split()],
                                        syn_docs=[
                                            'this is the document'.split()],
                                        model=self.model,
                                        model_config=self.model_config,
                                        tokenizer=self.tokenizer,
                                        batch_size=32,
                                        preds=None)
        return super().setUp()

    def test_preds(self):
        ret = self.sti.preds
        ret_should = do_prediction(self.model, self.model_config, self.tokenizer,
                                   docs_or_dataset=[
                                       'this is the document'.split()],
                                   batch_size=32,
                                   queries_or_none=['this is the query'.split()])[0]
        self.assertSequenceEqual(ret, ret_should)


class TestCausalSequenceTuples(unittest.TestCase):
    def setUp(self) -> None:
        args = Namespace(dataset_name='test_fever', model_name='bert')
        self.conf = SequenceMiningConfig(args)
        working_dir = '/home/zzhang/.sequence_mining/test/bert/'
        self.mock_mined_tuple_fname = working_dir + 'mined_results.csv'
        mock_mined_tuples = [['query_sequence', 'doc_sequence', 'cluster_label',
                              'tuple_support_in_REFUTES', 'tuple_support_in_SUPPORTS',
                              'tuple_npmi_REFUTES', 'tuple_npmi_SUPPORTS',
                              'model_pred_REFUTES', 'model_pred_SUPPORTS'],
                             ['the query', 'the doc', 0,
                              31, 80,
                              -0.001, 0.0005,
                              0.001, 0.999],
                             ['that query', 'this doc . and', 0,
                              31, 80,
                              -0.001, 0.0005,
                              0.001, 0.999],
                             ['no query', 'no doc', 0,
                              31, 80,
                              0.001, -0.0005,
                              0.001, 0.999]]
        with open(self.mock_mined_tuple_fname, 'w+') as fout:
            for line in mock_mined_tuples:
                fout.write('\t'.join(map(str, line)) + '\n')
        mock_sequences = ['this is the query . [SEP] and this is the doc .',
                          'there is no query . [SEP] neither is no doc .',
                          'i like that query . [SEP] it corresponds to this doc . and here is the second sentence . and i am the third .']
        self.mock_sequences_fname = working_dir + 'sequences.txt'
        with open(self.mock_sequences_fname, 'w+') as fout:
            for line in mock_sequences:
                fout.write(line + '\n')
        self.model_config, self.tokenizer, self.model = get_model(
            'bert', 'fever', 'cpu')

        self.mock_candidate_sequence_tuples = []
        queries = ['the query'.split(), 'that query'.split()]
        docs = ['the doc'.split(), 'this doc . and'.split()]
        for q, d in zip(queries, docs):
            self.mock_candidate_sequence_tuples.append(CandidateSequenceTuple(qd_tuple=(q, d)))
            self.mock_candidate_sequence_tuples[-1].init_functions(
                    dataset_name='test',
                    working_dataset_fname=self.mock_sequences_fname,
                    model=self.model,
                    model_config=self.model_config,
                    tokenizer=self.tokenizer,
                    sentences_as_context=False,
                    device=self.conf.DEVICE,
                    context_criterion=self.conf.CONTEXT_CRITERION,
                    batch_size=32)
        return super().setUp()

    def test_candidate_tuples(self):
        self.cst = CausalSequenceTuples(
            input_fnames=[self.mock_mined_tuple_fname], conf=self.conf)
        ret = self.cst.candidate_tuples
        ret_should = self.mock_candidate_sequence_tuples
        self.assertSequenceEqual(ret, ret_should)

    def test_synthesize_inputs(self):
        self.cst = CausalSequenceTuples(
            input_fnames=[self.mock_mined_tuple_fname], conf=self.conf)
        s = self.mock_candidate_sequence_tuples[0]
        self.cst.sentences_as_context = False
        ret = self.cst._get_synthetic_instances(s)
        ret_should = SyntheticTupleInputs(
            qd_tuple=s.qd_tuple,
            syn_queries=[
                'i like the query .'.split()],
            syn_docs=[
                'it corresponds to the doc here is the second sentence . and i am the third .'.split()],
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            batch_size=32,
            preds=None)
        self.assertEqual(ret, ret_should)

        # test cache
        self.cst.candidate_tuples = None
        ret = self.cst._get_synthetic_instances(s)
        self.assertEqual(ret, ret_should)

        self.cst.invalidate_candidate_cache('candidate_tuples')

        self.cst.sentences_as_context = True
        self.cst.candidate_tuples = None
        ret = self.cst._get_synthetic_instances(s)
        ret_should = SyntheticTupleInputs(
            qd_tuple=s.qd_tuple,
            syn_queries=[
                'i like the query .'.split()],
            syn_docs=[
                'it corresponds to the doc here is the second sentence .'.split()],
            model=self.model,
            model_config=self.model_config,
            tokenizer=self.tokenizer,
            batch_size=32,
            preds=None)
        self.assertEqual(ret, ret_should)

    def test_rules(self):
        self.cst = CausalSequenceTuples(
            input_fnames=[self.mock_mined_tuple_fname], conf=self.conf)
        self.cst.sentences_as_context = True
        self.maxDiff = None
        ret = self.cst.rules
        ret_should = {
            'the query [SEP] the doc':
            SyntheticTupleInputs(
                qd_tuple=self.mock_candidate_sequence_tuples[0].qd_tuple,
                syn_queries=['i like the query .'.split()],
                syn_docs=[
                    'it corresponds to the doc here is the second sentence .'.split()],
                model=self.model,
                model_config=self.model_config,
                tokenizer=self.tokenizer,
                batch_size=32,
                preds=None),
            # 'that query [SEP] this doc . and':  # this entry fails the causality test
            # SyntheticTupleInputs(
            #     qd_tuple=self.mock_candidate_sequence_tuples[0].qd_tuple,
            #     syn_queries=['this is that query .'.split()],
            #     syn_docs=['and this is this doc . and .'.split()],
            #     model=self.model,
            #     model_config=self.model_config,
            #     tokenizer=self.tokenizer,
            #     batch_size=32,
            #     preds=None)
        }
        self.assertDictEqual(ret, ret_should)


if __name__ == '__main__':
    unittest.main()
