from argparse import Namespace
import os.path
import unittest
from math import log
from unittest import TestCase

from config import SequenceMiningConfig
from miners.basic_miner import BasicMiner
from miners.double_sequence_miner import DoubleSequenceMiner
from miners.mono_sequence_miner import MonoSequenceMiner
from predictor import Predictor


class TestWithPredefinedData(TestCase):
    conf, miner = None, None

    def load_data(self):
        predictor = Predictor(self.conf)
        predictor.decoded_cls_preds = 'ababab'
        predictor.cluster_preds = [0, 0, 0, 0, 0, 0]
        predictor.instances = {'a': {0: ['that is a test'.split(),
                                         'what a test'.split(),
                                         'a testify is here'.split()]},
                               'b': {0: ['this to be a test , too'.split(),
                                         'and i am also one test'.split(),
                                         'leide kein test hier'.split()]}}
        predictor.dump()

    def remove_old_files(self):
        for c in 'ab':
            cluster = 0
            try:
                os.remove(
                    f'{self.miner.working_dataset_prefix}_{c}_{cluster}.txt')
            except FileNotFoundError:
                pass
        try:
            os.remove(f'{self.miner.working_dataset_fname}')
        except FileNotFoundError:
            pass

        try:
            os.remove(self.miner.output_fname)
        except FileNotFoundError:
            pass

    def remove_cache(self):
        try:
            os.remove(self.miner.mining_cache_fname)
        except FileNotFoundError:
            pass


class TestBasicMiner(TestWithPredefinedData):
    def setUp(self):
        args = Namespace(dataset_name='test')
        self.conf = SequenceMiningConfig(args)
        self.conf.CLASS_NAMES = 'a', 'b'
        self.conf.SEQ_MINING_MIN_SUPPORT = '3'
        self.miner = BasicMiner(self.conf)
        self.miner.patterns = ['(..)']
        self.remove_old_files()
        self.load_data()

    def test_convert_interval_to_patterns(self):
        ret = BasicMiner.convert_interval_to_patterns((3, 5))
        ret_should = ['(...)', '(....)', '(.....)']
        self.assertSequenceEqual(ret_should, ret)

    def test_frequent_object(self):
        t = self.miner.patterns
        self.remove_cache()
        self.miner.patterns = ['(..)',]
        ret = self.miner.frequent_objects
        ret_should = {'a test': 3}
        self.assertDictEqual(ret_should, ret)

        self.miner.patterns = ['(..)',  "('a')"]
        ret = self.miner.frequent_objects
        ret_should = {'a test': 3, 'a': 4}
        self.assertDictEqual(ret_should, ret)
        self.miner.patterns = t

    def test_count_docs_given_class(self):
        ret = self.miner._count_docs_given_class('a', 0)
        self.assertEqual(ret, 3)

    def test_n_docs_given_class(self):
        ret = self.miner.n_docs_given_class
        ret_should = {'a': {0: 3}, 'b': {0: 3}}
        self.assertDictEqual(ret_should, ret)

    def test_n_docs_all_class(self):
        ret = self.miner.n_docs_all_classes
        self.assertEqual(ret, {0: 6})


class TestMonoSequenceMiner(TestWithPredefinedData):
    def setUp(self):
        args = Namespace(dataset_name='test')
        self.conf = SequenceMiningConfig(args)
        self.conf.DOC_SEQ_LEN_INTERVAL = (2, 2)
        self.conf.SEQ_MINING_MIN_SUPPORT = '3'
        self.conf.CLASS_NAMES = ('a', 'b')
        self.miner = MonoSequenceMiner(self.conf)
        self.remove_old_files()
        self.load_data()

    def test_patterns(self):
        t = self.miner.doc_seq_len_interval
        self.miner.doc_seq_len_interval = (2,4)
        self.assertSequenceEqual(['(..)', '(...)', '(....)'], self.miner.patterns)
        self.miner.doc_seq_len_interval = t

    def test_get_dc_given_class(self):
        ret = self.miner._get_dc_given_class('a', 0, 'a test')
        self.assertEqual(2, ret)

    def test_preds(self):
        self.remove_cache()
        ret = self.miner.preds
        self.assertEqual(len(ret), 1)
        self.assertSequenceEqual(list(ret.keys()), ['a test'])
        self.assertIsInstance(ret['a test'][0], float)
        self.assertIsInstance(ret['a test'][1], float)

    def test_build_header(self):
        ret = self.miner.build_header()
        ret_should = ['sequence', 'cluster_label',
                      'dc_a', 'dc_b',
                      'npmi_a', 'npmi_b',
                      'model_pred_a', 'model_pred_b']
        self.assertSequenceEqual(ret_should, ret)

    def test_df_given_class(self):
        ret = self.miner.df_given_class
        ret_should = {'a': {0: {'a test': 2 / 3}},
                      'b': {0: {'a test': 1 / 3}}}
        self.assertDictEqual(ret_should, ret)

    def test_df_all_classes(self):
        ret = self.miner.df_all_classes
        ret_should = {0: {'a test': 1 / 2}}
        self.assertDictEqual(ret_should, ret)

    def test_npmi_between(self):
        ret = self.miner._npmi_between('a', 'a test')
        ret_should = log((2 / 3) / (1 / 2)) / log(2 / 6)
        self.assertEqual(ret, ret_should)

    def test_npmi(self):
        ret = self.miner.npmi
        ret_should = {'a': {0: {'a test': self.miner._npmi_between('a', 'a test')}},
                      'b': {0: {'a test': self.miner._npmi_between('b', 'a test')}}}
        self.assertDictEqual(ret_should, ret)

    def test_dc_given_class(self):
        ret = self.miner.dc_given_class
        ret_should = {'a': {0: {'a test': 2}},
                      'b': {0: {'a test': 1}}}
        self.assertDictEqual(ret_should, ret)

    def test_results(self):
        ret = self.miner.results
        ret_should = ['a test', 0,
                      2, 1,
                      self.miner._npmi_between('a', 'a test'),
                      self.miner._npmi_between('b', 'a test')]
        ret_should = list(map(str, ret_should))
        self.assertSequenceEqual(ret[0][:-2], ret_should)

    def test_save_results(self):
        self.miner.save_results()
        self.assertTrue(os.path.isfile(self.miner.output_fname))
        with open(self.miner.output_fname, 'r') as fin:
            ret1 = list(fin.readlines())

        os.remove(self.miner.output_fname)
        self.miner.save_results()
        self.assertTrue(os.path.isfile(self.miner.output_fname))
        with open(self.miner.output_fname, 'r') as fin:
            ret2 = list(fin.readlines())
        self.assertSequenceEqual(ret1, ret2)


class TestDoubleSequenceMiner(TestWithPredefinedData):
    def load_data(self):
        predictor = Predictor(self.conf)
        predictor.decoded_cls_preds = 'ababab'
        predictor.cluster_preds = [0, 0, 0, 0, 0, 0]
        predictor.instances = {'a': {0: [('the query 1'.split(), 'that is a test'.split()),
                                         ('the query 3'.split(), 'what a test'.split()),
                                         ('well query 5'.split(), 'a testify is here'.split())]},
                               'b': {0: [('the query 2'.split(), 'this to be a test , too'.split()),
                                         ('no query 4'.split(), 'and i am also one test'.split()),
                                         ('dies ist query 6'.split(), 'leide kein test hier'.split())]}}
        predictor.dump_queries = True
        predictor.dump()

    def setUp(self) -> None:
        args = Namespace(dataset_name='test')
        self.conf = SequenceMiningConfig(args)
        self.conf.CLASS_NAMES = ('a', 'b')
        self.miner = DoubleSequenceMiner(self.conf)
        self.remove_old_files()
        self.load_data()

    def test_patterns(self):
        td = self.miner.doc_seq_len_interval
        tq = self.miner.query_seq_len_interval
        self.miner.doc_seq_len_interval = (2, 4)
        self.miner.query_seq_len_interval = (2, 4)
        ret = self.miner.patterns
        self.assertEqual(ret, ["(..).*('[SEP]').*(..)",
                               "(..).*('[SEP]').*(...)",
                               "(..).*('[SEP]').*(....)",
                               "(...).*('[SEP]').*(..)",
                               "(...).*('[SEP]').*(...)",
                               "(...).*('[SEP]').*(....)",
                               "(....).*('[SEP]').*(..)",
                               "(....).*('[SEP]').*(...)",
                               "(....).*('[SEP]').*(....)"])
        self.miner.doc_seq_len_interval = td
        self.miner.query_seq_len_interval = tq

    def test_frequent_objects(self):
        self.remove_cache()
        ret = self.miner.frequent_objects
        ret_should = {('the query', 'a test'): 3}
        self.assertDictEqual(ret_should, ret)

    def test_get_dc_given_class(self):
        ret = self.miner._get_dc_given_class('a', 0, ('the query', 'a test'))
        self.assertEqual(ret, 2)

    def test_dc_given_class(self):
        ret = self.miner.dc_given_class
        ret_should = {'a': {0: {('the query', 'a test'): 2}},
                      'b': {0: {('the query', 'a test'): 1}}}
        self.assertDictEqual(ret_should, ret)

    def test_query_sequences(self):
        ret = self.miner.query_sequences
        self.assertSequenceEqual(['the query'], ret)

    def test_doc_sequences(self):
        ret = self.miner.doc_sequences
        self.assertSequenceEqual(['a test'], ret)

    def test_padded(self):
        ret = self.miner.padded('this test')
        self.assertEqual('. . this test . .', ret)

    def test_preds(self):
        ret = self.miner.preds
        self.assertEqual(len(ret), 1)
        self.assertSequenceEqual(list(ret.keys()), [('the query', 'a test')])
        self.assertIsInstance(ret[('the query', 'a test')][0], float)
        self.assertIsInstance(ret[('the query', 'a test')][1], float)

    def test_build_header(self):
        ret = self.miner.build_header()
        ret_should = ['query_sequence', 'doc_sequence',
                      'cluster_label',
                      'tuple_support_in_a', 'tuple_support_in_b',
                      'tuple_npmi_a', 'tuple_npmi_b',
                      'model_pred_a', 'model_pred_b']
        self.assertSequenceEqual(ret, ret_should)

    def test_df_given_class(self):
        ret = self.miner.df_given_class
        ret_should = {'a': {0: {('the query', 'a test'): 2 / 3}},
                      'b': {0: {('the query', 'a test'): 1 / 3}}}
        self.assertDictEqual(ret_should, ret)

    def test_df_all_classes(self):
        ret = self.miner.df_all_classes
        ret_should = {0: {('the query', 'a test'): 1 / 2}}
        self.assertDictEqual(ret_should, ret)

    def test_npmi_between(self):
        ret = self.miner._npmi_between('a', ('the query', 'a test'))
        ret_should = log((2 / 3) / (1 / 2)) / log(2 / 6)
        self.assertEqual(ret, ret_should)

    def test_npmi(self):
        ret = self.miner.npmi
        ret_should = {'a': {0: {('the query', 'a test'): self.miner._npmi_between('a', ('the query', 'a test'))}},
                      'b': {0: {('the query', 'a test'): self.miner._npmi_between('b', ('the query', 'a test'))}}}
        self.assertDictEqual(ret_should, ret)

    def test_results(self):
        ret = self.miner.results
        ret_should = ['the query', 'a test',
                      0,
                      2, 1,
                      self.miner._npmi_between('a', ('the query', 'a test')),
                      self.miner._npmi_between('b', ('the query', 'a test'))]
        ret_should = list(map(str, ret_should))
        self.assertSequenceEqual(ret[0][:-2], ret_should)

    def test_save_results(self):
        self.miner.save_results()
        self.assertTrue(os.path.isfile(self.miner.output_fname))
        with open(self.miner.output_fname, 'r') as fin:
            ret1 = list(fin.readlines())

        os.remove(self.miner.output_fname)
        self.miner.save_results()
        self.assertTrue(os.path.isfile(self.miner.output_fname))
        with open(self.miner.output_fname, 'r') as fin:
            ret2 = list(fin.readlines())
        self.assertSequenceEqual(ret1, ret2)


if __name__ == '__main__':
    unittest.main()
