from argparse import Namespace
import os
import unittest

import numpy as np
from config import SequenceMiningConfig
from predictor import Predictor
from utils import get_model


class TestPredictor(unittest.TestCase):
    def setUp(self) -> None:
        args = Namespace(dataset_name='test')
        self.conf = SequenceMiningConfig(args)
        self.predictor = Predictor(self.conf)
        (
            self.predictor.model_config,
            self.predictor.tokenizer,
            self.predictor.model
        ) = get_model(self.predictor.model_name,
                      self.predictor.dataset_name,
                      self.predictor.device)
        # self.predictor.preprocess()

    def test_clustering(self):
        self.predictor.n_clusters = 2
        X = np.array([[1, 2, 3], [4, 4, 5], [99, 99, 99]])
        ret = self.predictor.clustering(X)
        ret = ret.tolist()
        if ret[0] == 0:
            self.assertSequenceEqual(ret, [0, 0, 1])
        else:
            self.assertSequenceEqual(ret, [1, 1, 0])

    def test_dump_and_load_preds(self):
        ret = self.predictor.maybe_predict()
        try:
            os.remove(self.predictor.cache_fname)
        except FileNotFoundError:
            pass
        self.predictor.cache_preds(*ret)
        self.assertTrue(os.path.isfile(self.predictor.cache_fname))
        t = self.predictor.cache_fname
        self.predictor.cache_fname = 'no_such_file'
        with self.assertRaises(FileNotFoundError):
            self.predictor.load_preds()
        self.predictor.cache_fname = t
        ret2 = self.predictor.load_preds()
        self.assertSequenceEqual(ret[0], ret2[0])
        np.testing.assert_array_almost_equal(ret[1], ret2[1])
        self.assertSequenceEqual(ret[2], ret2[2])

    def test_maybe_predict(self):
        try:
            os.remove(self.predictor.cache_fname)
        except FileNotFoundError:
            pass
        ret1 = self.predictor.maybe_predict()
        self.assertTrue(os.path.isfile(self.predictor.cache_fname))
        ret2 = self.predictor.maybe_predict()
        self.assertSequenceEqual(ret1[0], ret2[0])
        for s1, s2 in zip(ret1[1], ret2[1]):
            np.testing.assert_array_almost_equal(s1, s2)
        for p1, p2 in zip(ret1[2], ret2[2]):
            self.assertSequenceEqual(p1, p2)

    def test_convert_to_dict(self):
        self.predictor.decoded_cls_preds = '010'
        self.predictor.cluster_preds = [0, 0, 1]
        inputs = [f'this is the doc number {i}' for i in range(3)]
        ret = self.predictor.convert_to_dict(inputs)
        print(ret)
        ret_should = {'0': {0: ['this is the doc number 0'],
                            1: ['this is the doc number 2']},
                      '1': {0: ['this is the doc number 1'],
                            1: []}}
        self.assertDictEqual(ret_should, ret)

    def test_preprocess(self):
        self.predictor.preprocess()

    def remove_dump_files(self):
        for cluster_to_dump in (0, 1):
            for class_to_dump in 'ab':
                fname = f'{self.predictor.working_dataset_prefix}_' \
                        f'{class_to_dump}_' \
                        f'{cluster_to_dump}.txt'
                try:
                    os.remove(fname)
                except FileNotFoundError:
                    pass
        return class_to_dump, cluster_to_dump

    def test_dump_instance_for(self):
        self.predictor.decoded_cls_preds = 'abab'
        self.predictor.cluster_preds = [0, 0, 1, 1]
        self.predictor.instances = {'a': {0: ['doc number 0'.split()],
                                          1: ['doc number 2'.split()]},
                                    'b': {0: ['doc number 1'.split()],
                                          1: ['doc number 3'.split()]}}
        self.remove_dump_files()
        i = 0
        for cluster_to_dump in (0, 1):
            for class_to_dump in 'ab':
                self.predictor.dump_instances_for(
                    class_to_dump, cluster_to_dump)
                fname = f'{self.predictor.working_dataset_prefix}_' \
                        f'{class_to_dump}_' \
                        f'{cluster_to_dump}.txt'
                self.assertTrue(os.path.isfile(fname))
                with open(fname, 'r') as fin:
                    ret = fin.readline()
                    self.assertSequenceEqual(f'doc number {i}\n', ret)
                i += 1

        self.remove_dump_files()
        self.predictor.decoded_cls_preds = 'abab'
        self.predictor.cluster_preds = [0, 0, 1, 1]
        self.predictor.instances = {'a': {0: [('query number 0'.split(), 'doc number 0'.split())],
                                          1: [('query number 2'.split(), 'doc number 2'.split())]},
                                    'b': {0: [('query number 1'.split(), 'doc number 1'.split())],
                                          1: [('query number 3'.split(), 'doc number 3'.split())]}}
        self.predictor.dump_queries = True
        self.predictor.dump_instances_for(class_to_dump, cluster_to_dump)
        i = 0
        for cluster_to_dump in (0, 1):
            for class_to_dump in 'ab':
                self.predictor.dump_instances_for(
                    class_to_dump, cluster_to_dump)
                fname = f'{self.predictor.working_dataset_prefix}_' \
                        f'{class_to_dump}_' \
                        f'{cluster_to_dump}.txt'
                self.assertTrue(os.path.isfile(fname))
                with open(fname, 'r') as fin:
                    ret = fin.readline()
                    self.assertSequenceEqual(
                        f'query number {i} [SEP] doc number {i}\n', ret)
                i += 1
        self.predictor.dump_queries = False

    def test_dump_sequence_for_all(self):
        fname = self.predictor.working_dataset_prefix + '.txt'
        try:
            os.remove(fname)
        except FileNotFoundError:
            pass
        self.predictor.decoded_cls_preds = '0101'
        self.predictor.cluster_preds = [0, 0, 1, 1]
        self.predictor.instances = {'0': {0: ['doc number 0'.split()],
                                          1: ['doc number 2'.split()]},
                                    '1': {0: ['doc number 1'.split()],
                                          1: ['doc number 3'.split()]}}
        self.predictor.dump_all_instances()
        with open(fname, 'r') as fin:
            for idx, line in enumerate(fin):
                self.assertSequenceEqual(f'doc number {idx}\n', line)

    def dump(self):
        self.predictor.decoded_cls_preds = 'ababab'
        self.predictor.cluster_labels = [0, 0, 0, 1, 1, 1]
        self.predictor.docs = [f'doc number {i}' for i in range(6)]
        self.predictor.dump()


if __name__ == '__main__':
    unittest.main()
