import unittest

import torch.cuda
from expred import BertClassifier, BertTokenizerWithSpans, Expred
from transformers import BertTokenizer
from utils import (_do_prediction, abs_diff, convert_to_word_split_strings,
                   do_prediction, get_bert_model, get_expred_model, get_model,
                   join_sentences_to_docs, locate_substr, mean, mean_criterion,
                   mean_stddev_criterion, sentence_seek,
                   split_doc_to_sentences, stddev, stddev_criterion)


class TestUtils(unittest.TestCase):
    def setUp(self) -> None:
        self.dataset_name = 'test'
        self.device = torch.device('cpu')
        self.docs = ['this movie is bad .'.split(), 'i like this movie .'.split()]
        self.sentences_docs = [[d] for d in self.docs]
        self.query = 'what is the sentiment of this review ?'.split()
        self.data = [(sum(sum(hash(w) for w in s) for s in doc),
                      self.query, doc, 'NEG') for doc in self.sentences_docs]

    def test_convert_to_word_split_strings(self):
        ret_should = ['this is doc 1'.split(),
                      'this is doc 2'.split()]
        _inputs = ['this is doc 1',
                   'this is doc 2']  # doc(query)
        ret = convert_to_word_split_strings(_inputs)
        for r, rs in zip(ret, ret_should):
            self.assertSequenceEqual(r, rs)

        _inupts = ['this is doc 1'.split(),
                   'this is doc 2'.split()]  # doc(query)-token
        ret = convert_to_word_split_strings(_inputs)
        for r, rs in zip(ret, ret_should):
            self.assertSequenceEqual(r, rs)

        _inupts = [['this is doc 1'.split()],
                   ['this is doc 2'.split()]]  # doc-sentence-token
        ret = convert_to_word_split_strings(_inputs)
        for r, rs in zip(ret, ret_should):
            self.assertSequenceEqual(r, rs)

    def test_get_expred_model(self):
        config, tokenizer, model = get_expred_model(self.dataset_name, self.device)
        print(config)
        self.assertTrue(isinstance(tokenizer, BertTokenizer))
        self.assertTrue(isinstance(tokenizer, BertTokenizerWithSpans))
        self.assertTrue(isinstance(model, Expred))
        self.assertEqual(next(model.parameters()).device, self.device)

    def test_get_bert_model(self):
        config, tokenizer, model = get_bert_model(self.dataset_name, self.device)
        print(config)
        self.assertTrue(isinstance(tokenizer, BertTokenizer))
        self.assertTrue(isinstance(model, BertClassifier))
        self.assertEqual(next(model.parameters()).device, self.device)

    def test_get_model(self):
        config, tokenizer, model = get_model('bert', self.dataset_name, self.device)
        print(config)
        self.assertTrue(isinstance(tokenizer, BertTokenizer))
        self.assertTrue(isinstance(model, BertClassifier))
        self.assertEqual(next(model.parameters()).device, self.device)
        config, tokenizer, model = get_model('expred', self.dataset_name, self.device)
        print(config)
        self.assertTrue(isinstance(tokenizer, BertTokenizerWithSpans))
        self.assertTrue(isinstance(model, Expred))
        self.assertEqual(next(model.parameters()).device, self.device)
        with self.assertRaises(NotImplementedError):
            _ = get_model('foo', 'bar', 'foo')

    def test_do_prediction(self):
        for model_name in ['bert', 'expred']:
            config, tokenizer, model = get_model(model_name, self.dataset_name, self.device)
            model_outputs = _do_prediction(model, config, tokenizer, self.data)
            cls_preds, decoded_cls_preds, doc_embeddings, decoded_docs = model_outputs
            self.assertEqual(doc_embeddings.shape, (2, 768))
            self.assertSequenceEqual(decoded_docs, self.docs)
            for p in decoded_cls_preds:
                self.assertIn(p, ['NEG', 'POS'])
            self.assertEqual(len(cls_preds), 2)
            for p in cls_preds:
                self.assertEqual(len(p), 2)

    def test_predict_docs_sentiment(self):
        for model_name in ['bert', 'expred']:
            config, tokenizer, model = get_model(model_name, self.dataset_name, self.device)
            cls_preds1 = do_prediction(model, config, tokenizer, self.sentences_docs, 2)[0]

            model.eval()
            model_outputs = _do_prediction(model, config, tokenizer, self.data)
            cls_preds2, _, _, _ = model_outputs

            self.assertSequenceEqual(cls_preds1, cls_preds2)

    def test_sentence_seek(self):
        sentences = ['this is sentence one .'.split(),
                     'this is sentence two .'.split()]
        ret = sentence_seek(sentences, 2)
        ret_should = (0, 0)
        self.assertSequenceEqual(ret_should, ret)

        ret = sentence_seek(sentences, 6)
        ret_should = (5, 1)
        self.assertSequenceEqual(ret_should, ret)

    def test_split_doc_to_sentences(self):
        doc = 'this is a doc . it has multiple sentences . .'.split()
        ret = split_doc_to_sentences(doc)
        ret_should = ['this is a doc .'.split(),
                      'it has multiple sentences .'.split(),
                      ['.']]
        self.assertSequenceEqual(ret_should, ret)

    def test_join_sentences_to_docs(self):
        sentences = ['this is a doc .'.split(),
                     'it has multiple sentences .'.split(),
                     ['.']]
        ret = join_sentences_to_docs(sentences)
        ret_should = 'this is a doc . it has multiple sentences . .'.split()
        self.assertSequenceEqual(ret_should, ret)

    def test_locate(self):
        string_list = 'marry has a little sheep , little sheep'.split()
        pattern = 'little sheep'.split()
        start, end = locate_substr(string_list, pattern)
        self.assertEqual(3, start)
        self.assertEqual(5, end)

    def test_stddev(self):
        # _input = [[1, 2], [2, 3], [3, 4], [4, 5], [5, 6]]
        _input = [1, 2, 3, 4, 5]
        ret_should = 2 ** 0.5
        ret = stddev(_input)
        self.assertEqual(ret_should, ret)

    def test_mean(self):
        _input = [1, 2, 3, 4, 5]
        ret_should = 3
        self.assertEqual(ret_should, mean(_input))

    def test_stddev_criterion(self):
        _input = [[i / 10.] for i in range(5)]
        self.assertFalse(stddev_criterion(_input))
        _input = [[i / 100.] for i in range(5)]
        self.assertTrue(stddev_criterion(_input))

    def test_mean_criterion(self):
        _input = [[(i + 6) / 10.] for i in range(5)]
        self.assertTrue(mean_criterion(_input))
        _input = [[i / 100. + 0.7] for i in range(5)]
        self.assertTrue(mean_criterion(_input))
        _input = [[i / 100. + 0.5] for i in range(5)]
        self.assertFalse(mean_criterion(_input))

    def test_mean_stddev_criterion(self):
        _input = [[(i + 6) / 10.] for i in range(5)]
        self.assertFalse(mean_stddev_criterion(_input))
        _input = [[i / 100. + 0.7] for i in range(5)]
        self.assertTrue(mean_stddev_criterion(_input))
        _input = [[i / 100. + 0.5] for i in range(5)]
        self.assertFalse(mean_stddev_criterion(_input))

    def test_abs_diff(self):
        self.assertFalse(abs_diff((0.39, 0.61)))
        self.assertTrue(abs_diff((0.41, 0.59)))


if __name__ == '__main__':
    unittest.main()
