import logging
import random
import re
import subprocess
from argparse import Namespace
from itertools import chain
from typing import List, Literal, Tuple, Union

import numpy as np
import torch
import tqdm
from expred import (BertClassifier, BertConfig, BertDataset, BertInput,
                    BertTokenizerWithSpans, CLSModel, Expred, ExpredConfig,
                    ExpredInput)
from sentence_transformers import InputExample
from torch import device
from transformers import BasicTokenizer, BertTokenizer

from config import *
from logging_utils import get_logger
from models.inputs import LSTMBatch
from models.lstm.model import BiLSTMModel
from models.model_config import LSTMConfig, SBERTConfig
from models.sbert.model import SBERT

EPSILON = 1e-8
INFINITY = 1e16

logger = get_logger('utils', logging.INFO)
if DEBUG:
    logger.setLevel(logging.DEBUG)
    

basic_tokenizer = BasicTokenizer()


def get_args() -> Namespace:
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset_name', dest='dataset_name', default='movies')
    parser.add_argument('--model_name', dest='model_name', default='bert')
    parser.add_argument('--d_interval', required=False, default=None, dest='d_interval')
    parser.add_argument('--q_interval', required=False, default=None, dest='q_interval')
    parser.add_argument('--decoy_id', required=False, type=int, default=None, dest='decoy_id')
    parser.add_argument('--focus', default=False, dest='focus', action='store_true')
    parser.add_argument('--rq3', default=False, dest='rq3', action='store_true')
    parser.add_argument('--support', default=None, dest='support', type=str)
    parser.add_argument('--force_recalculate_rules', default=False, dest='force_recalculate_rules', action='store_true')
    parser.add_argument('--explainer', required=False, default='lime', dest='explainer', choices=['lime', 'anchor', 'shap', 'expred'])
    args = parser.parse_args()
    if getattr(args, 'd_interval') is not None:
        args.d_interval = tuple(map(int, args.d_interval.split('-')))
    if getattr(args, 'q_interval') is not None:
        args.q_interval = tuple(map(int, args.q_interval.split('-')))
    return args

def get_expred_model(pretrained_dataset_name, base_dataset_name=None, device=None) -> Tuple[ExpredConfig, BertTokenizerWithSpans, Expred]:
    if base_dataset_name is None:
        if 'evidence_inference' in pretrained_dataset_name:
            base_dataset_name = 'evidence_inference'
        elif '_' in pretrained_dataset_name:
            base_dataset_name = pretrained_dataset_name.split('_')[0]
        else:
            base_dataset_name = pretrained_dataset_name
    config = ExpredConfig(
            pretrained_dataset_name=pretrained_dataset_name,
            base_dataset_name=base_dataset_name,
            device=device,
            load_from_pretrained=True
        )
    model = Expred.from_pretrained(config).to(device)
    tokenizer = BertTokenizerWithSpans.from_pretrained('bert-base-uncased')
    model.eval()
    return config, tokenizer, model


def get_bert_model(pretrained_dataset_name, base_dataset_name=None, device=None) -> Tuple[BertConfig, BertTokenizer, BertClassifier]:
    if base_dataset_name is None:
        if 'evidence_inference' in pretrained_dataset_name:
            base_dataset_name = 'evidence_inference'
        elif '_' in pretrained_dataset_name:
            base_dataset_name = pretrained_dataset_name.split('_')[0]
        else:
            base_dataset_name = pretrained_dataset_name
    config = BertConfig(pretrained_dataset_name, base_dataset_name, True)
    tokenizer = BertTokenizerWithSpans.from_pretrained('bert-base-uncased')
    if device is None:
        if torch.cuda.is_available():
            device = torch.device('cuda')
        else:
            device = torch.device('cpu')
    model = BertClassifier.from_pretrained(config).to(device)
    with open(SequenceMiningConfig.get_bert_model_fname(pretrained_dataset_name), 'rb') as fin:
        model.load_state_dict(torch.load(fin))
        logger = get_logger('utils')
        logger.info('best model loaded')
    model.eval()
    return config, tokenizer, model


def get_sbert_model(pretrained_dataset_name, base_dataset_name=None, device=None):
    config = SBERTConfig(
        dataset_name=pretrained_dataset_name,
        max_len=SequenceMiningConfig.MAXLEN,
        class_names=SequenceMiningConfig.get_class_names(base_dataset_name)
    )
    tokenizer = BertTokenizerWithSpans.from_pretrained('bert-base-uncased')
    # tokenizer = BertTokenizerWithSpans.from_pretrained('bert-base-uncased')
    # model = SBERT.from_pretrained(config).to(device)
    model = SBERT.from_pretrained(config).to('cuda:0')
    logger.info('pretrained sbert model loaded')
    model.eval()
    return config, tokenizer, model


def get_lstm_model(pretrained_dataset_name, base_dataset_name=None, device=None):
    config = LSTMConfig(
        dataset_name=pretrained_dataset_name,
        embed_size=SequenceMiningConfig.EMBEDDING_SIZE,
        max_vocab_size=SequenceMiningConfig.MAX_VOCAB_SIZE,
        max_len=SequenceMiningConfig.MAXLEN,
        dropout_rate=SequenceMiningConfig.DROPOUT_RATE,
        class_names=SequenceMiningConfig.get_class_names(base_dataset_name))
    tokenizer = BasicTokenizer()
    model = BiLSTMModel.from_pretrained(config).to(device)
    logger.info('pretrained lstm model loaded')
    model.eval()
    return config, tokenizer, model

def get_model(
    model_name: str,
    pretrained_dataset_name: str,
    device: Union[str, device],
    base_dataset_name: str) -> Tuple[Union[BertConfig, ExpredConfig],
                                     Union[BertTokenizerWithSpans, BertTokenizer],
                                     Union[Expred, torch.nn.Module]]:
    if model_name == 'expred':
        return get_expred_model(pretrained_dataset_name, base_dataset_name=base_dataset_name, device=device)
    elif model_name == 'bert':
        return get_bert_model(pretrained_dataset_name, base_dataset_name=base_dataset_name, device=device)
    elif model_name == 'lstm':
        return get_lstm_model(pretrained_dataset_name, base_dataset_name=base_dataset_name, device=device)
    elif model_name == 'sbert':
        return get_sbert_model(pretrained_dataset_name, base_dataset_name=base_dataset_name, device=device)
    else:
        raise NotImplementedError(f'model {model_name} not supported')


def _do_lstm_prediction(model: BiLSTMModel, config: LSTMConfig, tokenizer: BasicTokenizer, data):
    lstm_batch = LSTMBatch.from_bert_dataset(
            data, config,
            embedding=model.embedding, 
            tokenizer=tokenizer).preprocess()
    lstm_outputs = model(lstm_batch.input_embeddings)
    
    cls_pred = torch.softmax(lstm_outputs['cls_pred_logit'], dim=-1).tolist()
    doc_embedding = lstm_outputs['last_hidden'].detach().cpu().numpy()
    decoded_doc = lstm_batch.get_decoded_docs()
    decoded_cls_pred = lstm_batch.get_decoded_cls_preds(cls_pred)
    return cls_pred, doc_embedding, decoded_doc, decoded_cls_pred


def _do_sbert_prediction(model: SBERT, config: SBERTConfig, tokenizer:BertTokenizerWithSpans, data,
                         return_concatenated_embedding=False):
    sbert_batch = [
        InputExample(
            texts=(' '.join(q), ' '.join(chain.from_iterable(d))),
            label=config.class_names.index(l))
        for _, q, d, l in data]
    bpe_encoded_docs = [tokenizer(' '.join(chain.from_iterable(d)), add_special_tokens=False)['input_ids'][:512] for _, _, d, _ in data]
    decoded_docs = [basic_tokenizer.tokenize(tokenizer.decode(d)) for d in bpe_encoded_docs]

    logger.debug(f'stage 2 preprocess done')
    # 3 mins
    ret = model.predict_on_batch(
        sbert_batch,
        return_concatenated_embedding=return_concatenated_embedding)
    logger.debug(f'predicted on batch')

    if return_concatenated_embedding:
        cls_pred, _, concat_embedding = ret
    else:
        cls_pred, _ = ret
        concat_embedding = [None] * len(cls_pred)

    decoded_cls_pred = [config.class_names[p] for p in torch.argmax(cls_pred, dim=1)]
    cls_pred = torch.softmax(cls_pred, dim=-1).cpu().tolist()
    logger.debug(f'postprocess finished')

    return cls_pred, concat_embedding, decoded_docs, decoded_cls_pred


def _do_prediction(
    model: torch.nn.Module,
    config,
    tokenizer,
    data: List[Tuple[Literal[0], List[str], list[List[str]], str]],  # ann_id === 0, query, doc, label
    rationale_mask_char):
    logger.debug(f'inside _do_prediction')
    device = next(model.parameters()).device
    if isinstance(model, Expred):
        expred_inputs = ExpredInput.from_bert_dataset(
            data,
            config,
            tokenizer=tokenizer).preprocess().to(device)
        expred_outputs = model(expred_inputs)

        cls_pred = expred_outputs['cls_preds']['cls_pred'].tolist()
        doc_embedding = np.array(
            expred_outputs['cls_preds']['cls_hiddens'].detach().cpu().numpy())
        decoded_doc = expred_inputs.get_decoded_exp_preds(
            expred_outputs, 
            collapse_multiple_wildcards=True,
            rationale_mask_char=rationale_mask_char
        )
        decoded_cls_pred = expred_inputs.get_decoded_cls_preds(expred_outputs)
        hard_exp_preds = expred_outputs['mtl_preds']['exp_preds'].detach().to('cpu')
        exp_output = expred_inputs.select_no_overheads(hard_exp_preds)
        exp_output = list(map(lambda x: x.numpy(), exp_output))
    elif isinstance(model, BertClassifier) or isinstance(model, CLSModel):
        bert_inputs = BertInput.from_bert_dataset(
            data, config, tokenizer=tokenizer).preprocess().to(device)
        if isinstance(model, CLSModel):
            model_inputs = bert_inputs.bert_inputs
        else:
            model_inputs = bert_inputs
        bert_outputs = model(model_inputs)

        if isinstance(model, CLSModel):  # for expred compatibility
            bert_outputs['cls_preds'] = bert_outputs['cls_pred']
        cls_pred = bert_outputs['cls_preds'].tolist()
        doc_embedding = bert_outputs['bert_embeddings'].detach().cpu().numpy()
        decoded_doc = bert_inputs.get_decoded_docs()
        decoded_cls_pred = bert_inputs.get_decoded_cls_preds(bert_outputs)
    elif isinstance(model, BiLSTMModel):
        (
            cls_pred,
            doc_embedding, 
            decoded_doc, 
            decoded_cls_pred
        ) = _do_lstm_prediction(model, config, tokenizer, data)
    elif isinstance(model, SBERT):
        (
            cls_pred,
            doc_embedding,
            decoded_doc,
            decoded_cls_pred
        ) = _do_sbert_prediction(model, config, tokenizer, data)
    else:
        raise NotImplementedError
    ret = {
        'cls_pred': cls_pred,
        'decoded_cls_pred': decoded_cls_pred,
        'doc_embedding': doc_embedding,
        'decoded_doc': decoded_doc,
        'exp_output': None
    }
    if isinstance(model, Expred):
        ret['exp_output'] = exp_output
    return ret


def convert_to_word_split_strings(
    _inputs: Union[
        # word-split sentences
        List[List[List[str]]],
        # word-split query
        List[List[str]],
        # list of query/document strings
        List[str]]) \
        -> List[List[str]]:  # list of query/document strings
    ret = []
    for _input in _inputs:
        if _input is None:  # default sentiment analysis queries
            ret.append('what is the sentiment of this review ?'.split())
        elif isinstance(_input, str):  # list of document strings
            ret.append(_input.split())
        elif isinstance(_input[0], list):  # word-split sentences
            ret.append(chain.from_iterable(_input))
        else:  # word-split query
            ret.append(_input)
    return ret


# TODO: fix the flag
# TODO: tear down to multiple functions
def do_prediction(model: torch.nn.Module,
                  config: BertConfig,
                  tokenizer: BertTokenizer,
                  docs_or_dataset: Union[List[List[List[str]]],  # word-split sentences
                                         BertDataset,
                                         List[str]],  # list of document strings
                  batch_size=16,
                  queries_or_none: Union[List[str], List[List[str]]] = None,
                  return_queries=False,
                  rationale_mask_char=None,
                  tqdm_desc:str=None,
                  verbose:bool=True,
                  return_dict=False):
    logger.debug(f'inside do_prediction')
    if isinstance(docs_or_dataset, BertDataset):
        data = docs_or_dataset
        queries = [d[1] for d in data]
    else:
        queries = queries_or_none if queries_or_none is not None else [None] * len(docs_or_dataset)
        queries = convert_to_word_split_strings(queries)
        docs = convert_to_word_split_strings(docs_or_dataset)
        data = [(0, q, [d], config.class_names[0]) for q, d in zip(queries, docs)]
        logger.debug(f'doing prediction on documents = {docs}')
        logger.debug(f'doing prediction on queries = {queries}')
    data_iter = iter(data)
    cls_preds = []
    decoded_cls_preds = []
    input_embeddings = []
    decoded_docs = []
    exp_outputs = []
    logger.debug(f'finish preprocessing stage 1')
    logger.debug(f'batch size = {batch_size}')
    logger.debug(f'len(data) = {len(data)}')
    if len(data) > 2 * batch_size and verbose:
        iterator = tqdm.tqdm(range(0, len(data), batch_size), desc=tqdm_desc)
    else:
        iterator = range(0, len(data), batch_size)
    for i in iterator:
        batch = [next(data_iter) for _ in range(i, min(i + batch_size, len(data)))]

        pred_ret = _do_prediction(
            model, 
            config,
            tokenizer,
            batch,
            rationale_mask_char
        )
        cls_preds.extend(pred_ret['cls_pred'])
        decoded_cls_preds.extend(pred_ret['decoded_cls_pred'])
        input_embeddings.extend(pred_ret['doc_embedding'])
        decoded_docs.extend(pred_ret['decoded_doc'])
        if isinstance(model, Expred):
            exp_outputs.extend(pred_ret['exp_output'])

    if return_queries:
        decoded_inputs = list(zip(queries, decoded_docs))
    else:
        decoded_inputs = decoded_docs
    if return_dict:
        return {
            'cls_preds': cls_preds,
            'decoded_cls_preds': decoded_cls_preds,
            'input_embeddings': input_embeddings,
            'decoded_inputs': decoded_inputs,
            'exp_outputs': exp_outputs
        }
    return cls_preds, decoded_cls_preds, input_embeddings, decoded_inputs


def sentence_seek(sentences: List[List[str]],
                  point: int, sent_idx_st: int = 0, sentence_cnt: int=0) -> Tuple[int, int]:
    sent_offset_st = sum(map(len, sentences[:sent_idx_st]))
    while len(sentences[sent_idx_st]) + sent_offset_st < point and sent_idx_st < len(sentences):
        sent_offset_st += len(sentences[sent_idx_st])
        sent_idx_st += 1
    sent_idx_st += sentence_cnt
    if sentence_cnt > 0:
        sent_offset_st += sum(map(len, sentences[sent_idx_st: sent_idx_st + sentence_cnt]))
    elif sentence_cnt < 0:
        sent_offset_st -= sum(map(len, sentences[sent_idx_st + sentence_cnt: sent_idx_st]))
    return sent_offset_st, sent_idx_st


def split_doc_to_sentences(doc: List[str]) -> List[List[str]]:
    sents = [[]]
    for word in doc:
        sents[-1].append(word)
        if word == '.':
            sents.append([])
    if len(sents[-1]) == 0:
        sents = sents[:-1]
    return sents


def join_sentences_to_docs(sents: List[List[str]]) -> List[str]:
    return list(chain.from_iterable(sents))


def locate_substr(wordlist: List[str], pattern: List[str]) -> Tuple[int, int]:
    s = ' '.join(wordlist)
    p = ' '.join(pattern)
    pos = s.find(p)
    words_prev = s[:pos].split()
    start = len(words_prev)
    end = start + len(pattern)
    return start, end


def locate_tuple(word_list_tuple: Tuple[List[str], List[str]], q_pattern: List[str], d_pattern: List[str]) -> Tuple[int, int, int, int]:
    query, doc = word_list_tuple
    q_start, q_end = locate_substr(query, q_pattern)
    d_start, d_end = locate_substr(doc, d_pattern)
    return q_start, q_end, d_start, d_end


def abs_diff(pred):
    return abs(pred[0] - pred[1]) < SequenceMiningConfig.NEUTRAL_THRESHOLD


def mean(data) -> float:
    return sum(data) / float(len(data))


def stddev(data) -> float:
    var = sum(map(lambda x: (x - mean(data)) ** 2, data))
    return (var / len(data)) ** 0.5


def stddev_criterion(data) -> bool:
    return stddev([d[0] for d in data]) < SequenceMiningConfig.STDDEV_THRESHOLD


def mean_criterion(data, **kwargs) -> bool:
    m = mean([d[0] for d in data])
    if not hasattr(SequenceMiningConfig, "MEAN_THRESHOLD"):
        mean_threshold = kwargs.get('mean_threshold')
    else:
        mean_threshold = SequenceMiningConfig.MEAN_THRESHOLD
    return m > mean_threshold or m < 1 - mean_threshold


def mean_stddev_criterion(data) -> bool:
    if mean_criterion(data):
        return stddev_criterion(data)
    return False


def grep_tuple(query_seq: str, doc_seq: str, fname: str, count_only=True) -> List[str]:
    grep_pattern = re.escape(query_seq) + ' .*\[SEP\].* ' + re.escape(doc_seq)
    command = ['grep', '-Ew' + ('c' if count_only else ''),
               '--', grep_pattern, fname]
    # print(command)
    try:
        run = subprocess.run(command, stdout=subprocess.PIPE)
        if count_only:
            ret = int(run.stdout.decode('utf-8'))
        else:
            ret = run.stdout.decode('utf-8').strip().split('\n')
    except AttributeError:
        print(f'attribute error: {run.stderr} at {query_seq} - {doc_seq}')
        raise AttributeError
    return ret


def init_seeds():
    global_seed = SequenceMiningConfig.SEED

    random.seed(global_seed)
    np.random.seed(global_seed)
    torch.manual_seed(global_seed)
    torch.cuda.manual_seed(global_seed)
    torch.use_deterministic_algorithms(True)
    os.environ['PYTHONHASHSEED'] = str(global_seed)
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"  # https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility

 